(***************************************************************************
* Generic Variables (Atoms) for Programming Language Metatheory            *
* Brian Aydemir & Arthur Charguéraud, March 2007, Coq v8.1                 *
***************************************************************************)

Set Implicit Arguments.
Require Import List Max Omega OrderedType OrderedTypeEx.
Require Import Lib_Tactic Lib_ListFacts Lib_FinSet Lib_FinSetImpl.


(* ********************************************************************** *)
(** * Abstract Definition Atoms *)

Module Type ATOMS.

(** We leave the type of atoms abstract. *)

Parameter atom : Set.

(** This type is inhabited. *)

Parameter atom_default : atom.

(** Atoms are ordered. *)

Declare Module Atom_as_OT : UsualOrderedType with Definition t := atom.

(** We can form sets of atoms. *)

Declare Module Import AtomSet : FinSet with Module E := Atom_as_OT.
Open Local Scope set_scope.

Definition atoms := AtomSet.S.t.

(** Finally, we have a means of generating fresh atoms. *)

Parameter atom_generate : atoms -> atom.
Parameter atom_generate_spec : forall E, (atom_generate E) \notin E.
Parameter atom_fresh : forall (L : atoms), { x : atom | x \notin L }.

End ATOMS.


(* ********************************************************************** *)
(** * Concrete Implementation of Atoms *)

Module Atoms : ATOMS.

Definition atom := nat.

Definition atom_default : atom := O.

Module Atom_as_OT := Nat_as_OT.

Module Import AtomSet : FinSet with Module E := Atom_as_OT :=
  Lib_FinSetImpl.Make Atom_as_OT.
Open Local Scope set_scope.

Definition atoms := AtomSet.S.t.

Open Scope nat_scope.

Lemma max_lt_l :
  forall (x y z : nat), x <= y -> x <= max y z.
Proof.
  induction x; auto with arith.
  induction y; induction z; simpl; auto with arith.
Qed.

Lemma finite_nat_list_max : forall (l : list nat),
  { n : nat | forall x, In x l -> x <= n }.
Proof.
  induction l as [ | l ls IHl ].
  exists 0; intros x H; inversion H.
  inversion IHl as [x H].
  exists (max x l); intros y J; simpl in J; inversion J.
    subst; auto with arith.
    assert (y <= x); auto using max_lt_l.
Qed.

Lemma finite_nat_list_max' : forall (l : list nat),
  { n : nat | ~ In n l }.
Proof.
  intros l.
  case (finite_nat_list_max l); intros x H.
  exists (S x).
  intros J.
  assert (K := H _ J); omega.
Qed.

Definition atom_generate (L : atoms) : atom :=
  proj1_sig (finite_nat_list_max' (S.elements L)).

Lemma atom_generate_spec : forall E, (atom_generate E) \notin E.
Proof.
  unfold atom_generate. intros E.
  destruct (finite_nat_list_max' (S.elements E)) as [n pf].
  simpl. intros a.
  assert (In n (S.elements E)). rewrite <- InA_iff_In.
  auto using S.elements_1.
  intuition.
Qed.

Lemma atom_fresh : forall (L : atoms), { x : atom | x \notin L }.
Proof.
  intros L. exists (atom_generate L). apply atom_generate_spec.
Qed.

End Atoms.


(* ********************************************************************** *)
(** * Properties of atoms *)

Export Atoms.
Export Atoms.AtomSet.
Module Export AtomSetFacts := FinSetFacts AtomSet.

Open Scope set_scope.

(** Equality on atoms is decidable. *)

Module Import Atom_as_OT_Facts := OrderedTypeFacts Atoms.Atom_as_OT.

Lemma eq_atom_dec : forall x y : atom, {x = y} + {x <> y}.
Proof.
  exact Atom_as_OT_Facts.eq_dec.
Qed.

(* ********************************************************************** *)
(** ** Dealing with list of atoms *)

(** Converting a list of atoms into a set of atoms *)

Fixpoint atoms_of_list (xs : list atom) :=
  match xs with
  | nil => {}
  | x::xs' => {{x}} \u atoms_of_list xs'
  end.

(** Definition of a list of n atoms fresh from L *)

Definition fresh_atoms_length n L xs :=
  disjoint (atoms_of_list xs) L /\ length xs = n.

Section ListAtomProperties.

Hint Extern 1 (disjoint _ _) => disjoint_solve.
Hint Extern 1 (_ <> _ :> atom) => notin_neq_solve.

(** List of given size containing fresh atoms can be constructed *)

Lemma fresh_atoms_length_exists : forall L n,
  exists xs, fresh_atoms_length n L xs.
Proof.
  unfold fresh_atoms_length. induction n.
  exists (nil:list atom). simpl*.
  destruct IHn as [Xs [H1 H2]].
   destruct (atom_fresh L) as [X H3].
   exists (X :: Xs). simpl*.
Qed.

(** A atom fresh from a list of atoms is distinct from the 
  result of the application of List.nth to that list *)

Lemma fresh_list_atoms : forall (x:atom) K n,
  x \notin (atoms_of_list K \u {{atom_default}}) -> 
  nth n K atom_default <> x.
Proof.
  induction K; simpl; intros n Fr; destruct n; auto.
  apply IHK. notin_solve.
Qed.

End ListAtomProperties.


(* ********************************************************************** *)
(** ** Tactics: Case Analysis on Atoms *)

(** We define notations for the equality of atoms (our free atoms)
  and for the equality of naturals (our bound atoms represented using
  de Bruijn indices). *)

Notation "x == y" := (eq_atom_dec x y) (at level 67).
Notation "i === j" := (Peano_dec.eq_nat_dec i j) (at level 67).

(** Tactic for comparing two bound or free atoms. *)

Ltac case_nat :=
  let destr x y := destruct (x === y); [try subst x | idtac] in
  match goal with
  | H: context [?x === ?y] |- _ => destr x y
  | |- context [?x === ?y]      => destr x y
  end.

Tactic Notation "case_nat" "*" := case_nat; auto*.

Tactic Notation "case_atom" :=
  let destr x y := destruct (x == y); [try subst x | idtac] in
  match goal with
  | H: context [?x == ?y] |- _ => destr x y
  | |- context [?x == ?y]      => destr x y
  end.

Tactic Notation "case_atom" "*" := case_atom; auto*.


(* ********************************************************************** *)
(** ** Tactics: Picking Names Fresh from the Context *)

(** [gather_atoms_for_type T F] return the union of all the finite sets
  of atoms [F x] where [x] is a atom from the context such that
  [F x] type checks. In other words [x] has to be of the type of the
  argument of [F]. The resulting union of sets does not contain any
  duplicated item. This tactic is an extreme piece of hacking necessary
  because the tactic language does not support a "fold" operation on
  the context. *)

Ltac gather_atoms_with F :=
  let rec gather V :=
    match goal with
    | H: ?S |- _ =>
      let FH := constr:(F H) in
      match V with
      | {} => gather FH
      | context [FH] => fail 1
      | _ => gather (FH \u V)
      end
    | _ => V
    end in
  let L := gather {} in eval simpl in L.

(** [beautify_fset V] assumes that [V] is built as a union of finite
  sets and return the same set cleaned up: empty sets are removed and
  items are laid out in a nicely parenthesized way *)

Ltac beautify_fset V :=
  let rec go Acc E :=
     match E with
     | ?E1 \u ?E2 => let Acc1 := go Acc E1 in
                     go Acc1 E2
     | {}  => Acc
     | ?E1 => match Acc with
              | {} => E1
              | _ => constr:(Acc \u E1)
              end
     end
  in go {} V.

(** [pick_fresh_gen L Y] expects [L] to be a finite set of atoms
  and adds to the context a atom with name [Y] and a proof that
  [Y] is fresh for [L]. *)

Ltac pick_fresh_gen L Y :=
  let Fr := fresh "Fr" in
  let L := beautify_fset L in
  (destruct (atom_fresh L) as [Y Fr]).

(** Demo of pick_fresh_gen *)

Ltac test_pick_fresh_filter Y :=
  let A := gather_atoms_with (fun x : atoms => x) in
  let B := gather_atoms_with (fun x : atom => {{ x }}) in
  let C := gather_atoms_with (fun x : atom => {}) in
  pick_fresh_gen (A \u B \u C) Y.

Lemma test_pick_fresh : forall (x y z : atom) (L1 L2 L3: atoms), True.
Proof.
  intros. test_pick_fresh_filter k. auto.
Qed.

(** The above invokation of [pick_fresh] generates a
  atom [k] and the hypothesis
  [k \notin L1 \u L2 \u L3 \u {{x}} \u {{y}} \u {{z}}] *)


(* ********************************************************************** *)
(** ** Tactics: Applying Lemmas With Quantification Over Cofinite Sets *)

(** [apply_fresh_base] tactic is a helper to build tactics that apply an
  inductive constructor whose first argument should be instanciated
  by the set of names already used in the context. Those names should
  be returned by the [gather] tactic given in argument. For each premise
  of the inductive rule starting with an universal quantification of names
  outside the set of names instanciated, a subgoal with be generated by
  the application of the rule, and in those subgoal we introduce the name
  quantified as well as its proof of freshness. *)

Ltac apply_fresh_base lemma gather atom_name :=
  let L0 := gather in let L := beautify_fset L0 in
  first [apply (@lemma L) | eapply (@lemma L)];
  try match goal with |- forall _, _ \notin _ -> _ =>
    let Fr := fresh "Fr" in intros atom_name Fr end.

(** [inst_notin H y as H'] expects [H] to be of the form
  [forall x, x \notin L, P x] and creates an hypothesis [H']
  of type [P y]. It tries to prove the subgoal [y \notin L]
  by [auto]. This tactic is very useful to apply induction
  hypotheses given in the cases with binders. *)

Tactic Notation "inst_notin" constr(lemma) constr(atom)
                "as" ident(hyp_name) :=
  let go L := let Fr := fresh in assert (Fr : atom \notin L);
     [ auto | poses hyp_name (@lemma atom Fr); clear Fr ] in
  match type of lemma with
  | forall _, _ \notin ?L -> _ => go L
  | forall _, (_ \in ?L -> False) -> _ => go L
  end.

Tactic Notation "inst_notin" "*" constr(lemma) constr(atom)
                "as" ident(hyp_name) :=
  inst_notin lemma atom as hyp_name; auto*.
