(* 
 * Core.v  -- A metatheory of the core SOL language.
 *
 * Steve Zdancewic
 * Jeffrey A. Vaughan
 * Limin Jia
 * August 2007
 *)  


Set Implicit Arguments.

Require Import AdditionalTactics.
Require Import List Metatheory.
Require Import Lib_ListFacts.
Require Omega.

(*** An encoding of the core language in PTS-style, using multiple typing
 *   judgments to differentiate syntactic classes.
 *
 *   This approach is inspired by the PTS work and by the Henk intermediate
 *   language proposed by Simon Peyton Jones and Erik Meijer.
 *)


(** Terms -- these are the common datatype on which types, expressions, 
 *  kinds, etc. are built.
 *)

Definition bv  := nat.
Definition con := nat.
Definition ctr :=  prod nat  nat. 
(* the second nat indicate the number of params*)
(* e.g. cons = (n, 1), because List: Type -> Type*)

Inductive tm : Set :=
  | tm_fv   : atom -> tm            (* free variables *)
  | tm_bv   : bv -> tm              (* bound variables *)
  | tm_con  : con -> tm             (* built-in constants like Type, Prop, says, sign, or IO *)
  | tm_ctr  : ctr -> tm             (* user defined constructors, like Bool and List *)
  | tm_app  : tm -> tm -> tm        (* applications like (e1 e2) or e1[t]*)
  | tm_abs  : tm -> tm -> tm        (* abstractions \ (x:t).e   or  /\ (a::K).e *)
  | tm_arr  : tm -> tm -> tm        (* dependent products (x:t) -> e *)
  | tm_mat  : tm -> tm -> tm -> tm        (* single-level pattern matching *)
  | tm_brn  : ctr -> tm -> tm -> tm  (* branch in the case construct *)
  | tm_cast : tm -> tm -> tm        (* cast(e,t) : t *)
.

(** Intended invariants of terms:  *)

(* Pattern matches are to be represented as a list of tm_brn terms:

  Abstract syntax:
  match e {
  | ID1 => body1
  | ID2 => body2
  }

  Term representation
  tm_mat e result_type
  tm_brn ID1 (body1) (
  tm_brn ID2 (body2) (
  TM_NONE                           
  ))

  TM_NONE is defined to be [tm_con 0]

*)


(* Applications [tm_app e1 e2] mean e1 applied to e2 where e1:t. 

   The app rule looks like
       
    S; E |- e1: (x:u2) -> u     S; E |- e2: u2    ...
   ----------------------------------------------------
     S; E |- e1 e2 : u{e2/x},

   and is ambiguous in two ways.  First, the form of u needs to be
   guessed.  Second, because we have a conversion rule, we might be able
   to pick a wide variety of u2 types.  Making annontation t be the 
   full type of of the funtion (e.g. (x:u2)->u) clarifies both.
   Additionaly, if we were to make the annotation just u, it would not
   be locally closed and this would be tantamount to introducing a new
   binding construct.

   NOTE: if we add an explicit constructor for conversion (casts) we might 
   be able to eliminate the need for annotations on application:

   S; E |- e : t
   S; E |- t : u
   S; E |- s = t : u
   -------------------
   S; E |- conv(e, t, s) : s

*)


(* NOTE: the above definition includes space for a type annotation on application nodes to 
   facilitate type checking in the presence of dependent types.  Another option would be to add 
   an explicit tm_ann : tm -> tm -> tm constructor that annotates a given term with another. 

   Which is better?

   JAV: I think the explicit anotation is better.  Otherwise we need to
     worry about typing derivation that igore the annotation by
     using conversion immediately after the annotation.
*)

Lemma eq_ctr_dec:
   forall (c c1:ctr), {c=c1} + {~c=c1}.
Proof. 
 decide equality; try decide equality; auto.
Qed.

(* Term equality is decidable *)
Lemma eq_tm_dec:
  forall (t s: tm), {t = s} + {~t = s}.
Proof.
  decide equality; try solve [decide equality]; auto using eq_atom_dec, eq_ctr_dec.
Qed.


(***********************************)
(*** Syntactic operations on terms *)

(** OPEN: Substitute term s for bound variable j in term t  *)
Fixpoint open_tm_aux (j:nat) (s t:tm) {struct t} : tm :=
  match t with
  | tm_fv _         => t
  | tm_bv n         => if (n === j) then s else t
  | tm_con _        => t
  | tm_ctr _        => t
  | tm_app t1 t2    => tm_app (open_tm_aux j s t1) (open_tm_aux j s t2) 
  | tm_abs t1 t2    => tm_abs (open_tm_aux j s t1) (open_tm_aux (1+j) s t2)
  | tm_arr t1 t2    => tm_arr (open_tm_aux j s t1) (open_tm_aux (1+j) s t2)
  | tm_mat t1 t2 t3    => tm_mat (open_tm_aux j s t1) (open_tm_aux j s t2) (open_tm_aux j s t3)
  | tm_brn c t2 t3  => tm_brn c (open_tm_aux j s t2) (open_tm_aux j s t3)
  | tm_cast t1 t2   => tm_cast (open_tm_aux j s t1) (open_tm_aux j s t2)
  end.

Definition open_tm := open_tm_aux 0.


(** Free variables *)

Fixpoint fv (t:tm) : atoms :=
  match t with
  | tm_fv a         => {{ a }}
  | tm_bv _         => {}
  | tm_con _        => {}
  | tm_ctr _        => {}
  | tm_app t1 t2    => (fv t1) \u (fv t2)
  | tm_abs t1 t2    => (fv t1) \u (fv t2)
  | tm_arr t1 t2    => (fv t1) \u (fv t2)
  | tm_mat t1 t2 t3   => (fv t1) \u (fv t2) \u (fv t3)
  | tm_brn _ t2 t3  => (fv t2) \u (fv t3)
  | tm_cast t1 t2   => (fv t1) \u (fv t2)
  end.


(** Free variable substitution *)

Fixpoint subst_tm (a:atom) (s t:tm) {struct t} : tm :=
  match t with
  | tm_fv b         => if (b == a) then s else t
  | tm_bv _         => t
  | tm_con _        => t
  | tm_ctr _        => t
  | tm_app t1 t2    => tm_app (subst_tm a s t1) (subst_tm a s t2) 
  | tm_abs t1 t2    => tm_abs (subst_tm a s t1) (subst_tm a s t2)
  | tm_arr t1 t2    => tm_arr (subst_tm a s t1) (subst_tm a s t2)
  | tm_mat t1 t2 t3    => tm_mat (subst_tm a s t1) (subst_tm a s t2)  (subst_tm a s t3)
  | tm_brn c t2 t3  => tm_brn c (subst_tm a s t2) (subst_tm a s t3) 
  | tm_cast t1 t2   => tm_cast (subst_tm a s t1) (subst_tm a s t2) 
  end.

Check  subst_tm.
(** Local closure predicate *)

Inductive lc : tm -> Prop :=
  | lc_fv  : forall a,  lc (tm_fv a)
  | lc_con : forall a,  lc (tm_con a)
  | lc_ctr : forall a,  lc (tm_ctr a)
  | lc_app : forall t1 t2,   lc t1 -> lc t2 -> lc (tm_app t1 t2)
  | lc_abs : forall L t1 t2, lc t1 -> 
                (forall x, x \notin L -> lc (open_tm (tm_fv x) t2)) ->
                lc (tm_abs t1 t2)
  | lc_arr : forall L t1 t2, lc t1 -> 
                (forall x, x \notin L -> lc (open_tm (tm_fv x) t2)) ->
                lc (tm_arr t1 t2)
  | lc_mat : forall t1 t2 t3, lc t1 -> lc t2 -> lc t3 -> lc (tm_mat t1 t2 t3)
  | lc_brn : forall c t2 t3,  lc t2 -> lc t3 -> lc (tm_brn c t2 t3)
  | lc_cast : forall t1 t2,  lc t1 -> lc t2 -> lc (tm_cast t1 t2)
.

Hint Constructors lc.

Bind Scope tm_scope with tm.

Notation "t ^^ u"       := (open_tm u t)    (at level 67) : tm_scope.
Notation "[ x ~> u ] t" := (subst_tm x u t) (at level 68) : tm_scope.


(******************)
(*** Environments *)

(** Note: we have to define environments here because we need
    to define the gather_atoms tactic, and it depends on the 
    definition of environments.  This is an example of where 
    automation interferes (slightly) with the desired structure of 
    the theory.  An alternative would be to define a separate tactic
    later that works for environments too... *)

(** Bindings in the context are either variables and their types
    or equality assumptions *)
Inductive bnd : Set :=
| bnd_var : tm -> bnd
| bnd_eq  : tm -> tm -> tm -> bnd
.

Notation "a ~: t" := (a ~ bnd_var t)
  (at level 31, left associativity) : env_scope.

(* TODO? Add notation for equality constraints in the context *)

(** bnd equality is decidable *)
Lemma eq_bnd_dec: forall (a b:bnd), {a = b} + {~a = b}.
Proof.
  decide equality; auto using eq_tm_dec.
Qed.

Definition binds_var x t E :=
  binds x (bnd_var t) E.

Definition binds_eq x e1 e2 t E :=
  binds x (bnd_eq e1 e2 t) E.

Definition env := env bnd.

(* Limin: for substitution Lemma *)
Definition subst_bnd (a: atom) (s: tm) (b:bnd): bnd :=
  match b with 
  | bnd_var t => bnd_var (subst_tm a s t)
  | bnd_eq e1 e2 t  => bnd_eq (subst_tm a s e1) (subst_tm a s e2) (subst_tm a s t)
  end.

Definition subst_env (a: atom) (s: tm) (E: env) :=
  map (subst_bnd a s) E.

(**** Picking fresh atoms **********)
Ltac gather_atoms :=
  let A := gather_atoms_with (fun x : atoms => x) in
  let B := gather_atoms_with (fun x : atom  => {{ x }}) in
  let C := gather_atoms_with (fun x : tm    => fv x) in
  let D := gather_atoms_with (fun x : env   => dom x) in
  constr:(A \u B \u C \u D).

(** Tactic invocation [pick_fresh x] introduces a new atom, called [x],
    such that [x] is fresh from the set of atoms returned by [gather_atoms].*)

Ltac pick_fresh x :=
  let L := gather_atoms in (pick_fresh_gen L x).

Tactic Notation "apply_fresh" constr(T) "as" ident(x) :=
  apply_fresh_base T gather_atoms x.

Tactic Notation "apply_fresh" "*" constr(T) "as" ident(x) :=
  apply_fresh T as x; auto*.


(**************************************************)
Section Syntax.
Set Implicit Arguments.
Open Scope tm_scope.

(*** Properties of open and subst for tm *)

(** Substitution on indices is identity on well-formed expressions. *)
Let open_lc_id_aux : forall (t:tm) j (u v:tm) i, 
  i <> j -> (open_tm_aux j v t = open_tm_aux i u (open_tm_aux j v t)) ->
  t = open_tm_aux i u t.
Proof with auto*.
  induction t; intros j v u i Neq H; simpl in *; inversion H;
       f_equal...
  destruct (b === j)...
  destruct (b === i)...
  congruence.
Qed.

(** openening a locally closed tm is the identity *)
Lemma open_lc_id : forall (t u:tm),
  lc t -> forall k, t = (open_tm_aux k u) t.
Proof with auto*.
  intros t u H. induction H; intro k; simpl; f_equal...
  (* tm_abs case *)
  pick_fresh X. apply(@open_lc_id_aux t2 0 u (tm_fv X))...
  (* tm_arr case *)
  pick_fresh X. apply(@open_lc_id_aux t2 0 u (tm_fv X))...
Qed.

(** substituting for a fresh variable in t is the identity *)
Lemma subst_fresh : forall x u t,
  x \notin fv t -> [x~>u]t = t.
Proof with auto*.
  induction t; unfold fv in *; simpl; intro H; f_equal...
  (* tm_fv case *)
  destruct (a == x); subst... notin_contradiction.
Qed.

Let subst_open_aux : forall (s t:tm) (x:atom) (u:tm) (k:nat),
  lc u ->
  [x~>u](open_tm_aux k s t) = open_tm_aux k ([x~>u]s) ([x~>u]t).
Proof.
  introv LCu. generalize k.
  induction t; intros j; simpls; f_equal*.
  destruct (a == x); auto*. 
  rewrite <- open_lc_id; auto.
  destruct (b === j); auto.
Qed.

Lemma subst_open : forall (s t:tm) (x:atom) (u:tm),
  lc u ->
  [x~>u](s^^t) = ([x~>u]s) ^^ ([x~>u]t).
Proof.
  unfold open_tm. auto*.
Qed.

Lemma subst_open_commute : forall (x y:atom) (u t:tm),
  (x <> y) -> lc u -> open_tm (tm_fv y) ([x~>u]t) = [x~>u](open_tm (tm_fv y) t).
Proof.
  introv Neq LCu. rewrite* subst_open.
  simpl. destruct (y == x); subst; auto*.
Qed.

Lemma subst_intro : forall (x:atom) (t u:tm),
  x \notin fv t -> lc u ->
  (t ^^ u) = [x~>u](open_tm (tm_fv x) t).
Proof.
  introv Nin LCu. rewrite* subst_open.
  rewrite* (@subst_fresh x u t). simpl.
  destruct (x == x); auto*.
Qed.

Lemma subst_lc : forall (t:tm) (x:atom) (u:tm),
  lc u -> lc t -> lc ([x~>u]t).
Proof.
  induction 2; simpl in *; auto*.
  destruct (a == x); auto*.
  apply_fresh lc_abs as y. auto. rewrite* subst_open_commute.
  apply_fresh lc_arr as y. auto. rewrite* subst_open_commute.
Qed.

Hint Resolve subst_lc.

Lemma lc_abs_ex : forall (t u:tm) (x:atom),
  x \notin fv t -> lc u -> lc (t ^^ (tm_fv x)) -> lc (tm_abs u t).
Proof.
  intros t u x Nin LCu LCtx.
  apply_fresh(lc_abs) as y; auto.
  rewrite (@subst_intro x); auto.
Qed.


Lemma lc_arr_ex : forall (t u:tm) (x:atom),
  x \notin fv t -> lc u -> lc (t ^^ (tm_fv x)) -> lc (tm_arr u t).
Proof.
  intros t u x Nin LCu LCtx.
  apply_fresh(lc_arr) as y; auto.
  rewrite (@subst_intro x); auto.
Qed.
  
Lemma lc_abs_all : forall (t u:tm),
  lc (tm_abs u t) -> forall (x:atom), x \notin fv t -> lc (t ^^ (tm_fv x)). 
Proof.
  intros t u LCabs x NIx.
  inversion LCabs. subst.
  pick_fresh y.  
  forward~ (H2 y).
  rewrite (@subst_intro y t (tm_fv x)); auto.
Qed.

Lemma lc_arr_all : forall (t u:tm),
  lc (tm_arr u t) -> forall (x:atom), x \notin fv t -> lc (t ^^ (tm_fv x)). 
Proof.
  intros t u LCarr x NIx.
  inversion LCarr. subst.
  pick_fresh y.  
  forward~ (H2 y).
  rewrite (@subst_intro y t (tm_fv x)); auto.
Qed.

Close Scope tm_scope.
End Syntax.

(**************************************************)
(** Auxiliary definitions                         *) 

Definition disjoint (L1 L2:list ctr) := forall n, (In n L1) -> ~(In n L2).

Lemma disjoint_rewrite: forall L1 L2,
  (disjoint L1 L2) = (forall n, (In n L1) -> ~ (In n L2)).
auto.
Qed.


(**************************************************)
(*** Core language: Typing                        *)

(* Define the base constants of the language.
 * Each nat must be distinct.
 *) 

(* Case list terminator *)
Definition TM_NONE := tm_con 00.

(* Sort constants *)
Definition S_KIND := tm_con 01.  (* Box i.e.  Type : S_KIND  Prop : S_KIND *)

(* Kind constants *)
Definition K_TYPE := tm_con 10.
Definition K_PROP := tm_con 11.

(* Prop constants *)
Definition P_SAYS := tm_con 20.

(* Type constants *)
Definition T_PRIN := tm_con 30.
Definition T_PF   := tm_con 31.

(* Expr constants *)
Definition TM_PF_RETURN   := tm_con 40.
Definition TM_PF_BIND     := tm_con 41.
Definition TM_SAY         := tm_con 42.
Definition TM_PIF         := tm_con 43.
Definition TM_SELF        := tm_con 44.

(* Proof constants *)
Definition PF_SAYS_RETURN := tm_con 50.
Definition PF_SAYS_BIND   := tm_con 51.
Definition PF_SIGN        := tm_con 52.

(* Type and Prop constants *)
(*
Definition U_STATIC_ARG   := tm_con 60.   
Definition PE_STATIC_ABS  := tm_con 70.
Definition PE_STATIC_DIS  := tm_con 71.
Definition PE_STATIC_IF   := tm_con 72.

*) 
(* tm_arr (app U_STATIC_ARG P) T 
                  means [P] => T  *)




(**************************************************)
(* Value forms                                    *)

(* Notes: 

  1. This definition must change whenever new constants
   are added to the language.
    - other principal constants
    - built-in values like 0, 1, 2, "foo", etc. 
*)

Inductive value : tm -> Prop :=
| value_TYPE: value K_TYPE
| value_PROP: value K_PROP
| value_KIND: value S_KIND
| value_var : forall x, value (tm_fv x)
| value_abs : forall t e,
              value (tm_abs t e)
| value_app : forall t,
              applied_value t -> value t
| value_self : value (TM_SELF)

| value_arr : forall t1 t2, value (tm_arr t1 t2)

| value_sign: forall v p,  value v -> value (tm_app (tm_app PF_SIGN v)p)

| value_P_SAYS: forall v p, value  (tm_app (tm_app P_SAYS v) p)
| value_T_PRIN: value T_PRIN
| value_T_PF: forall P, value (tm_app T_PF P)

(* We don't care whether p contains redexes, since it is a proof *)
(* NOTE: Does having an embedded non-value expression cause a problem? *)
| value_TM_PF_RETURN : forall p,
                value p -> value (tm_app TM_PF_RETURN p)
| value_PF_SAYS_RETURN: forall a p,
               value (tm_app (tm_app PF_SAYS_RETURN a) p)
| value_PF_SAYS_BIND: forall e1 e2, value (tm_app (tm_app PF_SAYS_BIND e1) e2)

(* applied_values may safely use the tm_app construct without 
   introducing any evaluation *)
with applied_value : tm -> Prop :=
| applied_value_ctr : forall c, applied_value (tm_ctr c)
| applied_value_app : forall t1 t2,
                      applied_value t1 
                   -> value t2
                   -> applied_value (tm_app t1 t2)
.

Hint Constructors value applied_value.

Scheme tm_ctr_value_ind := Induction for value Sort Prop
with   ctr_tm_value_ind := Induction for applied_value Sort Prop.

Check tm_ctr_value_ind.
Definition mutual_tm_ctr_value_ind P P' :=
   fun h1 h2 h3 h4 h5 h6 h7 h8 h9 h10 h11 h12 h13 h14 h15 h16 h17 =>
     conj (tm_ctr_value_ind P P' h1 h2 h3 h4 h5 h6 h7 h8 h9 h10 h11 h12 h13 h14 h15 h16 h17)
          (ctr_tm_value_ind P P' h1 h2 h3 h4 h5 h6 h7 h8 h9 h10 h11 h12 h13 h14 h15 h16 h17).

Check mutual_tm_ctr_value_ind.

(**************************************************)
(** Signatures                                    *)

(** Datatype definitions:

    A signature [sig] is a list of bundles of mutually recursive Type (or Prop) declarations.   

    A bundle [bundle] is either a list of datatype definitions [data_decl] or an assertion
  
    Each data declaration contains: 
       - a Type (or Prop) constructor [tctr] identifier. 
       - the kind of the Type/Prop being declared
       - a list of constructor [ctr] declarations [ctr_decls]

    CONVENTION: We use the notation [tctr] to refer to a Type or Prop constructor identifier.
                We use the notation [ctr] to refer to an exp or proof constructor identifier.
                Both [tctr] and [ctr] values have Coq type [ctr] (both ranged over by [c])
 *)

(* Definition of datatype declarations in the context *)
Definition ctr_decls  := list (ctr  * tm).
Definition data_decl  := (ctr * tm * ctr_decls)%type.
Definition defns      := list data_decl.
Inductive  bundle : Set    := 
| bundle_defns  : defns -> bundle
| bundle_assert : ctr -> tm -> bundle
.
Definition sig        := list (bundle).


(* Utilities for working with signatures *)


(*Definition dom_ctr_decls (L:ctr_decls) := fst (split L).*)

Definition dom_ctr_decls(L:ctr_decls) :=
List.map (fun (x:ctr*tm) => let (n,_) := x in n) L.

Fixpoint ctr_decls_lookup (c:ctr) (cdl:ctr_decls) {struct cdl}: option tm :=
  match cdl with
    | nil          => None
    | (c', t)::cdl => 
        if (eq_ctr_dec c c')  then Some t
                    else ctr_decls_lookup c cdl
  end.


Fixpoint bundle_defns_ctr_lookup (c:ctr) (defns: defns) {struct defns}: option tm :=
  match defns with
    | nil              => None
    | (c', t, cdl)::ds => 
        if (eq_ctr_dec c c')  then Some t 
	            else match ctr_decls_lookup c cdl with
                           | None => bundle_defns_ctr_lookup c ds
			   | Some t => Some t
			 end
  end.

(* Gets the kind/type associated with constructor c in signature S
   Returns: [None] if c isn't in S
            [Some t] if c's kind/type is t in S
*)
Fixpoint sig_lookup_ctr (c:ctr) (S:sig) {struct S}: option tm :=
  match S with
    | nil   => None
    | b::bs => match b with 
		 | bundle_defns defns => 
		   match bundle_defns_ctr_lookup c defns with
		     | Some t => Some t
		     | None => sig_lookup_ctr c bs
		   end
		 | bundle_assert _ _ => sig_lookup_ctr c bs
	       end
  end.   

(* Gets the kind/type associated with an assertion c in signature S 
   returns: [None] if c isn't in S
            [Some t] if c's kind/type is t in S
*)
Fixpoint sig_lookup_assn (c:ctr) (S:sig) {struct S}: option tm :=
  match S with
    | nil   => None 
    | b::bs => match b with
                 | bundle_defns defns => sig_lookup_assn c bs
		 | bundle_assert c' t =>
		   if eq_ctr_dec c c' then Some t else sig_lookup_assn c bs
	       end
  end.


Definition ctr_decls_has (c:ctr) (t:tm) (cdl:ctr_decls) :=
ctr_decls_lookup c cdl = Some t.

Definition defns_has_ctr (c:ctr) (tp: tm) (d:defns) :=
 bundle_defns_ctr_lookup c d = Some tp.

Definition sig_has_ctr (c:ctr) (t:tm) (S:sig) : Prop :=
  sig_lookup_ctr c S = Some t.

Definition sig_has_assn (c:ctr) (t:tm) (S:sig) : Prop :=
  sig_lookup_assn c S = Some t.

Fixpoint get_defns_ctrs_of (c:ctr) (d:defns) {struct d} : option (list ctr) :=
  match d with
    | nil => None
    | (c', t, cdl)::ds =>
      if eq_ctr_dec c c'
	then Some (dom_ctr_decls cdl)
	else get_defns_ctrs_of c ds
  end.

(* Gets the list of expression constructors for type c from the signature S.
   Returns: [None] if c isn't in S
            [Some l], where l is the list of term constructors, otherwise
*)
Fixpoint get_ctrs_of (c:ctr) (S:sig) {struct S}: option (list ctr) :=
  match S with
    | nil   => None
    | b::bs => match b with
		 | bundle_defns d => 
		   match get_defns_ctrs_of c d with
		     | Some l => Some l
		     | None => get_ctrs_of c bs
		   end
		 | bundle_assert _ _ => get_ctrs_of c bs
	       end
  end.


Definition get_tctr_defns_of (d:defns) : defns :=
  let strip x := 
    match x with (n,t,_) => (n,t,nil) end
  in
  List.map strip d.

Fixpoint dom_defns (d:defns) {struct d}: list ctr :=
  match d with
  | nil => nil
  | (c,_,cdl)::ds => c::(dom_ctr_decls cdl)++(dom_defns ds)
  end.

(*
Fixpoint dom_defns (d:defns) {struct d}: list ctr :=
  let defns_ctrs_dom (cdl:ctr_decls): list ctr :=
    List.map (fun (x:ctr*tm) => let (n,_) := x in n) cdl in
  match d with
  | nil => nil
  | (c,_,cdl)::ds => c::(defns_ctrs_dom cdl)++(dom_defns ds)
  end.
*)


Fixpoint dom_bundle (b:bundle) : list ctr :=
match b with
  | bundle_defns d => dom_defns d
  | bundle_assert c t => c::nil
end.

Definition dom_sig (S: sig) : list ctr :=
  List.flat_map (dom_bundle) S.

Inductive is_tm_arr : tm -> Prop :=
| is_tm_arr_arr : forall t1 t2, is_tm_arr (tm_arr t1 t2)
.


Hint Constructors is_tm_arr.

(* [constructs tm s] is intended to hold when [tm] has form
     [t1 -> ... -> tn -> s] and s is not an arrow *)
Inductive constructs : tm -> tm -> Prop :=
| constructs_base : forall s, ~(is_tm_arr s) -> constructs s s
| constructs_arr  : forall s t1 t2, 
    constructs t2 s -> constructs (tm_arr t1 t2) s.

Hint Constructors constructs.


(*
  fully_applied t c (tn::t(n-1)::...t1::nil) k holds when
    t = (...((c t1) t2) ...tn) where k is the kind of c
    and the arity of k is n

  Note: args is in reverse application order.
*)
(*Inductive fully_applied : tm -> ctr -> list tm -> tm -> Prop :=
| fully_applied_TYPE : forall c, fully_applied (tm_ctr c) c nil K_TYPE
| fully_applied_PROP : forall c, fully_applied (tm_ctr c) c nil K_PROP
| fully_applied_app  : forall k1 k2 t tn c args,
    fully_applied t c args k2 ->
    fully_applied (tm_app t tn) c (tn::args) (tm_arr k1 k2)
.*)
Inductive fully_applied_args : tm -> ctr -> list tm -> Prop :=
| fully_applied__args_ctr : forall c, fully_applied_args (tm_ctr c) c nil 
| fully_applied_args_app  : forall  t tn c args,
    fully_applied_args t c args ->
    fully_applied_args (tm_app t tn) c (tn::args) 
.
Inductive fully_applied_typ : tm -> ctr -> tm -> Prop :=
| fully_applied_TYPE : forall c, fully_applied_typ (tm_ctr c) c  K_TYPE
| fully_applied_PROP : forall c, fully_applied_typ (tm_ctr c) c  K_PROP
| fully_applied_typ_app  : forall k1 k2 t tn c,
    fully_applied_typ t c k2 ->
    fully_applied_typ (tm_app t tn) c  (tm_arr k1 k2)
.

Hint Constructors fully_applied_args fully_applied_typ.

(*
  all_tm_bv holds when the list of terms consists entirely of bound variables.
*)

(* Limin: It appears that the ordering of bound variables are relevant*)
Inductive all_tm_bv : nat -> (list tm) ->  Prop :=
| all_tm_bv_nil  : forall n, all_tm_bv n nil
| all_tm_bv_cons : forall l n, all_tm_bv (n+1) l -> all_tm_bv n ((tm_bv n)::l).

Hint Constructors all_tm_bv.

Inductive all_tm_lc : (list tm) -> Prop :=
| all_tm_lc_nil  : all_tm_lc nil
| all_tm_lc_cons : forall a l, all_tm_lc l -> lc a -> all_tm_lc (a::l).

Hint Constructors all_tm_lc.

(* has_level t n if t = t1 -> ... -> tn -> s*)
Inductive has_level: tm -> nat -> Prop:=
| has_level_base: forall u, ~is_tm_arr u -> has_level u 0
| has_level_arr: forall u n t,
  has_level u n -> has_level (tm_arr t u) (1+n).

Hint Constructors has_level.

(* ctrs_of t = l where l is a list of (posibly non-unique)
   nats where each n in l occurence in t as (tm_ctr n) *)
Fixpoint ctrs_of (t:tm) : list ctr :=
  match t with
  | tm_fv _         => nil
  | tm_bv _         => nil
  | tm_con _        => nil
  | tm_ctr c        => c::nil
  | tm_app t1 t2    => (ctrs_of t1) ++ (ctrs_of t2) 
  | tm_abs t1 t2    => (ctrs_of t1) ++ (ctrs_of t2)
  | tm_arr t1 t2    => (ctrs_of t1) ++ (ctrs_of t2)
  | tm_mat t1 t2 t3    => (ctrs_of t1) ++ (ctrs_of t2) ++ (ctrs_of t3) 
  | tm_brn c t2 t3  => (ctrs_of t2) ++ (ctrs_of t3)
  | tm_cast t1 t2   => (ctrs_of t1) ++ (ctrs_of t2)
  end.

(* [positive (n_1, ... n_m) t] is intended to hold
   when construtor application of [n_i] appear at most
   positively in [t].  As definied this judgment is very
   conservative *)

(* Name? *)
Inductive positive (ctrs: list ctr): tm -> Prop :=
| positive_arr: forall t1 t2,
      disjoint ctrs (ctrs_of t1)
   -> positive ctrs t2
   -> positive ctrs (tm_arr t1 t2)
| positive_app: forall t1 t2,
      positive ctrs t1
   -> disjoint ctrs (ctrs_of t2)
   -> positive ctrs (tm_app t1 t2)
| postive_ctr: forall n,
      positive ctrs (tm_ctr n).

Hint Constructors positive.

Inductive branches_cover_aux : (list ctr) -> tm -> Prop :=
| bca_nil : branches_cover_aux nil TM_NONE
| bca_cons : forall c l t branches,
            branches_cover_aux l branches ->
            branches_cover_aux (c::l) (tm_brn c t branches).

Hint Constructors branches_cover_aux.

(* [branches_cover S branches ctr] is intended to hold when [branches] contains exactly
   one [tm_brn] per term constructor which creates a [ctr]. *)
Definition branches_cover (S:sig) (branches:tm) (ctr: ctr): Prop := 
  exists l, get_ctrs_of ctr S = Some l /\ branches_cover_aux l branches.


(**************************************************)
(* Conversion relation                            *)

(* Note: E may contain conversion assumptions of 
 * the form:    e1 = e2 : t
 *)
Inductive converts: env -> tm -> tm -> Prop :=
(* equivalence axioms *)
| conv_refl  : forall E t, lc t -> converts E t t

| conv_sym   : forall E t1 t2, 
                converts E t1 t2
             -> converts E t2 t1

| conv_trans : forall t3 E t1 t2,
                converts E t1 t2
             -> converts E t2 t3
             -> converts E t1 t3

(* definitional equality *)
| conv_axiom : forall E t1 t2 k x,
                binds_eq x t1 t2 k E
             -> converts E t1 t2

(* congruence axioms *)
| conv_app   : forall E t1 t2 u1 u2,
                converts E t1 u1
             -> converts E t2 u2
             -> converts E (tm_app t1 t2) (tm_app u1 u2)

| conv_abs   : forall L E t1 t2 u1 u2,
                converts E t1 u1
             -> (forall x, x\notin L ->  converts E (t2^^ tm_fv x) (u2^^tm_fv x) )
             -> converts E (tm_abs t1 t2) (tm_abs u1 u2)

| conv_arr   : forall L E t1 t2 u1 u2,
                converts E t1 u1
             -> (forall x, x\notin L -> converts E (t2^^tm_fv x) (u2^^tm_fv x))
             -> converts E (tm_arr t1 t2) (tm_arr u1 u2)

(* | conv_mat   : forall E t1 t2 t3 u1 u2 u3,
                converts E t1 u1
             -> converts E t2 u2   
             -> converts E t3 u3    	
             -> converts E (tm_mat t1 t2 t3) (tm_mat u1 u2 u3)

| conv_brn   : forall E c t1 t2 u1 u2,
                converts E t1 u1
             -> converts E t2 u2    
             -> converts E (tm_brn c t1 t2) (tm_brn c u1 u2)
                
| conv_cast  : forall E t1 t2 u1 u2,
                converts E t1 u1
             -> converts E t2 u2    
             -> converts E (tm_cast t1 t2) (tm_cast u1 u2)*)
.

(* first-order types *)

(* simple types  *)
Fixpoint check_simple_type (L:list ctr) (t:tm) {struct t}: bool :=
 match t with
  | tm_ctr c => if (In_dec eq_ctr_dec c L) then true else false
  | tm_con c => if (c===30) then true else false
  | _ => false
 end.

Definition simple_type L t := check_simple_type L t = true.

 Fixpoint atomic_ctr_type (L: list ctr) (t:tm) {struct t}: bool :=
 match t with
 | tm_ctr c => true
 | _ => false
 end.


 Fixpoint all_atomic_ctr_types  (L: list ctr) (cdl:ctr_decls)  {struct cdl}: bool :=
  match cdl with
  | nil => true
  | (c ,t)::cdl' => Bool.andb (atomic_ctr_type L t) (all_atomic_ctr_types L cdl')
  end. 

 Fixpoint all_atomic_decls_defns (L: list ctr) (d:defns)  {struct d}: bool :=
  match d with
  | nil => true
  | (tctr, k, cdls):: d =>
    if eq_tm_dec k K_TYPE then Bool.andb (all_atomic_ctr_types  L cdls) (all_atomic_decls_defns L d)
    else false
  end. 

 Fixpoint get_atomic_ctrs_sig (S:sig)  (L: list ctr) {struct S}: list ctr :=
  match S with
  | nil => L
  | d:: S' => let L' := get_atomic_ctrs_sig S' L in
             match d with 
	     | bundle_assert c k => L'
	     | bundle_defns decls => 
               if all_atomic_decls_defns  ((dom_defns(get_tctr_defns_of decls))++ L') decls 
               then L'++ (dom_defns(get_tctr_defns_of decls))
               else L'
	      end
 end
.

Definition atomic_type (S:sig) (t:tm) := 
   exists L, get_atomic_ctrs_sig S nil = L /\ simple_type L t.

(**************************************************)
(* Typing rules                                   *)

Inductive wf_env (S1:sig) (S2:sig): env -> Prop := 

(*
  S |- .
*)
| wf_env_nil      : wf_env S1 S2 empty

(*
    S |- E
    S;E |- t : k 
    x # E
    --------------------
    S |- E & x ~: t
*)
| wf_env_cons_var : forall k x t E,
                    wf_env S1 S2 E 
                ->  wf_tm S1 S2 E t k
                ->  x # E
                ->  wf_env S1 S2 (E & x ~: t)

(* TODO: ensure that e1 and e2 are values (i.e. in normal form)

   S |- E
   S;E |- e1 : Prin
   S;E |- e2 : Prin
   x # E
   ------------------------------
   S |- E & x ~ e1 = e2 :  Prin

*)
| wf_env_cons_eq  : forall x e1 e2 t E,
                    wf_env S1 S2 E 
                ->  wf_tm S1 S2 E e1 t 
                ->  wf_tm S1 S2 E e2 t 
                ->  atomic_type S2 t   (* TODO: relax restrictions on the form of t ?*)
                ->  x # E
		-> value e1 
	       -> value e2
               -> wf_tm S1 S2 E t K_TYPE
                ->  wf_env S1 S2 (E & x ~ (bnd_eq e1 e2 t))

(* S1 is the well-formed signature.
   S2 is either the same as S1 or contains more data constructors than S1.
   S2 is used for looking up constructors definitions *)
with wf_tm (S1:sig) (S2:sig): env -> tm -> tm -> Prop :=

(* 
 *    S;E |- ok
 * ---------------------
 *  S;E |- Type : Kind  
 *
 *)

| wf_tm_K_TYPE : forall E,
     wf_env S1 S2 E ->
     wf_tm S1 S2 E K_TYPE S_KIND

(*
 *  S;E |- ok
 * -------------------
 *  S;E |- Prop : Kind    
 *
 *)

| wf_tm_K_PROP : forall E,
     wf_env S1 S2 E ->
     wf_tm S1 S2 E K_PROP S_KIND



(*
 *  S; E, x:k1 |- u2: k2
 *  k2 = Type or Prop or Kind
 * ------------------------------
 *  S; E |- (x:k1) -> u2 : k2
 *)

| wf_tm_arr : forall L E u1 u2 k1 k2,
  wf_tm S1 S2 E u1 k1 ->
  (forall x, x \notin L -> wf_tm S1 S2 (E & x ~: u1) (u2 ^^ (tm_fv x)) k2) ->
   k2 = K_TYPE \/ k2 = K_PROP \/ k2 = S_KIND ->
  wf_tm S1 S2 E (tm_arr u1 u2) k2



(*
 *  S;E |- ok
 *  (c : t) \in S 
 * --------------------
 *  S;E |- c : t
 *
 *)
| wf_tm_ctr : forall E c t,
     wf_env S1 S2 E ->
     sig_has_ctr c t S1  ->
     wf_tm S1 S2 E (tm_ctr c) t

(*
 *  S;E |- ok
 *  (c : t) \in S
 * --------------------
 *  S;E |- c : t
 *
 *)
| wf_tm_assn : forall E c t,
     wf_env S1 S2 E ->
     sig_has_assn c t S1 ->
     wf_tm S1 S2 E (tm_ctr c) t

(*
 *  S; E |- ok
 *  (x : t) \in E 
 * ----------------
 *  S; E |- x: t
 *)
| wf_tm_var : forall E x t,
      wf_env S1 S2 E ->
      binds_var x t E ->
      wf_tm S1 S2 E (tm_fv x) t

(*
 *  S; E |- u_1: k_1
 *  S; E,x:u1 |- f^x: u_2
 *  S; E |- (x:u_1) -> u2 : k
 *  k \in {Type, Prop}
 * -------------------------------
 *  S; E |- abs f : (x: u_1) -> u_2
 *)
| wf_tm_abs: forall L k k1 E u1 u2 f,
    wf_tm S1 S2 E u1 k1 ->
    (forall x, x \notin L -> wf_tm S1 S2 (E & x ~: u1) (f ^^ tm_fv x) (u2 ^^ tm_fv x) ) ->
    wf_tm S1 S2 E (tm_arr u1 u2) k ->
    k = K_TYPE \/ k = K_PROP ->
    k1 = K_TYPE \/ k1 = K_PROP \/ k1 = S_KIND ->
    wf_tm S1 S2 E (tm_abs u1 f) (tm_arr u1 u2)


(*
 * S; E |- e1 : (x:t2) -> u
 * S; E |- e2 : t2
 * S; E |- u{e2/x} : ku
 * S; E |- t2 : k2
 *  value e2 or lc u
 * -------------------------------------
 * S; E |- e1 e2 : u{e2/x}
*)

| wf_tm_app: forall u (*ku  k2*) E e1 e2 t2,
    wf_tm S1 S2 E e1 (tm_arr t2 u) 
 -> wf_tm S1 S2 E e2 t2
(*
  -> wf_tm S E t2 k2 
  -> wf_tm S E (u ^^ e2) ku
*)
 -> (value e2 \/ lc u )
 -> wf_tm S1 S2 E (tm_app e1 e2) (u ^^ e2)


(** Language-specific constants *)

(*
 * S;E |- v : T_PRIN
 * S;E |- P : K_PROP    (value v)
 * --------------------------------
 * S;E |- v says p : K_PROP
 *)
| wf_tm_P_SAYS: forall E v P,
    wf_tm S1 S2 E v T_PRIN
 -> wf_tm S1 S2 E P K_PROP
 -> value v
 -> wf_tm S1 S2 E (tm_app (tm_app P_SAYS v) P) K_PROP

(* signatures: 
 *
 * S; . |- v : Prin
 * value v
 * S; . |- p : Prop
 * ------------------------
 * S;E |- sign(v, p) : a says p
 *)
| wf_tm_PF_SIGN : forall E v p,
    wf_tm S1 S2 empty v T_PRIN
 -> value v
 -> wf_tm S1 S2 empty p K_PROP
 -> wf_env S1 S2 E
 -> wf_tm S1 S2 E (tm_app (tm_app PF_SIGN v) p) (tm_app (tm_app P_SAYS v) p)

(*
 * S |- E wf
 * -----------
 * S;E |- Prin : Type
 *)
| wf_tm_T_PRIN : forall E, 
    wf_env S1 S2 E 
 -> wf_tm S1 S2 E T_PRIN K_TYPE

(*
 * S;E |- P : K_PROP
 * -------------------
 * S;E |- pf P : Type
 *)
| wf_tm_T_PF : forall E P,
    wf_tm S1 S2 E P K_PROP
 -> wf_tm S1 S2 E (tm_app T_PF P) K_TYPE

(*
 * S;E |- p : P
 * S;E |- P : K_PROP
 * ------------------
 * S;E |- return_pf p : (Pf P)
 *)
| wf_tm_TM_PF_RETURN : forall E p P,
    wf_tm S1 S2 E p P
 -> wf_tm S1 S2 E P K_PROP
 -> wf_tm S1 S2 E (tm_app TM_PF_RETURN p) (tm_app T_PF P)

(*
 * S;E |- e1 : (Pf P)
 * S;E |- e2 : (x:P) -> (Pf Q)    x \notin (fv Q)
 * -----------------------------------------------
 * S;E |- bind e1 e2 : (Pf Q)
 *)
(* NOTE: x can't do a 'dependent-let' style typing rule
   because there's nothing appropriate to substitute for x in Q *)
| wf_tm_TM_PF_BIND : forall P E e1 e2 Q,
    wf_tm S1 S2 E e1 (tm_app T_PF P)
 -> wf_tm S1 S2 E e2 (tm_arr P (tm_app T_PF Q))
 -> lc Q 
 -> wf_tm S1 S2 E (tm_app (tm_app TM_PF_BIND e1) e2) (tm_app T_PF Q)


(*
 * S;E |- P : K_PROP
 * --------------------
 * S;E |- say P : (Pf (self says P))
 *)
| wf_tm_TM_SAY : forall E P,
    wf_tm S1 S2 E P K_PROP
 -> wf_tm S1 S2 E (tm_app TM_SAY P) (tm_app T_PF (tm_app (tm_app P_SAYS TM_SELF) P))

(*
 * S;E |- v1 : T1    (value v1)
 * S;E |- v2 : T1    (value v2)
 * S |- atomic_type [] T1
 * S;E,v1=v2 |- e1 : T
 * S;E |- e2 : T
 *----------------------------------
 * S;E |- PIF v1 v2 e1 e2 : T
 *) 
| wf_tm_TM_PIF : forall L E v1 v2 e1 e2 T T1,
    wf_tm S1 S2 E v1 T1
 -> wf_tm S1 S2 E v2 T1
 -> atomic_type S2 T1
 -> value v1
 -> value v2
 -> (forall x, x \notin L -> wf_tm S1 S2 (E & x ~ (bnd_eq v1 v2 T1)) e1 T)
 -> wf_tm S1 S2 E e2 T
 -> wf_tm S1 S2 E T1 K_TYPE
 -> wf_tm S1 S2 E (tm_app (tm_app (tm_app (tm_app TM_PIF v1) v2) e1) e2) T


(*
 * S1 S2 |- E wf
 * ----------------------
 * S;E |- self : Prin
 *)
| wf_tm_TM_SELF : forall E,
    wf_env S1 S2 E
 -> wf_tm S1 S2 E TM_SELF T_PRIN

(*
 * S;E |- p : P
 * S;E |- P : K_PROP
 * S;E |- a : T_PRIN  (value a)
 * ------------------------------
 * S;E |- return_says a p : (a says P)
 *)
| wf_tm_PF_SAYS_RETURN : forall E p P a,
    wf_tm S1 S2 E p P
 -> wf_tm S1 S2 E P K_PROP
 -> wf_tm S1 S2 E a T_PRIN
 -> value a
 -> wf_tm S1 S2 E (tm_app (tm_app PF_SAYS_RETURN a) p) (tm_app (tm_app P_SAYS a) P)


(*
 * S;E |- p1 : a says P
 * S;E |- p2 : (x:P) -> a says Q   (x \notin (fv Q))
 * -------------------------------------------------
 * S;E |- bind p1 p2 : a says Q
 *)
| wf_tm_PF_SAYS_BIND : forall P E a p1 p2 Q,
    wf_tm S1 S2 E p1 (tm_app (tm_app P_SAYS a) P)
 -> wf_tm S1 S2 E p2 (tm_arr P (tm_app (tm_app P_SAYS a) Q))
 -> lc Q 
 -> wf_tm S1 S2 E (tm_app (tm_app PF_SAYS_BIND p1) p2) (tm_app (tm_app P_SAYS a) Q)

(*
 * Pattern Matching 
 *)
|  wf_tm_mat_TYPE: forall E e s branches result_typ ctr args k, 
                wf_tm S1 S2 E e s ->
                wf_tm S1 S2 E s K_TYPE ->                 (* s is a TYPE ... *)
                fully_applied_args s ctr args ->         (*   ... and is a fully applied to (List.reverse [args]) *)
                sig_has_ctr ctr k S1 ->                (*   ... which the sig says is a type con *)       
		fully_applied_typ s ctr k ->
                wf_tm S1 S2 E result_typ K_TYPE ->            (* Homogeneous elimination only *)
                wf_branches S1 S2 E s (List.rev args) branches result_typ ->     (* check branches *)
                branches_cover S2 branches ctr ->      (* Coverage check *)
                wf_tm S1 S2 E (tm_mat e result_typ branches) result_typ

|  wf_tm_mat_PROP: forall E e s branches result_typ ctr args k, 
                wf_tm S1 S2 E e s ->
                wf_tm S1 S2 E s K_PROP ->                 (* s is a PROP ... *)
                fully_applied_args s ctr args ->         (*   ... and is a fully applied to (List.reverse [args]) *)
                sig_has_ctr ctr k S1 ->                (*   ... which the sig says is a type con *)       
		fully_applied_typ s ctr k ->
                wf_tm S1 S2 E result_typ K_PROP ->            (* Homogeneous elimination only *)
                wf_branches S1 S2 E s (List.rev args) branches result_typ ->     (* check branches *)
                branches_cover S2 branches ctr ->      (* Coverage check *)
                wf_tm S1 S2 E (tm_mat e result_typ branches) result_typ


(* Conversion rule:
 *
 * S;E |- e : t1
 * S;E |- t1 : K_TYPE
 * S;E |- t2 : K_TYPE
 * S;E |- t1 == t2
 * ------------------
 * S;E |- cast(e, t2) : t2
 *)
| wf_tm_cast : forall E e t1 t2,
    wf_tm S1 S2 E e t1
 -> wf_tm S1 S2 E t2 K_TYPE (* added to make proof easier *)
 -> converts E t1 t2
 -> wf_tm S1 S2 E (tm_cast e t2) t2

(* need secondary judgment to type check branches that records the
     type we are matching *)
(*   wf_branches S E branches tctr args result_typ *)
with wf_branches (S1:sig) (S2:sig) : env -> tm -> list tm -> tm -> tm -> Prop :=
| wf_branches_None : forall E s args result_type,
        wf_branches S1 S2 E s args (TM_NONE) result_type
| wf_branches_Some : forall E s args ectr ectr_typ body body_type rest result_type,
        wf_branches S1 S2 E s args rest result_type ->
        sig_has_ctr ectr ectr_typ S1 ->
        wf_tm S1 S2 E body body_type ->
        wf_brn S1 S2 s args ectr_typ body_type result_type ->
        wf_branches S1 S2 E s args (tm_brn ectr body rest) result_type

(* wf_brn   *)
with wf_brn (S1:sig) (S2:sig)  : tm -> (list tm) -> tm -> tm -> tm -> Prop :=
  (* Base case: 
       - The constructor takes no arguments, the datatype has no remaining parameters
         so the type of the body must just be the result type of the match expression
       - Example: when checking the "Z" case of the datatype Nat:
 
         data Nat : Type {
         | Z : Nat
         | S : Nat -> Nat
         }

         when pattern matching against a Nat, we have:
           match e : Nat {
           | Z -> body_Z
           | S -> \ (n:Nat).body_S
           }
   *)  
| wf_brn_base :             
     forall s result_type,
     wf_brn S1 S2 s nil s result_type result_type
  (* Inductive case:
       - The constructor takes arguments in addition to the datatype parameters
       - The parameters have already been stripped off, so the type of the body
         must take an argument
       - Continuing the example, when checking the successor branch for type Nat:
         the constructor S has type Nat -> Nat so the body should have type 
         Nat -> result_typ
       - The body should be of the form \ ([x]:Nat).body
       - Note: This rule does not allow the result typ to depend on the 
         constructor argument
  *)
| wf_brn_constructor_arg :
     forall L s T1 T U result_typ,
     (forall x, 
        x \notin L ->
        x \notin (fv result_typ) ->
         wf_brn S1 S2 s nil (T^^(tm_fv x)) (U^^(tm_fv x)) result_typ
     ) ->
     wf_brn S1 S2 s nil (tm_arr T1 T) (tm_arr T1 U) result_typ

  (* Inductive case:
      - The datatype has a parameter (e.g. List : Type -> Type)
      - The type of the scrutinee is thus [tctr] applied to [args]
      - We witness the generated equality by opening up the types 
        using the arguments in [args]

      - For example: when checking the 'cons' case of a match against
        an expression of type (List Nat) we have [tctr] = List and 
        [args] = [Nat]  The branch should look like:

        match e {
        | nil -> body_nil
        | cons -> \ (hd:0).\ (tl:List 1).body_cons
        }

  *)
| wf_brn_datatype_param:
     forall s arg args T1 T U result_type,
       wf_brn S1 S2 s args (T ^^ arg) U result_type ->
       wf_brn S1 S2 s (arg::args) (tm_arr T1 T) U result_type
.

Hint Constructors wf_env  wf_tm wf_branches wf_brn.


Scheme ind_wf_tm_branches := Induction for wf_tm Sort Prop
with   ind_wf_branches_tm := Induction for wf_branches Sort Prop.



Definition mutual_wf_tm_branches_ind (S1:sig) (S2:sig)
    (P: forall e t s, wf_tm S1 S2 e t s -> Prop)
    (P': forall e s args b t_res, wf_branches S1 S2 e s args b t_res -> Prop) :=

fun   h1 h2 h3 h4 h5 h6 h7 h8 h9 h10 h11 h12 h13 h14 h15 h16 h17 h18 h19 h20 h21 h22 h23 h24=>
    conj
    (@ind_wf_tm_branches  S1 S2 P P'   h1 h2 h3 h4 h5 h6 h7 h8 h9 h10 h11 h12 h13 h14 h15 
                                                         h16 h17 h18 h19 h20 h21 h22 h23 h24) 
    (@ind_wf_branches_tm S1 S2 P P'   h1 h2 h3 h4 h5 h6 h7 h8 h9 h10 h11 h12 h13 h14 h15 
     h16 h17 h18 h19 h20 h21 h22 h23 h24).

Check mutual_wf_tm_branches_ind.


Scheme ind_wf_tm_branches_brn_env := Induction for wf_tm Sort Prop
with   ind_wf_branches_brn_env_tm := Induction for wf_branches Sort Prop
with   ind_wf_brn_env_tm_branches := Induction for wf_brn Sort Prop
with   ind_wf_env_tm_branches_brn := Induction for wf_env Sort Prop.

Definition mutual_wf_tm_branches_brn_env_ind (S1:sig) (S2:sig)
   (P: forall e t s, wf_tm S1 S2 e t s -> Prop)
   (P': forall e s args b t_res, wf_branches S1 S2 e s args b t_res -> Prop)
   (P'': forall s args ectr_typ t_b t_res, wf_brn S1 S2 s args ectr_typ t_b t_res  -> Prop) 
   (P''': forall E, wf_env S1 S2 E -> Prop)  :=
  fun h1 h2 h3 h4 h5 h6 h7 h8 h9 h10 h11 h12 h13 h14 h15 h16 h17 h18 
     h19 h20 h21 h22 h23 h24 h25 h26 h27 h28 h29 h30=>
    conj
    (@ind_wf_tm_branches_brn_env S1 S2 P P' P'' P''' h1 h2 h3 h4 h5 h6 h7
                            h8 h9 h10 h11 h12 h13 h14 h15 h16 h17 h18  h19 h20 h21 
                            h22 h23 h24 h25 h26 h27 h28 h29 h30) 
    ( conj
    (@ind_wf_branches_brn_env_tm S1 S2 P P' P'' P''' h1 h2 h3 h4 h5 h6 h7 
                            h8 h9 h10 h11 h12 h13 h14 h15 h16 h17 h18 h19 h20 h21 
                            h22 h23 h24 h25 h26 h27 h28 h29 h30)
    (conj
    (@ind_wf_brn_env_tm_branches S1 S2 P P' P'' P''' h1 h2 h3 h4 h5 h6 h7 
                            h8 h9 h10 h11 h12 h13 h14 h15 h16 h17 h18 h19 h20 h21
                            h22 h23 h24 h25 h26 h27 h28 h29 h30)
    ( @ind_wf_env_tm_branches_brn S1 S2 P P' P'' P''' h1 h2 h3 h4 h5 h6 h7 
                            h8 h9 h10 h11 h12 h13 h14 h15 h16 h17 h18 h19 h20 h21 
                            h22 h23 h24 h25 h26 h27 h28 h29 h30))).

Check mutual_wf_tm_branches_brn_env_ind.

(* wf_sig: Ensures that a list of (mutually recursive sets of) data type declarations is 
 * well formed.

  |- S wf

 *)
Inductive wf_sig : sig -> Prop :=

(*
  --------------
  |- nil wf_sig
*)
| wf_sig_nil : wf_sig nil


(* assertion *)
| wf_assertion : forall c k S,
                   wf_sig S
                -> wf_tm S S empty k S_KIND
                -> constructs k K_PROP
                -> ~(In c (dom_sig S))
                -> wf_sig ((bundle_assert c k)::S)

(*  Note: [b] means the bundle derived from b by considering only the type constructors in b
    (and ignoring the constructors associated with them).

 |- S wf_sig
 S |- b :: Type wf_bundle_tctrs
 nil ; S,[b] |- b wf_bundle_ctrs
 -------------------------------
  |- S,b wf_sig
*)

| wf_sig_cons_Type : forall (d:defns) (S:sig),
                   wf_sig S
                -> wf_bundle_tctrs S d K_TYPE   	
		-> disjoint (dom_defns d) (dom_sig S)
                -> wf_bundle_ctrs nil (bundle_defns(get_tctr_defns_of d) ::S)((bundle_defns d) :: S) d  
                -> wf_sig ((bundle_defns d)::S)


(*  Note: [b] means the bundle derived from b by considering only the type constructors in b
    (and ignoring the constructors associated with them).

 |- S wf_sig
 S |- b :: Prop wf_bundle_tctrs
 [b] ; S,[b] |- b wf_bundle_ctrs
 -------------------------------
  |- S,b wf_sig
*)
| wf_sig_cons_Prop : forall (d:defns) (S:sig),
                   wf_sig S
                -> wf_bundle_tctrs S d K_PROP   
		-> disjoint (dom_defns d) (dom_sig S)	
                -> wf_bundle_ctrs (dom_defns d) (bundle_defns(get_tctr_defns_of d) ::S) ((bundle_defns d)::S) d  
                -> wf_sig ((bundle_defns d)::S)


(* Ensures that each type constructor in the mutually recursive datatype 
   constructs an object (Type or Prop) of the same kind [k]
   This means that props and types can't be mutually recursively defined.

*)
with wf_bundle_tctrs : sig -> defns -> tm -> Prop :=

(*
  S;. |- k :: SORT
  -----------------------
  S |- nil :: k wf_bundle_tctrs
*)
| wf_bundle_tctrs_nil  : forall S k, 
                      wf_tm S S empty k S_KIND    (* Requires that k = Type or Prop *)
                   -> wf_bundle_tctrs S nil k

(*
   S |- b :: k wf_bundle_tctrs
   c \notin dom (S,b)
   S;. |- k : KIND
   k == (x1:t1) -> (x2:t2) -> ... -> K
   ------------------------------------ 
   S |- (c,k,cdl) :: K wf_bundle_tctrs

*)
| wf_bundle_tctrs_cons : forall S c k K cdl d,
                      wf_bundle_tctrs S d K   
                   -> ~(In c (dom_defns d)) (* c not already defined *)
		   -> ~(In c (dom_ctr_decls cdl))
                   (* -> ~(In c (dom_sig S))  *)
                   -> wf_tm S S empty k S_KIND
                   -> constructs k K
                   -> wf_bundle_tctrs S ((c, k, cdl)::d) K


(* First argument is a list of type constructors that must appear positively in expression
   constructor types *)
with wf_bundle_ctrs : (list ctr) -> sig -> sig -> defns -> Prop := 

(* Note: could add |- S wf_sig   (but that does a lot of re-checking of the signature)

   ----------------------------
   P ; S |- nil wf_bundle_ctrs
*)
| wf_bundle_ctrs_nil  : forall P S1 S2,  
                         wf_bundle_ctrs P S1 S2 nil

(*
   
  P ; S |- b wf_bundle_ctrs
  P ; S |- c : k {cdl} wf_ctr_decls
  dom (ctrs) \cap dom (b) = empty
  dom (ctrs) \cap dom S = empty
  -----------------------------------
  P ; S |- b and c:t {cdl} wf_bundle_ctrs

*)
| wf_bundle_ctrs_cons : forall P S1 S2 c k cdl d, 
                         wf_bundle_ctrs P S1 S2 d
                      -> wf_ctr_decls P S1 S2 c k cdl     (* cdl is a well-formed constructor declaration list for c *)
                      -> disjoint (dom_ctr_decls cdl) (dom_defns d)
                     (* -> disjoint (dom_ctr_decls cdl) (dom_sig S)*)
                      -> wf_bundle_ctrs P S1 S2 ((c, k, cdl)::d)



with wf_ctr_decls : (list ctr) -> sig -> sig -> ctr -> tm -> ctr_decls -> Prop := 
(*
  -----------------
  P ; S |- c : k {} wf_ctr_decls
*)
| wf_ctr_decls_nil  : forall P S1 S2 c k, 
                    wf_ctr_decls S1 S2 P c k nil

(*
  P ; S |- c : k {cdl} wf_ctr_decls
  ctr \notin dom(cdl)
  S ; . |- t : K
  t == (x1:t1) -> (x2:t2) -> ... -> (c x1 ... xm)  
  c_inst = (c x1 ... xm)  is saturated at kind k
  constructors in P appear strictly positively in t
  --------------------------------------------------
  P ; S |- 
*)
(*Limin: add one more condition to check that the number of arguments of a data type is correctly recorded *)
| wf_ctr_decls_cons : forall K args c_inst P S1 S2 c k ctr n m t cdl,
                      wf_ctr_decls P S1 S2 c k cdl
                   -> ~(In (ctr, n) (dom_ctr_decls cdl))  (* Constructor is used only once *)
                   -> wf_tm S1 S2 empty t K              (* Make sure the type definition is of the right kind *)
		   -> has_level t m
                   -> constructs t c_inst            (* t constructs an instance c_inst *)
                   -> fully_applied_args c_inst c args   (*   that is fully applied [c] of kind [k] *)
		   -> fully_applied_typ c_inst c k
		   -> n = List.length args           (* n is the number of args *)
                   -> ~(m < n) 
                   -> all_tm_bv (m-n) args                 (*   and all the arguments are bound variables *)
                   -> positive P t                   (* Make sure tycons in P only occur positively *)
                   -> wf_ctr_decls P S1 S2 c k (((ctr, n), t)::cdl)
.

Hint Constructors wf_sig  
                  wf_bundle_tctrs
                  wf_bundle_ctrs
                  wf_ctr_decls.



Lemma construct_unique: forall t c1 c2,
constructs t c1 ->
constructs t c2 -> 
c1 = c2.
Proof.
intros t c1 c2 Hc1 Hc2.
induction Hc1.
Case"~is_tm_arr".
inversion Hc2; subst; auto.
assert(J:= is_tm_arr_arr t1 t2); contradiction.

Case "arr".
apply IHHc1.
inversion Hc2; subst; auto.
assert(J:= is_tm_arr_arr t1 t2 ); contradiction.
Qed.

Lemma fully_applied_args_unique: forall t c args c' args',
fully_applied_args t c args ->
fully_applied_args t c' args'  ->
(c = c' /\ args = args').
Proof.
introv H H'.
gen c' args'.
induction H.
Case "Base".
introv H'. inversion H'; subst; auto.
introv H'. inversion H'; subst; auto.


assert (J:= IHfully_applied_args c' args0 H4).
destruct J.
split; auto. subst; auto.
Qed.


Lemma sig_has_ctr_strong_inv: forall c t S,
  wf_sig S ->
  sig_has_ctr c t S ->
  ( exists d, 
   (exists S', exists k, wf_bundle_tctrs S' d k) 
   /\ (exists P, exists S1', exists S2', wf_bundle_ctrs P S1' S2' d) /\ defns_has_ctr c t d).
Proof.
  introv Hwf Hhas.
  induction Hwf. 
  Case "nil".
    inversion Hhas.
     
  Case "assert".
    inversion Hhas.
    apply IHHwf.
    unfold sig_has_ctr. apply H3.

  Case "defns1".
    inversion Hhas.
    remember (bundle_defns_ctr_lookup c d) as Lookup.
    destruct Lookup.
    inversion H3.
    exists d. repeat split.
    exists S K_TYPE. auto.
    exists (nil:(list ctr)) (bundle_defns (get_tctr_defns_of d) :: S) (bundle_defns  d :: S). auto.
    unfold defns_has_ctr. symmetry. subst. auto.
    apply IHHwf.
    unfold sig_has_ctr. auto.

  Case "defns2".
    inversion Hhas.
    remember (bundle_defns_ctr_lookup c d) as Lookup.
    destruct Lookup.
    inversion H3.
    exists d. repeat split.
    exists S K_PROP. auto.
    exists (dom_defns  d)  (bundle_defns (get_tctr_defns_of d) :: S) (bundle_defns d :: S). auto.
    unfold defns_has_ctr. symmetry. subst. auto.
    apply IHHwf.
    unfold sig_has_ctr. auto.
Qed.



Lemma ctr_decls_has_type_inv: forall P S1 S2 c k cdl t c1,
 wf_ctr_decls P S1 S2 c k cdl ->
 ctr_decls_has c1 t cdl ->
 exists S1', exists S2', exists k, wf_tm S1' S2' empty t k.
Proof. 
intros. 
induction H.
inversion H0.
clear H1 H3 H4 H5 H6.
unfold ctr_decls_has in H0.
unfold ctr_decls_lookup in H0.
destruct c1 as (ctr1, n1).
destruct(eq_ctr_dec (ctr1, n1)(ctr0, n)).
inversion H0. rewrite <- H3; exists S1 S2 K; apply H2.

apply IHwf_ctr_decls. 
apply H0.
Qed.


Lemma bundle_has_type_inv: forall P S1 S2 S3 d k c t,
  wf_bundle_tctrs S1 d k ->
  wf_bundle_ctrs P S2 S3 d ->
  defns_has_ctr c t d ->
  exists S1', exists S2', exists k, wf_tm S1' S2' empty t k.
Proof. 
 intros.
 induction d.
 Case "nil".
   inversion H1.

 Case "cons".
   unfold defns_has_ctr in H1.
   unfold bundle_defns_ctr_lookup in H1.
   destruct a. destruct p.
   destruct (eq_ctr_dec c c1).

 Case "c=c1". 
   inversion H. inversion H1. subst.
     exists S1 S1 S_KIND. auto.

 Case "c<>c1".
   remember (ctr_decls_lookup c c0) as CTRDecl.
   destruct CTRDecl.

 Case "CTRDecl = Some t".
   inversion H1. subst.
   inversion H0. subst.


eapply ctr_decls_has_type_inv; eauto.  
assert (J: ctr_decls_lookup c c0 = Some t). symmetry; auto.
unfold ctr_decls_has; eapply J.

Case "CTRDecls = None".
eapply IHd; eauto. 
inversion H. subst. auto.
inversion H0. subst. auto.

Qed.


Lemma sig_has_ctr_typ_inv: forall c t S,
  wf_sig S ->
  sig_has_ctr c t S ->
  exists S1', exists S2', exists k, wf_tm S1' S2' empty t k.
Proof. 
 intros c t S Hwf Hhas.
  assert(J := sig_has_ctr_strong_inv Hwf Hhas).
  destruct  J as [b [Hwf_tctrs [Hwf_ctrs H_has]]].
  destruct Hwf_tctrs as [S' [K Htctrs]]. 
destruct Hwf_ctrs as [P [S1 [S2 Hctrs]]].
  destruct (bundle_has_type_inv Htctrs Hctrs H_has). firstorder.
Qed.

Lemma sig_has_assn_typ_inv: forall c t S,
  wf_sig S ->
  sig_has_assn c t S ->
  exists S1', exists S2', exists k, wf_tm S1' S2' empty t k.
Proof. 
intros.
induction H.

unfold sig_has_assn in H0.
unfold sig_lookup_assn in H0.
inversion H0. 

unfold sig_has_assn in H0.
unfold sig_lookup_assn in H0.
destruct (eq_ctr_dec c c0).
inversion H0. 
rewrite <- H5.
exists S S S_KIND; auto.

eapply IHwf_sig; eauto.

eapply IHwf_sig ; eauto.
eapply IHwf_sig; eauto.
Qed.



Lemma ctr_decls_has_strong_type_inv: forall P S1 S2 c k cdl t c1 n1,
 wf_ctr_decls P S1 S2 c k cdl ->
 ctr_decls_has (c1, n1) t cdl ->
 (exists c_inst, exists args, exists nt,  
   has_level t nt /\ constructs t c_inst /\ fully_applied_args c_inst c args 
  /\ fully_applied_typ c_inst c k
   /\ length args = n1
   /\ ~(nt<n1) 
   /\ all_tm_bv (nt -n1) args).
Proof. 
intros. 
induction H.
inversion H0.
unfold ctr_decls_has in H0.
unfold ctr_decls_lookup in H0.
destruct(eq_ctr_dec (c1, n1)(ctr0, n)).
inversion H0; inversion e; subst.
exists c_inst args m.  repeat (split); trivial. 

apply IHwf_ctr_decls. 
apply H0.
Qed.

 Lemma get_defns_ctr_some_in: forall ctr0 d l, 
  get_defns_ctrs_of ctr0 d = Some l -> In ctr0 (dom_defns d).
Proof.
 intros.
 induction d. 
Case "nil". inversion H.
Case "cons". 
  unfold get_defns_ctrs_of in H.
  destruct a. destruct p. 
  destruct (eq_ctr_dec ctr0 c0). 
  Case "ctr0 = c0". 
   subst. unfold dom_defns; simpl.  left; auto.
  Case "ctr0 <> c0".
     fold get_defns_ctrs_of in H. 
     assert (J:= IHd H).
    unfold dom_defns. fold dom_defns.
    simpl. right. 
   assert (J1: In ctr0 (dom_ctr_decls c) \/ In ctr0 (dom_defns d)).  right; auto.
   apply (in_or_app (dom_ctr_decls c) (dom_defns d) ctr0 J1).
Qed. 

 Lemma get_defns_ctr_some_in2: forall c ctr0 d l, 
  get_defns_ctrs_of ctr0 d = Some l -> In c l -> In c (dom_defns d).
Proof.
 intros.
 induction d. 
Case "nil". inversion H.
Case "cons". 
  unfold get_defns_ctrs_of in H.
  destruct a. destruct p. 
  destruct (eq_ctr_dec ctr0 c1). 
  Case "ctr0 = c1".
    inversion H.   subst. unfold dom_defns. simpl. right.    fold dom_defns.  
      assert (J1: In c (dom_ctr_decls c0) \/ In c (dom_defns d)).  left; auto.
   apply (in_or_app (dom_ctr_decls c0) (dom_defns d) c J1).

  Case "ctr0 <> c1".
     fold get_defns_ctrs_of in H. 
     assert (J:= IHd H).
    unfold dom_defns.     simpl. right. fold dom_defns. 
    remember (dom_ctr_decls c0) as A.
   assert (J1: In c A\/ In c (dom_defns d)).  right; auto.
   apply (in_or_app A (dom_defns d) c J1).
Qed. 

Lemma get_ctrs_of_sig_in: forall ctr0 c l S, 
  get_ctrs_of ctr0 S = Some l -> In c l -> In c (dom_sig S).
Proof. 
intros. 
induction S. 
 Case "nil". simpl in H.  inversion H.
 simpl in H. destruct a.
 Case " a = bundle_defns d".   
    remember (get_defns_ctrs_of ctr0 d) as L.
    destruct L.
    Case "L = some".
      rewrite H in HeqL. symmetry in HeqL.
      unfold dom_sig; simpl.
      apply in_or_app. left. 
      apply (get_defns_ctr_some_in2 c ctr0 d HeqL); auto.
    Case "L = None".
    unfold dom_sig; simpl.
    apply in_or_app. right. apply IHS; auto. 

 Case "a=bundle_assert..".
  simpl.  right. apply IHS; auto.
Qed.
 
 Lemma get_defns_ctr_some_in1: forall ctr0 d l, 
  get_defns_ctrs_of ctr0 d = Some l -> In ctr0 (dom_defns (get_tctr_defns_of d)).
Proof.
 intros.
 induction d. 
Case "nil". inversion H.
Case "cons". 
  unfold get_defns_ctrs_of in H.
  destruct a. destruct p. 
  destruct (eq_ctr_dec ctr0 c0). 
  Case "ctr0 = c0". 
   subst. unfold dom_defns; simpl.  left; auto.
  Case "ctr0 <> c0".
     fold get_defns_ctrs_of in H. 
     assert (J:= IHd H).
    unfold dom_defns. 
    simpl. right. 
    apply J.
Qed. 


Lemma ctr_decls_lookup_some_in: forall ctr0 c0 k,
   ctr_decls_lookup  ctr0 c0 = Some k -> In ctr0 (dom_ctr_decls c0).
Proof.
intros.
induction c0.
 Case "nil". inversion H.
 Case "cons".
  unfold ctr_decls_lookup in H.
  destruct a.  destruct (eq_ctr_dec ctr0 c).
    Case "ctr0 = c".
    subst. simpl.  left; auto. 
   Case "ctr0 <>c". 
    fold ctr_decls_lookup in H.
    simpl. right. eapply IHc0; eauto.
Qed.

Lemma  defns_lookup_some_in: forall c d k,
  bundle_defns_ctr_lookup c d = Some k -> In c (dom_defns d).
Proof. 
 intros.
 induction d. 
Case "nil". inversion H.
Case "cons". 
   simpl in H.
  destruct a. destruct p. 
  destruct (eq_ctr_dec c c1). 
  Case "c = c1". 
   subst. unfold dom_defns; simpl.  left; auto.
  Case "c <> c1".
     remember (ctr_decls_lookup c c0) as L. 
     destruct L. 
     Case "L=Some". 
     symmetry in HeqL.
     unfold dom_defns; simpl.  right. apply in_or_app.  left. 
     eapply (ctr_decls_lookup_some_in); eauto.
     Case "L=None".
      unfold dom_defns; simpl.  right. apply in_or_app.  right.
      fold dom_defns.
     apply IHd; auto.
Qed. 

Lemma get_ctr_sig_in: forall c S l,
  get_ctrs_of c S = Some l -> In c (dom_sig S).
Proof.
intros. 
induction S. 
 Case "nil". simpl in H.  inversion H.
 simpl in H. destruct a.
 Case " a = bundle_defns d".   
    remember (get_defns_ctrs_of c d) as L.
    destruct L.
    Case "L = some".
      rewrite H in HeqL. symmetry in HeqL.
      unfold dom_sig; simpl.
      apply in_or_app. left. 
      apply (get_defns_ctr_some_in c  d HeqL); auto.
    Case "L = None".
    unfold dom_sig; simpl.
    apply in_or_app. right. apply IHS; auto. 

 Case "a=bundle_assert..".
  simpl.  right. apply IHS; auto.
Qed.

Lemma sig_lookup_ctr_some_in: forall c S t, 
  sig_lookup_ctr c S = Some t -> In c (dom_sig S).
Proof. 
intros. 
induction S. 
 Case "nil". simpl in H.  inversion H.
 simpl in H. destruct a.
 Case " a = bundle_defns d".   
    remember (bundle_defns_ctr_lookup c d) as L.
    destruct L.
    Case "L = some".
      rewrite H in HeqL. symmetry in HeqL.
      unfold dom_sig; simpl.
      apply in_or_app. left. 
      eapply (defns_lookup_some_in); eauto.
    Case "L = None".
    unfold dom_sig; simpl.
    apply in_or_app. right. apply IHS; auto. 

 Case "a=bundle_assert..".
  simpl.  right. apply IHS; auto.
Qed.

Lemma sig_lookup_assn_some_in: forall c S t, 
  sig_lookup_assn c S = Some t -> In c (dom_sig S).
Proof. 
intros. 
induction S. 
 Case "nil". simpl in H.  inversion H.
 simpl in H. destruct a.
 Case " a = bundle_defns d".   
    unfold dom_sig; simpl.
    apply in_or_app. right. apply IHS; auto. 

 Case "a=bundle_assert..".
  destruct (eq_ctr_dec c c0). 
  subst.  simpl.  left; auto.
  simpl.  right. apply IHS; auto.
Qed.

Lemma dom_defns_distrib: forall ds ds1, 
  dom_defns (ds ++ds1) = (dom_defns ds) ++ (dom_defns ds1).
Proof.  
 intros. 
 induction ds. 
 Case "nil". auto.
 Case "cons". 
  unfold dom_defns; simpl. 
  destruct a. destruct p. fold dom_defns.
 rewrite (IHds). rewrite <- app_comm_cons. rewrite app_ass. auto.
Qed. 


Lemma bundle_has_strong_type_inv: forall P S1 S2 d k c n  ctr0 l k1 t ds,
  wf_bundle_tctrs S1 d k1 ->
  wf_bundle_ctrs P (bundle_defns(get_tctr_defns_of(ds++d))::S2) (bundle_defns(ds++d)::S2) d ->
  defns_has_ctr ctr0 k d ->
  get_defns_ctrs_of ctr0 d = Some l ->
  In (c,n) l ->
  defns_has_ctr (c,n) t d -> 
 (exists c_inst, exists args, exists nt,
   has_level t nt /\ constructs t c_inst /\ fully_applied_args c_inst ctr0 args 
   /\ fully_applied_typ c_inst ctr0 k
   /\ length args = n
   /\ ~(nt < n) 
   /\ all_tm_bv (nt -n) args).

Proof. 
 intros.
 remember (bundle_defns(ds++d) ::S2) as S3.
gen ds.
 induction d.
 Case "nil".
   inversion H1.

 Case "cons".
  intros. 
   unfold defns_has_ctr in H1.
   unfold bundle_defns_ctr_lookup in H1.
   destruct a. destruct p.
   destruct (eq_ctr_dec ctr0 c1).

   Case "ctr0=c1". 
      inversion H. inversion H1. subst. 
      unfold get_defns_ctrs_of in H2.
     destruct (eq_ctr_dec c1 c1).
         Case "c1 = c1".
         inversion H2. inversion H0; subst.  
         clear IHd H2.
         inversion H4.  destruct (eq_ctr_dec (c,n) c1). 
             Case " (c,n) = c1".
             rewrite e0 in H3; contradiction.

             Case "(c,n)<>c1".
             remember(ctr_decls_lookup (c,n) c0) as CTR.
             destruct CTR.
                 Case "(c,n) = c0".
                 inversion H5; subst. 
                 symmetry in HeqCTR. apply (ctr_decls_has_strong_type_inv H19 HeqCTR).
                 Case "(c,n) <> c0".         
                 symmetry in HeqCTR.
                 assert (J:= defns_lookup_some_in (c, n) d H5).
                 destruct( H20 (c,n) H3 J).
 
        Case "c1<>c1". 
          tauto.
     
 Case "ctr0<>c1". 
   remember (ctr_decls_lookup ctr0 c0) as CTRDecl.
   destruct CTRDecl.
      Case "CTRDecl = Some t".
      unfold get_defns_ctrs_of in H2. 
      destruct (eq_ctr_dec ctr0 c1). contradiction.  
      clear IHd. fold get_defns_ctrs_of in H2. 
      inversion H0. 
      inversion H; subst. fold get_defns_ctrs_of in H2.
      assert (J1:= get_defns_ctr_some_in ctr0 d H2).
      symmetry in HeqCTRDecl.
      assert (J2:= ctr_decls_lookup_some_in ctr0 c0 HeqCTRDecl).
      unfold disjoint in H13. assert (J:= H14 ctr0 J2). contradiction.

    Case "CTRDecls = None".
    fold bundle_defns_ctr_lookup in H1.
    inversion H.  inversion H0.   inversion H2.   inversion H4. subst. 
    destruct (eq_ctr_dec ctr0 c1).
        Case "ctr0 = c1".  contradiction.
        Case "ctr0 <>c1". 
        destruct (eq_ctr_dec (c,n) c1).
            Case "(c,n) = c1". clear IHd.
            assert (J:= get_defns_ctr_some_in2 (c,n) ctr0 d H27 H3).
            rewrite e in J; contradiction.
            Case "(c,n) <>c1".
            remember(ctr_decls_lookup (c,n) c0) as CTR.
             destruct CTR.
                 Case "(c,n) = c0". clear IHd.
                  assert (J1:= get_defns_ctr_some_in2 (c,n ) ctr0 d H27 H3).
                  symmetry in HeqCTR.
                  assert (J2:= ctr_decls_lookup_some_in (c,n) c0 HeqCTR). 
                  unfold disjoint in H24. destruct (H25 (c,n) J2 J1).
                 Case "(c,n) <> c0".
                 eapply (IHd H9 H1 H27 H28 (ds++(c1,t0,c0)::nil));  eauto.
                  rewrite  app_ass. rewrite <- app_comm_cons. auto.
                  rewrite app_ass.  rewrite <- app_comm_cons. auto.
Qed.


Lemma get_ctrs_lookup_same:  forall ctr0 d,  
  bundle_defns_ctr_lookup ctr0 d =  None ->  get_defns_ctrs_of ctr0 d = None.
Proof.    
  intros. induction d.
   Case "nil". auto. 
   Case "cons".
     simpl in H. destruct a. destruct p. destruct (eq_ctr_dec ctr0 c0). 
          Case "ctr0 = c0". inversion H.   
          Case "ctr0 <> c0". destruct (ctr_decls_lookup ctr0 c). inversion H.
            assert (J:= IHd H).
           simpl.  destruct (eq_ctr_dec ctr0 c0); tauto.  
Qed.

Lemma sig_has_ctr_strong_inv1: forall ctr0 k S l c n t,
  wf_sig S ->
  sig_has_ctr ctr0 k S ->
  get_ctrs_of ctr0 S = Some l ->
  sig_has_ctr (c,n) t S ->
  In (c,n) l ->
  ( exists d, 
   (exists S', exists K, wf_bundle_tctrs S' d K) 
   /\ (exists P, exists S', wf_bundle_ctrs P (bundle_defns (get_tctr_defns_of d) ::S')  (bundle_defns d ::S') d) /\ defns_has_ctr ctr0 k d
   /\ get_defns_ctrs_of ctr0 d = Some l  /\ defns_has_ctr (c,n) t d).
Proof. 
  intros ctr0 k  S l c n t Hwf_sig Hhas Hget_ctrs Hhas_ctr Hin.
  induction Hwf_sig. 
  Case "nil".
    inversion Hhas.
     
  Case "assert".
    inversion Hhas.
    apply IHHwf_sig.
    unfold sig_has_ctr. apply H3.
    inversion Hget_ctrs; auto.
    inversion Hhas_ctr; auto.

  
  Case "defns1".
    inversion Hhas. 
    remember (bundle_defns_ctr_lookup ctr0 d) as Lookup.
    destruct Lookup.
      Case "bundle_defns_ctr_lookup ctr0 d = Some t0". 
      inversion H3. subst. 
      exists d. split.
         exists S K_TYPE; auto. split. 
           exists (nil:(list ctr)) S;auto. split. 
              symmetry in HeqLookup; auto.
               assert (J: get_defns_ctrs_of ctr0 d = Some l).
               inversion Hget_ctrs.
               remember (get_defns_ctrs_of ctr0 d) as Getctrs. 
               destruct Getctrs; auto.
                  Case "get_ctr_of ctr0 S = None".  clear IHHwf_sig. 
                  symmetry in HeqLookup.
                  destruct (H0 ctr0 (defns_lookup_some_in  ctr0 d HeqLookup)(get_ctr_sig_in ctr0 S H4) ). 
                 Case "get_ctr_of ctr0 S = Some l".  clear IHHwf_sig.
                   split; auto.
                   inversion Hhas_ctr.
                   remember (bundle_defns_ctr_lookup (c,n) d) as Bundlectr.
                   destruct Bundlectr.
                   Case "bundle_defns_lookup (c,n) d = Some t0". 
                   inversion H4; subst; symmetry in HeqBundlectr; auto.
                   Case "bundle_defns_lookup (c,n) d = None". 
                   assert (Jin:= get_defns_ctr_some_in2 (c,n) ctr0 d J Hin ).
                     assert (J1:= sig_lookup_ctr_some_in  (c,n) S H4).  
                     destruct (H0 (c,n) Jin J1 ).
         Case "bundle_defns_ctr_lookup ctr0 d = None".
         apply IHHwf_sig; auto. clear IHHwf_sig.
         inversion Hget_ctrs.  
         remember (get_defns_ctrs_of ctr0 d) as Getdefns. 
         destruct Getdefns; auto.
          symmetry in HeqGetdefns. 
            destruct (H0 ctr0 (get_defns_ctr_some_in  ctr0 d HeqGetdefns)
              (sig_lookup_ctr_some_in ctr0 S H3) ). 

            inversion Hhas_ctr. 
            remember (bundle_defns_ctr_lookup (c,n) d) as Bundlectr.
            destruct Bundlectr; auto. inversion H4. subst. clear IHHwf_sig.
             inversion Hget_ctrs.
               remember (get_defns_ctrs_of ctr0 d) as Getctrs. 
               destruct Getctrs. 
                  inversion H5; subst.  symmetry in HeqLookup. symmetry in HeqGetctrs.  
                   assert (J:= get_ctrs_lookup_same ctr0 d HeqLookup). rewrite J in HeqGetctrs. inversion HeqGetctrs.  
                  symmetry in HeqBundlectr. 
                  destruct( H0  (c,n) (defns_lookup_some_in  (c,n) d HeqBundlectr) (get_ctrs_of_sig_in ctr0 (c,n) S H5 Hin)). 
                                                           

 Case "defns2".
   inversion Hhas. 
    remember (bundle_defns_ctr_lookup ctr0 d) as Lookup.
    destruct Lookup.
      Case "bundle_defns_ctr_lookup ctr0 d = Some t0". 
      inversion H3. subst. 
      exists d. split.
         exists S K_PROP; auto. split. 
           exists (dom_defns d) S;auto. split. 
              symmetry in HeqLookup; auto.
               assert (J: get_defns_ctrs_of ctr0 d = Some l).
               inversion Hget_ctrs.
               remember (get_defns_ctrs_of ctr0 d) as Getctrs. 
               destruct Getctrs; auto.
                  Case "get_ctr_of ctr0 S = None".  clear IHHwf_sig. 
                  symmetry in HeqLookup.
                  destruct (H0 ctr0 (defns_lookup_some_in  ctr0 d HeqLookup)(get_ctr_sig_in ctr0 S H4) ). 
                 Case "get_ctr_of ctr0 S = Some l".  clear IHHwf_sig.
                   split; auto.
                   inversion Hhas_ctr.
                   remember (bundle_defns_ctr_lookup (c,n) d) as Bundlectr.
                   destruct Bundlectr.
                   Case "bundle_defns_lookup (c,n) d = Some t0". 
                   inversion H4; subst; symmetry in HeqBundlectr; auto.
                   Case "bundle_defns_lookup (c,n) d = None". 
                   assert (Jin:= get_defns_ctr_some_in2 (c,n) ctr0 d J Hin ).
                     assert (J1:= sig_lookup_ctr_some_in  (c,n) S H4).  
                     destruct (H0 (c,n) Jin J1 ).
         Case "bundle_defns_ctr_lookup ctr0 d = None".
         apply IHHwf_sig; auto. clear IHHwf_sig.
         inversion Hget_ctrs.  
         remember (get_defns_ctrs_of ctr0 d) as Getdefns. 
         destruct Getdefns; auto.
          symmetry in HeqGetdefns. 
            destruct (H0 ctr0 (get_defns_ctr_some_in  ctr0 d HeqGetdefns)
              (sig_lookup_ctr_some_in ctr0 S H3) ). 

            inversion Hhas_ctr. 
            remember (bundle_defns_ctr_lookup (c,n) d) as Bundlectr.
            destruct Bundlectr; auto. inversion H4. subst. clear IHHwf_sig.
             inversion Hget_ctrs.
               remember (get_defns_ctrs_of ctr0 d) as Getctrs. 
               destruct Getctrs. 
                  inversion H5; subst.  symmetry in HeqLookup. symmetry in HeqGetctrs.  
                   assert (J:= get_ctrs_lookup_same ctr0 d HeqLookup). rewrite J in HeqGetctrs. inversion HeqGetctrs.  
                  symmetry in HeqBundlectr. 
                  destruct( H0  (c,n) (defns_lookup_some_in  (c,n) d HeqBundlectr) (get_ctrs_of_sig_in ctr0 (c,n) S H5 Hin)). 
                                            
Qed.

Lemma sig_has_ctr_strong_typing_inv: forall S k ctr0 l c n t,
 wf_sig S ->
 sig_has_ctr ctr0 k S ->
 get_ctrs_of  ctr0 S = Some l ->
 In (c, n) l ->
 sig_has_ctr (c, n) t S ->
 (exists c_inst, exists args, exists nt, 
   constructs t c_inst /\ fully_applied_args c_inst ctr0 args 
   /\ fully_applied_typ c_inst ctr0 k 
   /\ length args = n 
   /\ has_level t nt /\ ~(nt < n) 
   /\ all_tm_bv (nt -n) args).
Proof.
 intros S k ctr0 l c n t Hwf_sig Hhas Hget_ctr Hin Hsig_has.
  assert(J := sig_has_ctr_strong_inv1 Hwf_sig Hhas Hget_ctr Hsig_has Hin).
  destruct  J as [d [Hwf_tctrs [Hwf_ctrs [H_has [H_getctr Hdefns_has]]]]].
  destruct Hwf_tctrs as [S' [K Htctrs]]. 
destruct Hwf_ctrs as [P [S1 Hctrs]].
 assert (J:=bundle_has_strong_type_inv nil Htctrs Hctrs H_has H_getctr Hin Hdefns_has).
  firstorder. 
Qed.

Lemma get_ctrs_typ_inv: forall S c l,
 get_ctrs_of c S = Some l ->
 exists t, sig_has_ctr c t S.
Proof.
 intros. 
 induction S. 
 Case "nil". inversion H.
 Case "cons".
  simpl in H. destruct a.
   Case "bundle_defns".
   remember (get_defns_ctrs_of c d) as G.
   destruct G.
     Case "G = Some".
     inversion H; subst.  unfold sig_has_ctr. unfold sig_lookup_ctr. 
     remember (bundle_defns_ctr_lookup c d) as B.
     destruct B. 
         Case "B = some ".
         exists t; auto.
         Case "B=None".
         assert (HeqB': bundle_defns_ctr_lookup c d = None); auto.
         assert (J:=  get_ctrs_lookup_same c d HeqB').    
         rewrite J in HeqG; inversion HeqG.

     Case "G=None".
     destruct (IHS H) as [u J].
     unfold sig_has_ctr. unfold sig_lookup_ctr. 
     remember (bundle_defns_ctr_lookup c d) as B.
     destruct B. 
         Case "B = some ".
         exists t; auto.
         Case "B=None".
         exists u; auto.
 
   Case "bundle_assn". 
      destruct (IHS H) as [u J].
      exists u. unfold sig_has_ctr. unfold sig_lookup_ctr; auto.  
Qed.


(**************************************************)
(* Lemmas about uniqueness of types and conversion *)

Lemma wf_ctr_unique: forall S ctr t1 t2,
 wf_sig S 
-> sig_has_ctr ctr t1 S
-> sig_has_ctr ctr t2 S
-> t1 = t2.
Proof.
intros. 
 induction H.
 Case "nil". inversion H0.
 Case "assn". 
      unfold sig_has_ctr in H0; simpl in H0.
      unfold sig_has_ctr in H1; simpl in H1.
      apply IHwf_sig; auto.
 Case "defn_type". 
       unfold sig_has_ctr in H0; simpl in H0.
       remember (bundle_defns_ctr_lookup ctr0 d) as B.
       destruct B.
       Case "B = some". 
          unfold sig_has_ctr in H1; simpl in  H1.
            remember (bundle_defns_ctr_lookup ctr0 d) as B1.
            destruct B1. inversion H0.  inversion H1. inversion HeqB; subst; auto.
            inversion HeqB.
       Case "B = None". 
           unfold sig_has_ctr in H1; simpl in  H1.
            remember (bundle_defns_ctr_lookup ctr0 d) as B1.
            destruct B1.  inversion HeqB.
            apply IHwf_sig; auto.
Case "defn_prop". 
       unfold sig_has_ctr in H0; simpl in H0.
       remember (bundle_defns_ctr_lookup ctr0 d) as B.
       destruct B.
       Case "B = some". 
          unfold sig_has_ctr in H1; simpl in  H1.
            remember (bundle_defns_ctr_lookup ctr0 d) as B1.
            destruct B1. inversion H0.  inversion H1. inversion HeqB; subst; auto.
            inversion HeqB.
       Case "B = None". 
           unfold sig_has_ctr in H1; simpl in  H1.
            remember (bundle_defns_ctr_lookup ctr0 d) as B1.
            destruct B1.  inversion HeqB.
            apply IHwf_sig; auto.
Qed.


Lemma wf_assn_unique: forall S ctr t1 t2,
 wf_sig S 
-> sig_has_assn ctr t1 S
-> sig_has_assn ctr t2 S
-> t1 = t2.
Proof.
intros.
 induction H.
 Case "nil". inversion H0.
 Case "assn".
    unfold sig_has_assn in H0; simpl in H0.
    unfold sig_has_assn in H1; simpl in H1.
    destruct (eq_ctr_dec ctr0 c).
     inversion H0. inversion H1. subst; auto.
   apply IHwf_sig; auto.
 Case "defn_type". 
   unfold sig_has_ctr in H0; simpl in H0.
   unfold sig_has_ctr in H1; simpl in H1.
   apply IHwf_sig; auto.
   
 Case "defn_prop". 
   unfold sig_has_ctr in H0; simpl in H0.
   unfold sig_has_ctr in H1; simpl in H1.
   apply IHwf_sig; auto.
Qed.    

(**************************************************)


Ltac solve_disjoint := 
  try solve[unfold dom_sig; simpl; unfold disjoint;
            simpl; eauto; intuition (eauto;omega)].

(* Extending a signature with a fresh type declaration preserves
   well-formedness. *)

(*
Lemma wf_env_fresh_sig : forall n S k,
  wf_env S empty ->
  k = K_TYPE \/ k = K_PROP ->
  ~ (In n (dom_sig S)) ->
  wf_env ( bundle_defns((n,k,nil)::nil)::S) empty.
Proof with try solve[simpl; intuition eauto; solve_disjoint].
  intros n S k Hwf Hkind Hnotin.
  inversion Hwf.
  apply wf_env_nil.
Qed.
*)

Tactic Notation "gather_atoms'" "as" ident(name) :=
  let tmp := gather_atoms in
  let tmp' := beautify_fset tmp in
  let l := fresh name in
  set(l:=tmp').

(******   Lemmas for proving substitution: Copied from Static.v  *********)
(**** Well-formedness  ****)
Lemma wf_env_ok : forall S1 S2 E,
  wf_env S1 S2 E -> ok E.
Proof.
  intros.
  induction H; auto.
Qed.

Hint Resolve wf_env_ok.

Lemma wf_env_contraction : forall S1 S2 E F,
  wf_env S1 S2 (E & F) ->
  wf_env S1 S2 E.
Proof.
  intros.
  induction F. simpl in H. trivial.
  apply IHF. inversion H; trivial.
Qed.

Lemma wf_tm_gives_wf_env: forall S1 S2 E e t,
  wf_tm S1 S2 E e t ->
  wf_env S1 S2 E.
Proof.
  intros. induction H;eauto.
Qed.


Lemma wf_tm_branches_lc_mutual: forall S1, forall S2, 
  (forall E e t, wf_tm S1 S2 E e t -> lc e)
/\
  (forall E s  args b  t_res,
    wf_branches  S1 S2 E s args b t_res -> lc b).
Proof.
  intros.
  apply mutual_wf_tm_branches_ind; eauto.

 intros. unfold K_TYPE. auto.
 intros. unfold K_PROP. auto.

intros. repeat (eapply lc_app).
unfold P_SAYS; auto. 
trivial. trivial.

Case "a sign p".
intros. repeat (eapply lc_app).
unfold PF_SIGN; auto. trivial. trivial.

intros. unfold T_PRIN; auto.

intros. eapply lc_app.
unfold T_PF; auto. trivial.

intros. eapply lc_app. unfold TM_PF_RETURN; auto. trivial.

intros. repeat (eapply lc_app).
unfold TM_PF_BIND; auto. trivial. trivial.

intros. repeat (eapply lc_app).
unfold TM_SAY; auto. trivial.

intros. repeat (eapply lc_app).
unfold TM_PIF; auto. trivial. trivial.
pick_fresh z. apply H1 with (x:=z). auto.
trivial.

intros; unfold TM_SELF; auto.

intros. repeat (eapply lc_app).
unfold PF_SAYS_RETURN; auto. trivial. trivial.

intros. repeat (eapply lc_app).
unfold PF_SAYS_BIND; auto. trivial. trivial.

intros. unfold TM_NONE; trivial.
Qed.



(***** Binding properties ******)

Lemma binds_var_weakening: forall S1 S2 E F G x t,
  binds_var x t (E & G) ->
  wf_env S1 S2 (E & F & G) ->
  binds_var x t (E & F & G).
Proof.
  intros. unfold binds_var. apply binds_weaken; eauto.
Qed.

Lemma binds_eq_weakening: forall S1 S2 E F G x e1 e2 t,
  binds_eq x e1 e2 t (E & G) ->
  wf_env S1 S2 (E & F & G) ->
  binds_eq x e1 e2 t (E & F & G).
Proof.
  intros. unfold binds_eq. apply binds_weaken; eauto.
Qed.

Lemma binds_fresh_neq: forall b (E: env) x y,
  binds x b E ->
  y # E ->
  x <> y.
Proof.
intros.
  intro. eapply binds_fresh; eauto.
Qed.

Corollary binds_var_fresh_neq: forall t (E: env) x y,
  binds_var x t E ->
  y # E ->
  x <> y.
Proof.
  intros. unfold binds_var in *. 
  eapply binds_fresh_neq; eauto.
Qed.

(**** Environement Rewrite ******)


Lemma subst_env_commute_push: forall e z F x t,
  (subst_env z e F) & x ~: ([z~>e] t) =
      subst_env z e (F & x ~: t).
Proof.
  reflexivity.
Qed.


Lemma subst_env_commute_push': forall e z E F x t,
  E & (subst_env z e F) & x ~: ([z~>e] t) =
      E & subst_env z e (F & x ~: t).
Proof.
  reflexivity.
Qed.

(***** Some facts about fv ******)
(* Might be in Lib_FinSet.v ? *)
Lemma notin_union: forall (A B: atoms) a,
  ~ (a \in A) -> ~ (a \in B) ->
  ~ (a \in (A \u B)).
Proof.
  auto*.
Qed.

Lemma fv_open_tm_aux : forall n s t x,
  ~ (x \in (fv (open_tm_aux n s t))) ->
  ~ (x \in (fv t)). 
Proof.
  intros. generalize dependent n.
  induction t; intros; simpl in *.
    auto. auto. auto. auto.
    eauto using notin_union, (IHt1 n), (IHt2 n).
    eauto using notin_union, (IHt1 n), (IHt2 (S n)).
    eauto using notin_union, (IHt1 n), (IHt2 (S n)).

    repeat (apply notin_union).  eauto using  (IHt1 n).  eauto using (IHt2 n). eauto using  (IHt3 n).

    eauto using notin_union, (IHt1 n), (IHt2 n).
    eauto using notin_union, (IHt1 n), (IHt2 n).
Qed.

Open Scope tm_scope. 
Lemma fv_open_tm: forall s t x,
  ~( x \in fv (t ^^ s) ) ->
  ~( x \in  fv t ).
Proof.
  unfold open_tm.
  apply fv_open_tm_aux.
Qed.
Close Scope tm_scope.


Lemma notin_open: forall x u e n,
 x\notin fv u -> x\notin fv e -> x \notin fv (open_tm_aux n e u).
Proof.
 intros.
 gen n.
 induction u; subst; intros; simpl in H; simpl; auto.
 destruct (b===n); auto.
 apply notin_union. apply IHu1; auto. apply IHu2; auto.
  apply notin_union. apply IHu1; auto. apply IHu2; auto.
  apply notin_union. apply IHu1; auto. apply IHu2; auto.
  repeat (apply notin_union). apply IHu1; auto. apply IHu2; auto.  apply IHu3; auto.
  apply notin_union. apply IHu1; auto. apply IHu2; auto.
  apply notin_union. apply IHu1; auto. apply IHu2; auto.
Qed.

Lemma wf_tm_fresh_l_mutual: 
  forall S1 S2, (forall E e t, wf_tm S1 S2 E e t ->
     forall x,
       x # E -> x \notin fv e)
/\
  (forall E s args b t_res,
   wf_branches S1 S2 E s args b t_res ->
     forall x, 
       x # E -> x \notin fv b).

Proof.
introv.
 apply mutual_wf_tm_branches_ind.
  Case ("K_TYPE"). 
  intros.  eapply notin_empty.

 Case ("K_PROP").
    intros. eapply notin_empty.  
 
  Case ("arr").
   intros. simpl. apply notin_union. 
    apply H; auto.
    pick_fresh z.
    eapply fv_open_tm. apply (H0 z). auto*.
    simpl. auto*.


 Case ("ctr").
 intros. simpl. auto.

Case "wf_tm_assn".
intros. simpl.  auto.

 Case ("fv").
 intros. simpl. unfold not. intro. replace x with x0 in b. 
assert (J : x0 <> x0).
unfold binds_var in b.
eapply binds_fresh_neq with  (b := bnd_var t) (E:=E) (x:=x0) (y:=x0).
auto. auto. auto. 
assert (J:= in_singleton x0 x).
destruct J. auto.

  Case ("abs"). 
    intros.
    simpl. apply notin_union. auto.
    pick_fresh z.
    eapply fv_open_tm. apply (H0 z). auto*.
    simpl. auto*.

 Case ("app").
    intros. simpl. apply notin_union; auto.

  Case ("says"). 
    intros.
    simpl. repeat (apply notin_union); auto.

Case "a sign p".
intros; simpl; repeat (apply notin_union); auto.

  Case ("principal"). 
    intros.
    simpl.  auto.

  Case ("pf"). 
    intros.
    simpl.  apply notin_union; auto.

(*  Case ("pf ret"). *)
    intros.
    simpl.  apply notin_union; auto.

(*  Case ("pf bind"). *)
    intros.
    simpl.  repeat (apply notin_union); auto.

(*  Case ("says"). *)
    intros.
    simpl. (apply notin_union); auto.

(*  Case ("PIF"). *)
    intros.
    simpl. repeat (apply notin_union); auto.

 pick_fresh z.
 apply (H1 z). auto*.
    simpl. auto*.

(*  Case ("TM_SELF"). *)
    intros.
    simpl.  auto.

(*  Case ("PF_SAYS_RETURN"). *)
    intros.
    simpl.   repeat (apply notin_union); auto. 

(*  Case ("PF_SAYS_BIND"). *)
    intros.
    simpl.   repeat (apply notin_union); auto. 

(*  Case ("match"). *)
    intros.
    simpl.  repeat (apply notin_union); auto. 

(*  Case ("match"). *)
    intros.
    simpl.  repeat (apply notin_union); auto.

(*  Case ("cast"). *)
    intros.
    simpl.  repeat (apply notin_union); auto.

(* Case ("tm_sign"). *)
   intros.
   simpl. repeat (apply notin_union); auto.

(*  Case ("TM_NONE"). *)
    intros.
    simpl.  auto.

(*  Case ("tm_brn"). *)
    intros.
    simpl.  repeat (apply notin_union); auto.
    
Qed.


Lemma wf_tm_fresh_l: forall S1 S2 E e t x,
  x # E ->
  wf_tm S1 S2 E e t ->
  x \notin fv e.
Proof.
  intros. 
  eapply (proj1 (wf_tm_fresh_l_mutual S1 S2)); eauto.
Qed.

Lemma binds_var_fresh: forall S1 S2 E x t,
  wf_env S1 S2 E ->
  binds_var x t E ->
  x \notin fv t.
Proof.
  intros.
  induction E. inversion H0.
  destruct a as [z s].
  unfold binds_var in H0. 
  simpl_env. binds_cases H0.
    apply IHE; auto.
      inversion H; auto.
    inversion H. subst.
    inversion H. subst.
    eapply wf_tm_fresh_l with (S1:=S1) (E:=E) (e:= t) (t:=k) ; eauto.
   inversion H2.  subst t0. apply H5.
  subst s. inversion H2.
Qed.

Lemma sig_has_ctr_fv: forall c t S x,
  wf_sig S ->
  sig_has_ctr c t S ->
  x \notin fv t.
Proof.
intros.
 destruct (sig_has_ctr_typ_inv H H0) as [S1' [S2' [k J]]].
  eapply wf_tm_fresh_l with (E:=(empty:env)). eapply notin_empty. 
 eauto.
 Qed.

Lemma sig_has_assn_fv: forall c t S x,
  wf_sig S ->
  sig_has_assn c t S ->
  x \notin fv t.
Proof.
intros.
 destruct (sig_has_assn_typ_inv H H0) as [S1' [S2' [k J]]].
  eapply wf_tm_fresh_l with (E:=(empty:env)). eapply notin_empty. 
 eauto.
 Qed.

Inductive step_matches_ctr: tm -> ctr -> tm -> (tm * nat * list tm * list tm) -> Prop :=
| step_matches_ctr_base: forall c body n,
       step_matches_ctr (tm_ctr (c, n)) (c, n) body (body, n, nil, nil)

| step_matches_ctr_strip_type_arg: forall v1 v2 m c body args,
       value v2
    -> step_matches_ctr v1 c body (body, m, args, nil)
    -> m > 0
    -> step_matches_ctr (tm_app v1 v2) c body (body, m-1, args++(v2::nil), nil)

| step_matches_ctr_rec: forall v1 v2  c body e args args1,
       value v2
    -> step_matches_ctr v1 c body (e, 0, args, args1)
    -> step_matches_ctr (tm_app v1 v2) c body ((tm_app e v2), 0, args, args1++ v2::nil)
.

Inductive step_matches_branches : tm -> tm -> tm -> Prop :=
| step_matches_here: forall v c body rest e args args1,
       step_matches_ctr v c body (e, 0, args, args1)
    -> step_matches_branches v (tm_brn c body rest) e

| step_matches_earlier: forall v c body rest e,
       step_matches_branches v rest e
    -> step_matches_branches v (tm_brn c body rest) e
.

Inductive step : tm -> tm -> Prop :=
(* Computational rules *)

| step_app: forall t e v,
       value v 
    -> step (tm_app (tm_abs t e) v) (e ^^ v)

| step_pf_bind: forall p f,
       value f 
    -> step (tm_app (tm_app TM_PF_BIND (tm_app TM_PF_RETURN p)) f)
            (tm_app f p)

| step_say: forall p,
       step (tm_app TM_SAY p) (tm_app TM_PF_RETURN (tm_app (tm_app PF_SIGN TM_SELF) p))

| step_pif1: forall v1 v2 e1 e2, 
       v1 = v2 (* note -- in principle, this should be equality defined for runtime-principal values *)
    -> step (tm_app (tm_app (tm_app (tm_app TM_PIF v1) v2) e1) e2) e1 

| step_pif2: forall v1 v2 e1 e2, 
       v1 <> v2 (* note -- in principle, this should be equality defined for runtime-principal values *)
    -> step (tm_app (tm_app (tm_app (tm_app TM_PIF v1) v2) e1) e2) e2

| step_mat: forall v t branches e,
       step_matches_branches v branches e
    -> step (tm_mat v t branches) e

| step_cast: forall v t,
       value v
    -> step (tm_cast v t) v

(* Stuctural congruence rules *)

| step_app_cong1: forall e1 e1' e2,
       step e1 e1'
    -> step (tm_app e1 e2) (tm_app e1' e2)

| step_app_cong2: forall t e1 e2 e2',
       step e2 e2'
    -> step (tm_app (tm_abs t e1) e2) (tm_app (tm_abs t e1) e2')

| step_app_cong3: forall v e2 e2',
       applied_value v  
    -> step e2 e2'
    -> step (tm_app v e2) (tm_app v e2')

| step_pf_ret_cong: forall e1 e1',
       step e1 e1'
  ->  step (tm_app TM_PF_RETURN e1) (tm_app TM_PF_RETURN e1')
 
| step_pf_bind_cong1: forall e1 e1' e2,
       step e1 e1'
    -> step (tm_app (tm_app TM_PF_BIND e1) e2)
            (tm_app (tm_app TM_PF_BIND e1') e2)

| step_pf_bind_cong2: forall e1 e2 e2',
       value e1
    -> step e2 e2'
    -> step (tm_app (tm_app TM_PF_BIND e1) e2)
            (tm_app (tm_app TM_PF_BIND e1) e2')

| step_mat_cong: forall e1 e1' branches t,
       step e1 e1'
    -> step (tm_mat e1 t branches) (tm_mat e1' t branches)

| step_cast_cong: forall e1 e1' t,
       step e1 e1'
    -> step (tm_cast e1 t) (tm_cast e1' t)
.



(***** Metatheory *****)

Lemma converts_weakening: forall S1 S2 E0 t s,
  converts E0 t s ->
  forall E F G,
  E0 = E & G ->
   wf_env S1 S2 (E & F & G) ->
  converts (E & F & G) t s.
Proof. introv Hc.
  induction Hc; intros.
 apply conv_refl; auto.
 apply conv_sym; auto.
 eapply conv_trans with (t2:=t2); auto.
 eapply conv_axiom with (x:=x) (k:=k); auto.
 apply binds_eq_weakening with (S1:=S1) (S2:=S2) (x:=x) (e1:=t1) (e2 :=t2) (t:=k) (E:=E0) (F:=F) (G:=G).
 rewrite H0 in H. auto. assumption.
 eapply conv_app; eauto.
 eapply conv_abs; eauto.
 eapply conv_arr; eauto. (*
 eapply conv_mat; eauto.
 eapply conv_brn; eauto. 
 eapply conv_cast; eauto. *)
Qed.

Corollary converts_weakening': forall S1 S2 E F t s,
  converts E t s ->
  wf_env S1 S2 (E & F) ->
  converts (E & F) t s.
Proof.
  intros. rewrite_env (E & F & empty).
  eapply converts_weakening; simpl; eauto.
Qed.


Lemma typing_branches_weakening_mutual: 
  forall S1 S2,  (forall E0 e t,
   wf_tm S1 S2 E0 e t ->
   forall E F G,
   E0 = E & G ->
   wf_env S1 S2 (E & F & G) ->
   wf_tm S1 S2 (E & F & G) e t)
/\
  (forall E0 b cs t_0 t_res,
   wf_branches S1 S2 E0 b cs t_0 t_res ->
   forall E F G,
   E0 = E & G ->
   wf_env S1 S2 (E & F & G) ->
   wf_branches S1 S2 (E & F & G) b cs t_0 t_res).

Proof with auto*.
  intros. apply mutual_wf_tm_branches_ind; auto.

Case "arr".   
intros. subst. gather_atoms' as L'.
    eapply wf_tm_arr with (L:=L'); subst L'...
      intros. rewrite_env (E0 & F & (G & x~: u1)).
        apply H0...
        simpl_env.  
constructor 2 with (S1:=S1) (S2:=S2) (E:=E0&F&G) (x:=x) (t:=u1) (k:=k1); auto.

Case "fv".
intros.  subst. constructor...
 eapply binds_var_weakening with (x:=x) (t:=t) ( E:=E0) (F:=F) (G:=G); eauto.


Case "abs".
intros.  subst. gather_atoms' as L'.
    eapply wf_tm_abs with (L:=L'); subst L'...
      intros. rewrite_env (E0 & F & (G & x~: u1)).
        apply H0...
        simpl_env. 
constructor 2 with (E:=E0&F&G) (x:=x) (t:=u1) (k:=k1); eauto.
 (*assert (J: x\notin dom E0 \u dom F \u dom G). auto .
    repeat(rewrite <- dom_concat  in J). auto.
*)

Case "app". (*TODO*)
   intros. subst.
    eapply wf_tm_app...

(* Case "PF_BIND"*)
intros. subst. eapply wf_tm_TM_PF_BIND...

(* Case "PF_PIF"*)
intros.  subst. gather_atoms' as L'.
    eapply wf_tm_TM_PIF with (L:=L'); subst L'...
intros. apply H1 with (E:=E0) (F:=F) (x:=x) (G0:= G& single x (bnd_eq v1 v2 T1)) ...
simpl_env. 
constructor 3 with (E:= E0&F&G) (x:=x) (e1:= v1) (e2:= v2) (t:= T1). auto.

eapply H... eapply H0... trivial.
rewrite_env (concat E0 (concat F G)).
repeat rewrite dom_concat.  auto. auto. auto.
eapply H3; eauto.

(*Case "PF_SAYS_BIND"*)
intros. subst.
eapply  wf_tm_PF_SAYS_BIND...

(*Case "tm_mat"*)
intros. subst.
eapply wf_tm_mat_TYPE...

(*Case "tm_mat"*)
intros. subst.
eapply wf_tm_mat_PROP...

(*Case "cast"*)
intros. subst.
eapply wf_tm_cast...
eapply converts_weakening...


(*Case "branch"*)

intros. subst.
eapply wf_branches_Some...
Qed.


Lemma typing_weakening : forall S1 S2 E F G e t,
  wf_tm S1 S2 (E & G) e t ->
  wf_env S1 S2 (E & F & G) ->
  wf_tm S1 S2 (E & F & G) e t.
Proof.
  intros.
  assert(J:= typing_branches_weakening_mutual S1 S2).
  destruct (J ). eauto using H1.
Qed.


Corollary typing_weakening': forall S1 S2 E F e t,
  wf_tm S1 S2 E e t ->
  wf_env S1 S2 (E & F) ->
  wf_tm S1 S2 (E & F) e t.
Proof.
  introv H J. rewrite_env (E & F & empty).
  auto using typing_weakening.
Qed.

(*  Need to make environment rewrite work. simpl_env *)
Lemma binds_inversion: forall x (b:bnd) E,
  binds x b E -> 
  exists L, exists R, (L & x~b & R) = E.
Proof. 
 intros. induction E.
    inversion H.
    inversion H. destruct a as [y s].
    simpl_env. binds_cases H.
 (*   Case "x<>y". *)
      unfold binds_var in *.
      destruct (IHE B) as [L [R H']].
      exists L. exists (R & y ~ s). subst.
      auto.
  (*  Case "x=y".*)
exists E. exists (@empty bnd).
rewrite H0. auto.
Qed.

Lemma binds_var_inversion: forall x t E,
  binds_var x t E ->
  exists L, exists R, (L & x~: t & R) = E.
(* need binds_inverstion*)
Proof.  unfold binds_var.
firstorder (apply binds_inversion with (x:=x) (b:= bnd_var t) (E:= E);auto).
Qed.

Lemma wf_env_strong_inv_var: forall S1 S2 E x t F,
  wf_env S1 S2 (E & x ~: t & F) ->
  exists k, wf_tm S1 S2 E t k.
Proof.
  intros. induction F.
    simpl_env. inversion H. firstorder.
    apply IHF. inversion H; auto.
Qed.

Lemma binds_eq_inversion: forall x e1 e2 t E,
  binds_eq x e1 e2 t E ->
  exists L, exists R, (L & x~(bnd_eq e1 e2 t)& R) = E.
Proof. unfold binds_eq.
firstorder ( auto).
apply binds_inversion with (x:=x) (b:= bnd_eq  e1 e2 t) (E:= E);auto.
Qed.

Lemma wf_env_strong_inv_eq: forall S1 S2 E x e1 e2 t F,
  wf_env S1 S2 (E & x ~bnd_eq e1 e2  t & F) ->
  (wf_tm S1 S2 E e1 t) /\ (wf_tm S1 S2 E e2 t) /\ (exists k, wf_tm S1 S2 E t k).
Proof.   intros. induction F.
    simpl_env. inversion H.
    repeat split; subst; auto. 
    exists K_TYPE; auto.
    apply IHF. inversion H; auto.
Qed.

Lemma env_fresh_concat_l: forall (E: env) F x,
  x # (E & F) ->
  x # E.
Proof.
  intros.
  intro.
 assert (J:=in_union x (dom E) (dom F)).  
apply H. simpl_env.
destruct J. auto. Qed.
 
Lemma env_fresh_concat_r: forall (E: env) F x,
  x # (E & F) ->
  x # F.
Proof. 
  intros.
  intro. apply H.
assert (J:=in_union x (dom E) (dom F)).  
simpl_env.
destruct J. auto.
Qed. 


Lemma wf_env_fresh_l : forall S1 S2 E F z u,
  wf_env S1 S2 (E & z ~: u & F) ->
  z # E.
Proof.
  intros. induction F.
    simpl_env; inversion H; subst; trivial.
    destruct a as [y t]; simpl_env.
apply IHF.
eapply wf_env_contraction with (E:=  E&z~:u & F) (F:= single y t); eauto.
Qed.

Lemma wf_env_fresh_l' : forall S1 S2 E F z v1 v2 k,
  wf_env S1 S2 (E & z ~bnd_eq v1 v2 k & F) ->
  z # E.
Proof.
  intros. induction F.
    simpl_env; inversion H; subst; trivial.
    destruct a as [y t]; simpl_env.
apply IHF.
eapply wf_env_contraction with (E:=  E&z~bnd_eq v1 v2 k  & F) (F:= single y t).
auto.
Qed.

Lemma wf_env_fresh_r: forall S1 S2 E F x u,
 wf_env S1 S2 (E&x~:u&F) -> x\notin  dom F.
Proof.
intros.
 induction F.  auto.
 destruct a as (y, b). rewrite cons_concat. rewrite cons_concat in H.
 rewrite dom_push.
 replace (E&x~:u& F & (single y b)) with ((E&x~:u&F) & (single y b)) in H; auto.
 apply notin_union.
 destruct b.  
 replace ( (E&x~:u&F) & (y~:t)) with (E&x~:u&F& y ~: t & empty ) in H; auto.
 assert (J:=  wf_env_fresh_l (E &x~:u & F) empty  t H). 
 assert (J1:=  env_fresh_concat_l (E & x~:u)  F J).
  assert (J2:=  env_fresh_concat_r E  (x~:u) J1). simpl in J2; auto.

 replace (E&x~:u & (F& y~ bnd_eq t t0 t1)) with (E&x~:u & F& y~ bnd_eq t t0 t1); auto.
  assert (J:=  wf_env_fresh_l' (E &x~:u & F) empty  t t0 t1 H).
 assert (J1:=  env_fresh_concat_l (E & x~:u)  F J).
  assert (J2:=  env_fresh_concat_r E  (x~:u) J1). simpl in J2; auto.

 apply IHF.  replace (E& x~:u & (F&y~b)) with (E& x~:u & F&y~b) in H; auto. 
eapply wf_env_contraction; eauto.
Qed.


Lemma binds_var_fresh_ext : forall S1 S2 E x z t,
  binds_var x t E ->
  wf_env S1 S2 E ->
  z # E ->
  z \notin (fv t).
Proof.
  introv Hbd Hwf Hfr.
    destruct (binds_var_inversion Hbd) as [L [R H]].
    subst E.
    destruct (@wf_env_strong_inv_var S1 S2 L x t R Hwf) as [k Hwf'].
      eapply wf_tm_fresh_l with (E:=L) (t:=k); auto*.
Qed.



Lemma binds_eq_fresh_ext: forall S1 S2 E x e1 e2 z t,
  binds_eq x e1 e2 t E ->
  wf_env S1 S2 E ->
  z # E ->
  z\notin (fv e1) /\ z \notin (fv e2) /\ z \notin (fv t).
Proof.
  introv Hbd Hwf Hfr.
    destruct (binds_eq_inversion Hbd) as [L [R H]].
    subst E.
    destruct (@wf_env_strong_inv_eq S1 S2 L x e1 e2 t R Hwf) as [Hwf1 [Hwf2 [k Hwft]]].
   repeat split.
 eapply wf_tm_fresh_l with (E:=L) (t:=t); auto*.
 eapply wf_tm_fresh_l with (E:=L) (t:=t); auto*.
  eapply wf_tm_fresh_l with (E:=L) (t:=k); auto*.
Qed.


Lemma wf_tm_fresh_t: forall S1 S2 E e t x, 
 wf_sig S1 -> wf_tm S1 S2 E e t -> x # E -> x \notin fv t.
Proof.
intros.
induction H0; simpl; auto. 
 
Case "arr".
 pick_fresh z. assert (J: z \notin L); auto. eapply (H3 z); auto.
   simpl. auto.
Case "ctr". eapply sig_has_ctr_fv with (c:=c); eauto. 
Case "assn". eapply sig_has_assn_fv with (c:=c); eauto.
Case "var". eapply binds_var_fresh_ext; eauto.
Case "abs".  
    assert (J:= wf_tm_fresh_l H1 H0_0).  auto.
Case "app".
   simpl in IHwf_tm1.
   apply notin_open with (n:=0).
   assert (J:= IHwf_tm1 H1); auto.
   eapply wf_tm_fresh_l with (E:= E) ;eauto. 
Case "prin".
 repeat (apply notin_union); auto. 
 eapply wf_tm_fresh_l with (E:=empty:env) ;eauto. 
 eapply wf_tm_fresh_l with (E:= empty:env) ;eauto. 
Case "pf".
  apply notin_union; auto.
Case "pf bind".
  apply notin_union; auto.
  simpl in IHwf_tm2. assert (J:= IHwf_tm2 H1); auto.

  repeat (apply notin_union); auto.
  eapply wf_tm_fresh_l with (E:= E) ;eauto. 

  repeat (apply notin_union); auto.
  eapply wf_tm_fresh_l with (E:= E) ;eauto. 

  repeat (apply notin_union); auto.
 simpl in IHwf_tm1. assert (J:= IHwf_tm1 H1); auto.
 simpl in IHwf_tm2. assert (J:= IHwf_tm2 H1); auto.
   eapply wf_tm_fresh_l with (E:= E) ;eauto. 
  eapply wf_tm_fresh_l with (E:= E) ;eauto.
    eapply wf_tm_fresh_l with (E:= E) ;eauto.
Qed.

Lemma atomic_type_fresh: forall S t x,
   atomic_type S t
  -> x \notin fv t.
Proof.
  intros.
  unfold atomic_type in H.
  destruct H as [L [Jtr J]].
  unfold simple_type in J. 
  destruct t ;subst; unfold check_simple_type in J; simpl J; try solve[ inversion J; simpl; auto]. 
Qed.

Lemma converts_strengthening_var: forall S1 S2 E x u F t1 t2, 
 converts (E&x~:u&F) t1 t2 -> wf_env S1 S2 (E&F) -> converts (E&F) t1 t2.
Proof.
intros.
 remember (E&x~:u&F) as E0. 
 gen E F. induction H; subst; intros.
 apply conv_refl; auto.
 apply conv_sym; eapply IHconverts; auto.
  apply conv_trans with (t2:=t2).  eapply IHconverts1; auto. eapply IHconverts2; auto.
  eapply conv_axiom with (x:=x0) (t1:=t1) (t2:=t2) (k:=k).
  rewrite HeqE0 in H. 
  binds_cases H. 
  rewrite_env  ( E0 & F& empty). replace E0 with (E0&empty) in B0; auto.
  eapply binds_eq_weakening; eauto; auto. 
  rewrite_env (empty & E0& F). replace F with (empty & F) in B0; auto.
  eapply binds_eq_weakening; eauto. rewrite_env (E0&F); eauto.   apply empty_concat.  
  
 eapply conv_app; eauto.
 eapply conv_abs; eauto.
 eapply conv_arr; eauto. (*
  eapply conv_mat; eauto.
 eapply conv_brn; eauto.
  eapply conv_cast; eauto. *)
Qed.
 
Lemma converts_strengthening_eq: forall S1 S2 E x v k F t1 t2, 
 converts (E&x~bnd_eq v v k &F) t1 t2 -> lc v -> wf_env S1 S2 (E&F) -> converts (E&F) t1 t2.
Proof. 
 intros.
 remember (E&x~bnd_eq v v k &F)  as E0.
 induction H.
 apply conv_refl; auto.
 apply conv_sym; auto.
 apply conv_trans with (t2:=t2); auto.
 Case "axiom".
   rewrite HeqE0 in H. binds_cases H.
   apply conv_axiom with (x:=x0) (k:=k0); auto.
   rewrite_env (E&F&empty).  replace E with (E&empty) in B0 ; auto.
   eapply binds_eq_weakening; eauto.
   
   inversion H2; subst. apply conv_refl; auto.

  apply conv_axiom with (x:=x0) (k:=k0); auto.
   rewrite_env (empty&E&F).  replace F with (empty& F) in B0 ; auto.
   eapply binds_eq_weakening; eauto. 
   rewrite_env (E&F); eauto. apply empty_concat.

 apply conv_app; auto.
 Case "abs".
    gather_atoms' as L'.
    apply conv_abs with (L:=L'); auto.
    intros. unfold L' in H4. eapply H3; eauto.

Case "arr".
    gather_atoms' as L'.
    apply conv_arr with (L:=L'); auto.
    intros. unfold L' in H4. eapply H3; eauto.
(*
 apply conv_mat; auto.
 apply conv_brn; auto.
 apply conv_cast; auto. *)
Qed.

Open Scope tm_scope.

Lemma typing_strengthening_var_mutual:
forall S1 S2 ,
wf_sig S1 ->
((forall E0 e k,  wf_tm S1 S2 E0 e k ->
  forall E F x t,  E0 = E& x~: t& F ->
   x\notin fv e -> 
   wf_env S1 S2 (E& F) -> 
   wf_tm S1 S2 (E&F) e k)
/\ (forall E0 s args branches r_tp,  wf_branches S1 S2 E0 s args branches r_tp ->
             forall E F x t, E0 = E& x~: t& F ->
            x\notin (fv branches  \u fv s) ->
            wf_env S1 S2 (E& F) -> 
            wf_branches S1 S2 (E& F) s args branches r_tp )).
Proof with auto*.
intros.
 eapply  mutual_wf_tm_branches_ind; intros. 
Case "K_TYPE".
  apply wf_tm_K_TYPE. auto.

Case "K_PROP".
  apply wf_tm_K_PROP. auto.

Case "arr".
 gather_atoms' as L'.  simpl in H3. 
 apply wf_tm_arr with (L:=L') (k1:=k1); auto.
 eapply H0; eauto.  
 intros. unfold L' in H5.  
 rewrite_env (E0&(F&(x0~:u1))).   
eapply H1 with (x:=x0) (E0:= E0) (F:= F&x0~:u1) (x0:=x) (t:=t); eauto.  rewrite H2; auto. 
 unfold open_tm; eapply notin_open; eauto. simpl; auto.
rewrite_env ( (E0&F) & (x0~:u1)).
eapply wf_env_cons_var; eauto. 

Case "ctr". eapply wf_tm_ctr; eauto.
 Case "assn".  eapply wf_tm_assn; eauto.
Case "fv".  eapply wf_tm_var; eauto.
 simpl in H1.  
 rewrite H0 in b.  binds_cases b. 
   Case "x in E0".
    unfold binds_var.  apply binds_concat_fresh; auto.
  Case "x = x0".
    destruct (notin_same H1).
   auto.
  Case "x in F".
   unfold binds_var.  apply binds_prepend; auto.

Case "abs".
   gather_atoms' as L'. simpl in H4.
   assert (Ju2: forall x0, x0 \notin L' -> wf_tm S1 S2 (E0&F&x0~:u1) (f^^ tm_fv x0) (u2^^tm_fv x0)).
  intros. unfold L' in H6.   
  eapply H1 with (x:=x0) (E0:= E0)  (F:= (F&x0~:u1))  (x0:= x) (t := t); eauto.  rewrite H3; auto. 
  unfold open_tm; eapply notin_open; eauto. simpl; auto.
  rewrite_env (concat (E0&F) (x0~:u1)).
  eapply wf_env_cons_var; eauto. 

  eapply wf_tm_abs with (L:= L'); eauto.  
  eapply H2; eauto.
    simpl. apply notin_union; auto.
  pick_fresh z.  assert (Jz: z \notin L'); auto.
  assert (J:= Ju2 z Jz).
  assert (Jx: x\notin dom(E0&F&z~:u1)).
  assert (Jenv := wf_tm_gives_wf_env w). rewrite H3 in Jenv.
  simpl. rewrite dom_concat; repeat (apply notin_union); auto.
  eapply wf_env_fresh_l; eauto.
  eapply wf_env_fresh_r; eauto.
  assert (J1:= wf_tm_fresh_t H J Jx).
  assert (J2:= fv_open_tm (tm_fv z)  u2 J1); auto.
 
Case "app". 
    simpl in H3; auto.
    assert (J: wf_tm S1 S2 (E0&F) e1 (tm_arr t2 u)).
    eapply H0; eauto.   
    assert (Jx: x\notin dom(E0&F)). 
    assert (Jenv := wf_tm_gives_wf_env w).  rewrite H2 in Jenv. 
    rewrite dom_concat; apply notin_union.
    eapply wf_env_fresh_l; eauto.
    eapply wf_env_fresh_r; eauto.

    assert (J1:= wf_tm_fresh_t H J Jx). simpl in J1.
   eapply wf_tm_app with (t2:=t2); eauto.
(*
   eapply H3; eauto. 
   unfold open_tm; eapply notin_open; eauto. 
 *)
Case "says".
eapply wf_tm_P_SAYS; eauto. eapply H0; eauto. simpl in H3; auto. 
  eapply H1; eauto. simpl in H3; auto.

Case "sign". 
  simpl in H3.
  eapply wf_tm_PF_SIGN; eauto. 

Case "prin". eapply wf_tm_T_PRIN; eauto.

Case "pf". 
eapply wf_tm_T_PF; eauto.  eapply H0; eauto. simpl in H2; auto.  
Case "return".
 assert (J: wf_tm S1 S2 (E0&F) p P).   eapply H0; eauto. simpl in H3; auto.  
 eapply wf_tm_TM_PF_RETURN; eauto. 
 eapply H1; eauto. simpl in H3; auto.
 eapply wf_tm_fresh_t  with (E:= E0 &F); eauto.
 assert (Jenv := wf_tm_gives_wf_env w). rewrite H2 in Jenv.
 rewrite dom_concat; apply notin_union.
 eapply wf_env_fresh_l; eauto.
 eapply wf_env_fresh_r; eauto.
  
Case "pf_bind". 
eapply wf_tm_TM_PF_BIND; eauto. eapply H0; eauto. simpl in H3; auto.   
eapply H1; eauto. simpl in H3; auto.

eapply wf_tm_TM_SAY; eauto. eapply H0; eauto. simpl in H2; auto.

   
simpl in H6.
gather_atoms' as L'.
eapply wf_tm_TM_PIF with (L:= L'); eauto. 
  intros. rewrite_env (E0 & (F & single x0 (bnd_eq v1 v2 T1))).
  unfold L' in H8.
  assert (Jx: x0 \notin L); auto.
 eapply (H2 x0 Jx E0 (F& single x0 (bnd_eq v1 v2 T1)) x t); eauto. 
 rewrite H5 ; auto. 
rewrite_env ( (E0 & F)  & (single x0 (bnd_eq v1 v2 T1))). 
 eapply wf_env_cons_eq; eauto.
 eapply H4; eauto. eapply atomic_type_fresh; eauto.
 eapply H4; eauto. eapply atomic_type_fresh; eauto.

  eapply wf_tm_TM_SELF; eauto.

  assert (J: wf_tm S1 S2 (E0&F) p P).    eapply H0; eauto. simpl in H4; auto. 
  eapply wf_tm_PF_SAYS_RETURN; eauto.
  eapply H1; eauto. simpl in H4; auto. 
 eapply wf_tm_fresh_t  with (E:= E0 &F); eauto.
   assert (Jenv := wf_tm_gives_wf_env w). rewrite H3 in Jenv.
   rewrite dom_concat; apply notin_union.
   eapply wf_env_fresh_l; eauto.
   eapply wf_env_fresh_r; eauto.
  eapply H2; eauto. simpl in H4; auto.

 eapply wf_tm_PF_SAYS_BIND; eauto.
 eapply H0; eauto. simpl in H3; auto. 
  eapply H1; eauto. simpl in H3; auto.

 simpl in H5. assert (J: wf_tm S1 S2 (E0&F) e s). eapply H0; eauto.
 assert (Js: x\notin fv s). eapply wf_tm_fresh_t; eauto.  
 assert (Jenv := wf_tm_gives_wf_env w). rewrite H4 in Jenv.
 rewrite dom_concat; apply notin_union.
 eapply wf_env_fresh_l; eauto.
 eapply wf_env_fresh_r; eauto.
 eapply wf_tm_mat_TYPE with (s:=s); eauto.  

  simpl in H5. assert (J: wf_tm S1 S2 (E0&F) e s). eapply H0; eauto.
  assert (Js: x\notin fv s). eapply wf_tm_fresh_t; eauto. 
  assert (Jenv := wf_tm_gives_wf_env w). rewrite H4 in Jenv.
  rewrite dom_concat; apply notin_union.
  eapply wf_env_fresh_l; eauto.
  eapply wf_env_fresh_r; eauto.
  eapply wf_tm_mat_PROP with (s:=s); eauto.  
    
    assert (J: wf_tm S1 S2 (E0 & F) e t1).  
    eapply H0; eauto. simpl in H3; auto.
   eapply wf_tm_cast with (t1:=t1); eauto.  
  eapply H1; eauto. simpl in H3; auto.   
 (*  eapply  wf_tm_fresh_t; eauto.
   assert (Jenv := wf_tm_gives_wf_env w). rewrite H3 in Jenv.
   rewrite dom_concat; apply notin_union.
   eapply wf_env_fresh_l; eauto.
   eapply wf_env_fresh_r; eauto.
   eapply H2; eauto. simpl in H4; auto.*)
   rewrite H2 in c.
   eapply  converts_strengthening_var; eauto.
     
   eapply wf_branches_None; eauto.
     
   eapply wf_branches_Some; eauto.
   eapply H0; eauto. simpl in H3; eauto.
   eapply H1; eauto. simpl in H3; eauto.
Qed.



Lemma typing_strengthening_eq_mutual:
forall S1 S2 ,
wf_sig S1 ->
((forall E0 e t,  wf_tm S1 S2 E0 e t ->
  forall E F x v k,  E0 = (E& x~ bnd_eq v v k & F) ->
   wf_env S1 S2 (E&F) ->
   wf_tm S1 S2 (E & F) e t)
/\ (forall E0 s args branches r_tp,  wf_branches S1 S2 E0 s args branches r_tp ->
             forall E F x v k,  E0 = (E& x~ bnd_eq v v k& F) -> wf_env S1 S2 (E&F) ->
            wf_branches S1 S2 (E&F) s args branches r_tp )).
Proof.
intros.
 apply  mutual_wf_tm_branches_ind; intros; auto. 

  Case "arr".
  gather_atoms' as L'.
  eapply wf_tm_arr with (L:=L'); eauto.
  intros. unfold L' in H4. 
  rewrite_env (E0 & (F& x0~:u1)). eapply H1 with (x0:=x) (x:=x0) (v:=v) (k:=k); eauto.  rewrite H2; auto.
  rewrite_env ((E0 & F) & (x0~:u1)).
  eapply wf_env_cons_var; eauto.

 Case "fv".
  rewrite H0 in b. binds_cases b. 
 apply wf_tm_var; auto.
  replace E0 with (E0 & empty) in B0; auto. rewrite_env ( E0 & F & empty). eapply binds_var_weakening; eauto.
  apply wf_tm_var; auto.
  rewrite_env (empty&E0&F). replace F with (empty&F) in B0; auto. eapply binds_var_weakening; eauto.
  rewrite_env (E0&F); eauto. apply empty_concat.

Case "abs".
  gather_atoms' as L'.
 eapply wf_tm_abs with (L:=L'); eauto.
 intros. unfold L' in H5. rewrite_env (E0 & (F&x0 ~:u1)). 
  eapply H1 with (x0:=x) (F:= F&x0~:u1) (k:=k0) (v:=v) ; eauto. 
  rewrite H3; eauto. 
  rewrite_env (E0 &F&x0~:u1); eapply wf_env_cons_var; eauto.

Case "app".
  eapply wf_tm_app with (t2:=t2); eauto.
Case "SAYS".
  eapply wf_tm_P_SAYS; eauto.
   eapply wf_tm_T_PF; eauto.
   eapply wf_tm_TM_PF_RETURN; eauto.
   eapply wf_tm_TM_PF_BIND; eauto.
   eapply wf_tm_TM_SAY; eauto.

Case "pif".
   gather_atoms' as L'.
   eapply wf_tm_TM_PIF with (L:=L'); eauto.
   intros. unfold L' in H7. rewrite_env (E0& (F& x0~bnd_eq v1 v2 T1)). 
   eapply H2 with (F:=F &x0~bnd_eq v1 v2 T1) (x:=x0) (v:=v3) (k:=k) (x0:=x); eauto.
   rewrite H5; auto.
   rewrite_env (E0 & F &x0~bnd_eq v1 v2 T1).
   apply wf_env_cons_eq; auto.
  
  eapply H0;eauto.
  eapply H1; eauto.
  eapply H4; eauto.

 eapply wf_tm_PF_SAYS_RETURN; eauto.
  eapply wf_tm_PF_SAYS_BIND; eauto.

 eapply wf_tm_mat_TYPE; eauto.
 eapply wf_tm_mat_PROP; eauto.
  eapply wf_tm_cast; eauto. rewrite H2 in c. eapply converts_strengthening_eq; eauto.
     rewrite H2 in w. assert (J:= wf_env_strong_inv_eq E0 x v v k F (wf_tm_gives_wf_env w)).
     destruct J as [J1 J2].
     eapply (proj1 (wf_tm_branches_lc_mutual S1 S2)); eauto.
 
  eapply wf_branches_Some; eauto.
Qed.
  
 

Lemma tm_ctr_value_subst_mutual :
  (forall v1, value v1 ->
     forall x v2,
     value v2 ->
     value ([x~>v2]v1)) /\
  (forall v1, applied_value v1 ->
     forall x v2,
     value v2 ->
     applied_value ([x~>v2]v1)).

Proof. 
  apply mutual_tm_ctr_value_ind;
      try solve[ intros; simpl*; auto].
    (* Case ("var case"). *)
      intros. simpl*. destruct (x==x0); auto.
Qed.


Lemma value_subst : forall x v1 v2,
  value v1 ->
  value v2 ->
  value ([x~>v2]v1).
Proof.
  intuition.
  assert (J:= tm_ctr_value_subst_mutual).
  destruct J. auto.
Qed.



Lemma ok_map_right: forall E x v F,
  ok (E & F) ->
  ok (E & subst_env x v F).
Proof.   induction F.
    intro. simpl in *. auto.
  intros.
  destruct a as [y b].
  simpl_env.
  rewrite_env (E & (subst_env x v F) & y ~ subst_bnd x v b).
 constructor 2 with  (E:= E& subst_env x v F) (x:=y) (a:= subst_bnd x v b).
apply IHF. inversion H. trivial.
inversion H. 
rewrite_env (concat E (subst_env x v F)).
rewrite dom_concat.
unfold subst_env. 
rewrite_env (map (subst_bnd x v) F). 
rewrite dom_map. rewrite dom_concat in H4. trivial.
Qed.

Open Scope tm_scope.
Lemma subst_converts: forall S1 S2 E F x v u s t k,
  wf_tm S1 S2 E v k ->
(*  value v ->*)
  wf_env S1 S2 (E & x ~: u & F) ->
  converts (E & x ~: u & F) s t ->
  converts (E & (subst_env x v F)) 
    ([x ~> v] s) ([x ~> v] t).
Proof.
intros.
remember (E & x ~: u & F) as G.
induction H1.
eapply conv_refl. apply subst_lc; auto. 
 eapply  (proj1 (wf_tm_branches_lc_mutual S1 S2)); eauto.
eapply conv_sym. eapply IHconverts; eauto.
eapply conv_trans with (t2 := [x ~> v] t2).
eapply IHconverts1; eauto.
eapply IHconverts2; eauto.

Case "conv_axiom".
 unfold binds_eq in H1. 
 rewrite HeqG in H1.
binds_cases H1.

assert (J1: x\notin fv t1 /\ x\notin fv t2 /\ x \notin fv k0).
replace (binds x0 (bnd_eq t1 t2 k0) E) with (binds_eq x0 t1 t2 k0 E) in B0; auto.
eapply binds_eq_fresh_ext with (e1:=t1) (e2:=t2); eauto.
rewrite HeqG in H0.
eapply wf_env_contraction with (F:= x~:u & F).
rewrite (concat_assoc E (x~:u) F)  in H0.   apply H0.
rewrite HeqG in H0.
eapply wf_env_fresh_l; eauto.  auto.
destruct J1 as [Ht1 [Ht2 Hk]].
rewrite subst_fresh. rewrite subst_fresh.
eapply conv_axiom with (k:=k0) (x:=x0); eauto.
(* rewrite_env (concat E (subst_env x v F)).*)
unfold binds_eq.
eapply binds_concat_ok; eauto. 
eapply ok_map_right.
eapply ok_remove with (F:= x~:u).
rewrite HeqG in H0. 
eapply wf_env_ok; eauto. auto. auto. auto.

assert (J:binds x0 (subst_bnd x v (bnd_eq t1 t2 k0)) (subst_env x v F)).
unfold subst_env.
apply binds_map with (x:=x0) (f:=subst_bnd x v) (E:=F) (a:= bnd_eq t1 t2 k0); auto.

simpl in J. eapply conv_axiom with (k:= [x~>v]k0) (x:=x0); eauto.
unfold binds_eq.
eapply binds_prepend with (E:=E) (F:= subst_env x v F); eauto.

simpl. eapply conv_app; eauto.

gather_atoms' as L'.
simpl. eapply conv_abs with (L:=L'); eauto. 
intros. assert (J:= H3 x0).
unfold L' in H4.
rewrite subst_open in J; auto. rewrite subst_open in J; auto.
rewrite subst_fresh in J; auto. 
simpl; auto.
eapply(proj1 (wf_tm_branches_lc_mutual S1 S2)); eauto.
eapply(proj1 (wf_tm_branches_lc_mutual S1 S2)); eauto.

simpl.
gather_atoms' as L'.
simpl. eapply conv_arr with (L:=L'); eauto. 
intros. assert (J:= H3 x0).
unfold L' in H4.
rewrite subst_open in J; auto. rewrite subst_open in J; auto.
rewrite subst_fresh in J; auto. 
simpl; auto.
eapply(proj1 (wf_tm_branches_lc_mutual S1 S2)); eauto.
eapply(proj1 (wf_tm_branches_lc_mutual S1 S2)); eauto.
(*
simpl. eapply conv_mat; eauto.
simpl. eapply conv_brn; eauto.
simpl. eapply conv_cast; eauto. *)
Qed.


Lemma subst_fully_applied_args: forall k c args z v,
  fully_applied_args k c args ->
  fully_applied_args (subst_tm z v k) c (List.map (subst_tm z v) args).
Proof.
intros.
induction H; try solve[simpl; auto].
Qed.

Lemma subst_fully_applied_typ: forall k c u z v,
  fully_applied_typ k c u ->
  fully_applied_typ (subst_tm z v k) c ([z~>v] u).
Proof.
intros.
induction H; try solve[simpl; auto].
Qed.

Lemma subst_branches_cover_aux: forall l branches z v,
 branches_cover_aux l branches -> 
 branches_cover_aux l ([z~> v] branches).
Proof.
intros. induction H; try solve[simpl; auto].
Qed.

Lemma subst_branches_cover: forall S branches ctr z v,
 branches_cover S branches ctr -> 
 branches_cover S ([z~> v] branches) ctr.
Proof. 
intros.
unfold branches_cover.
unfold branches_cover in H.
destruct H as [l [Hctr Hcover]].
exists l; split; auto.
eapply subst_branches_cover_aux; eauto.
Qed.

Lemma substitution : forall S1 S2,
  wf_sig S1 ->
  (forall E0 e t, wf_tm S1 S2 E0 e t->
     forall E F z v u,
       E0 = E & z ~: u & F ->
       wf_tm S1 S2 E v u ->
       value v ->
       wf_tm S1 S2 (E & (subst_env z v F)) ([z~>v]e) ([z~>v]t))
/\
  (forall E0 t_e args t_b t_res, wf_branches S1 S2 E0 t_e args t_b t_res ->
     forall E F z v u,
       E0 = E & z ~: u & F ->
       wf_tm S1 S2 E v u ->
       value v ->
       wf_branches S1 S2 (E & (subst_env z v F))
                 ([z~>v]t_e) (List.map (subst_tm z v) args) ([z~>v]t_b) ([z~>v]t_res))
/\
  (forall s args t_ctr t_b t_res, wf_brn S1 S2 s args t_ctr t_b t_res -> 
     forall z v ,
(*       wf_tm S E v u ->*)
	lc v ->
       value v ->
       wf_brn S1 S2 ([z~>v]s) (List.map (subst_tm z v) args) ([z~>v]t_ctr) ([z~>v]t_b) ([z~>v]t_res))
/\
  (forall E0, wf_env S1 S2 E0 ->
     forall E F z v u,
       E0 = E & z ~: u & F ->
       wf_tm S1 S2 E v u ->
       value v ->        
       wf_env S1 S2 (E & (subst_env z v F)))
 .
Proof with auto*.
introv Swf.
apply mutual_wf_tm_branches_brn_env_ind; eauto.

Case "wf_tm_arr".
intros; simpl.
gather_atoms' as L' .
 eapply wf_tm_arr with (k1:=[z~>v]k1) (L:=L').
eapply H with (u:=u); auto.
intros.
rewrite (subst_env_commute_push' v z E0 F x u1).
rewrite subst_open_commute.

eapply H0 with (x:=x) (E0:=E0) (F:= F&x~:u1) (z:=z) (v:=v) (u:= u). 
 unfold L' in H4. auto. rewrite H1. auto.
auto. auto. unfold L' in H1.
apply notin_singleton_r.  apply notin_singleton_swap. unfold L' in H4. auto. 
 eapply (proj1 (wf_tm_branches_lc_mutual S1 S2)); eauto. 
destruct o; subst. left; auto.
destruct H4; subst. right. left; auto.
right. right; auto.

Case "wf_tm_ctr".
intros. simpl. constructor.
eapply H. eapply H0. auto. auto.
rewrite subst_fresh. auto. 
eapply sig_has_ctr_fv with (c:=c); eauto.

Case "wf_tm_assn".
intros. simpl. eapply wf_tm_assn; eauto. 
rewrite subst_fresh. auto.
eapply sig_has_assn_fv; eauto.

Case "wf_tm_var".
intros. simpl. 
destruct (x==z).
subst. unfold binds_var in b.
assert(J: bnd_var t = bnd_var u).
eapply binds_mid_eq with (z:= z) (E:=E0) (b:= bnd_var u) (F:=F).
rewrite_env(Env.concat(E0 & z~:u) F).
auto.
eapply wf_env_ok with (S1:= S1) (S2 :=S2); eauto.

inversion J. subst.
rewrite subst_fresh.
apply typing_weakening'; auto.
eapply H with (z0:=z) (u0:=u); auto.
eapply binds_var_fresh with (E:=E0 & z~:u & F); auto.

Case "x<>z".
assert(Jwf: wf_env S1 S2 E0).
eapply wf_tm_gives_wf_env; eapply H1. eapply w.

subst. binds_cases b.
 Case "x in E0". 
rewrite subst_fresh. 
eapply typing_weakening' with (F:= subst_env z v F); eauto.
eapply wf_tm_var; auto.

eapply wf_tm_gives_wf_env; eapply H1; eauto. 
eapply binds_var_fresh_ext with (E:=E0) (x:=x) (S1:=S1) (S2:=S2); eauto.
eapply wf_tm_gives_wf_env; eapply H1; eauto. 
assert (J:wf_env S1 S2 (E0 & z~:u)). eapply wf_env_contraction; eauto.
inversion J; eauto.
 
constructor.
eapply H with (z0:=z) (F0:=F); auto.
unfold binds_var.
eapply binds_prepend.
unfold subst_env.
change (bnd_var ([z ~> v] t))
          with (subst_bnd z v (bnd_var t)).
 eapply binds_map; auto.

Case "wf_tm_abs".
intros; simpl. 
gather_atoms' as L'.
eapply wf_tm_abs with (L:=L') (k :=[z~>v] k) (k1:= [z~>v] k1).
eapply H. eapply H2. auto. auto.
intros. 
rewrite (subst_env_commute_push' v z E0 F x u1).
rewrite subst_open_commute.
rewrite subst_open_commute.

eapply (H0 x) with (E0:=E0) (F:= F & (x~:u1)) (z:=z)(v:=v) (u:= u).
unfold L' in H5. auto.
rewrite H2; auto. auto. auto. 
unfold L' in H5. auto.
destruct (wf_tm_branches_lc_mutual S1 S2).
apply (H6 E0 v u); auto.
unfold L' in H5; auto.
destruct (wf_tm_branches_lc_mutual S1 S2).
apply (H6 E0 v u); auto.

simpl in H1. apply (H1 E0 F z v u).
rewrite cons_concat.
rewrite H2. auto. 
auto.  auto. 
destruct o. 
left; rewrite H5; auto.
right; rewrite H5; auto.
repeat(destruct o0).
left; rewrite H5; auto.
destruct H5.
right; left; rewrite H5; auto.
right; right; rewrite H5; auto.

Case "wf_tm_app".
intros; simpl. rewrite subst_open.
eapply wf_tm_app with (t2 := [z~>v] t2).

simpl in H.  eapply H with (v:=v) (u0:=u0) (E0:=E0) (F:=F); auto.
(* eapply H1 with (v:=v) (u0:=u0) (E0:=E0) (F:=F); auto.  *)
eapply H0; eauto.
(* eapply H1; eauto.
rewrite <- subst_open.
eapply H2; eauto.
eapply(proj1 (wf_tm_branches_lc_mutual S)); eauto.
intros.  *)
destruct o.  left.  eapply value_subst; eauto.
right.  apply subst_lc; auto.
eapply(proj1 (wf_tm_branches_lc_mutual S1 S2)); eauto.
eapply(proj1 (wf_tm_branches_lc_mutual S1 S2)); eauto.

Case "wf_tm_P_SAYS".
intros;simpl. eapply wf_tm_P_SAYS.
simpl in H.
eapply H; eauto.
rewrite cons_concat; rewrite H1; auto.
simpl in H0.
eapply H0; eauto.
rewrite cons_concat; rewrite H1; auto.
apply value_subst; auto.

Case "wf_tm_PF_SIGN".

intros; simpl. apply wf_tm_PF_SIGN. 
rewrite subst_fresh; auto. 
eapply wf_tm_fresh_l with (E:= empty:env); eauto.
eapply value_subst; eauto.
rewrite subst_fresh; auto. 
eapply wf_tm_fresh_l with (E:= empty:env); eauto.
eapply H1; eauto.

Case "wf_tm_T_PF".
intros;simpl. apply wf_tm_T_PF. 
simpl in H.
eapply H; eauto.
rewrite cons_concat; rewrite H0; auto.

Case "wf_tm_TM_PF_RETURN".
intros;simpl. apply wf_tm_TM_PF_RETURN. 
eapply H; eauto.
simpl in H0. eapply H0; eauto.
rewrite cons_concat; rewrite H1; auto.

Case "wf_tm_TM_PF_BIND".
intros;simpl. eapply wf_tm_TM_PF_BIND.
simpl in H. eapply H; eauto.
rewrite cons_concat; rewrite H1; auto.
simpl in H0. eapply H0; eauto.
rewrite cons_concat; rewrite H1; auto.

apply subst_lc.
destruct (wf_tm_branches_lc_mutual S1 S2).
apply (H4 E0 v u); auto. auto.


Case "wf_tm_TM_SAY".
intros;simpl. eapply wf_tm_TM_SAY.
simpl in H. eapply H; eauto.
rewrite cons_concat; rewrite H0; auto.


Case "wf_tm_TM_PIF".
intros;simpl.
gather_atoms' as L'.
 eapply wf_tm_TM_PIF with (L:=L').
simpl in H. eapply H; eauto.
rewrite cons_concat; rewrite H4; auto.
simpl in H0. eapply H0; eauto.
rewrite cons_concat; rewrite H4; auto.
rewrite subst_fresh; auto.
eapply atomic_type_fresh with (S:=S2); eauto.  
apply value_subst; auto.
apply value_subst; auto.
intros; simpl.
rewrite cons_concat.
change (single x (bnd_eq ([z~>v3]v1) ([z~>v3]v2) T1)) with
 (subst_env z v3 (x~ (bnd_eq v1 v2 T1))).
rewrite_env (E0 & subst_env z v3( F&(x~bnd_eq v1 v2 T1))).
eapply H1 with (v:=v3) (u:=u) (x:=x); eauto.

unfold L' in H7; auto.
rewrite H4; auto.
eapply H2; eauto.
simpl in H3. eapply H3; eauto. rewrite cons_concat; rewrite H4; auto.

Case " wf_tm_PF_SAYS_RETURN ".
intros; simpl.
eapply wf_tm_PF_SAYS_RETURN; eauto.
simpl in H0.
eapply H0; eauto.
rewrite cons_concat. rewrite H2; auto.
simpl in H1; eapply H1; eauto.
rewrite cons_concat. rewrite H2; auto.
eapply value_subst; eauto.

Case "wf_tm_PF_SAYS_BIND".
intros; simpl.
eapply  wf_tm_PF_SAYS_BIND with (P:= [z~>v]P); eauto.
simpl in H.
eapply H; eauto.
rewrite cons_concat; rewrite H1; eauto.
simpl in H0; eapply H0; eauto.
rewrite cons_concat; rewrite H1; eauto.
eapply subst_lc; auto.
destruct ( wf_tm_branches_lc_mutual S1 S2).
apply (H4 E0 v u); auto.

Case " wf_tm_mat_TYPE".
intros; simpl.
eapply wf_tm_mat_TYPE with (ctr:=ctr0) (args:=List.map (subst_tm z v) args) (k:=[z~>v]k) (s:=[z~>v]s); eauto.
simpl in H0; eapply H0; eauto.
rewrite cons_concat; rewrite H3; auto.
eapply subst_fully_applied_args; eauto.
rewrite subst_fresh; auto.
eapply sig_has_ctr_fv; eauto.
eapply subst_fully_applied_typ; eauto.

simpl in H1; eapply H1; eauto.
rewrite cons_concat; rewrite H3; eauto.

rewrite <- map_rev.
eapply H2; eauto.

eapply subst_branches_cover; eauto.

Case " wf_tm_mat_PROP".
intros; simpl.
eapply wf_tm_mat_PROP with (ctr:=ctr0) (args:=List.map (subst_tm z v) args) (k:=[z~>v]k); eauto.
simpl in H0; eapply H0; eauto.
rewrite cons_concat; rewrite H3; auto.
eapply subst_fully_applied_args; eauto.
rewrite subst_fresh; auto.
eapply sig_has_ctr_fv; eauto.
eapply subst_fully_applied_typ; eauto.


simpl in H1; eapply H1; eauto.
rewrite cons_concat; rewrite H3; eauto.

rewrite <- map_rev.
eapply H2; eauto.

eapply subst_branches_cover; eauto.

Case "wf_tm_cast".
intros; simpl.
eapply wf_tm_cast; eauto.
replace K_TYPE with ([z~>v]K_TYPE) ; auto.
eapply H0; eauto.
(* replace K_TYPE with ([z~>v]K_TYPE) ; auto.
eapply H1; eauto.*)
eapply subst_converts with (u:=u).
eapply H2. 
rewrite <- H1; auto.
eapply wf_tm_gives_wf_env; eapply w.
rewrite <- H1; auto.

Case "wf_branches_Some".
intros. simpl.
eapply wf_branches_Some with (ectr_typ:=[z~>v] ectr_typ); eauto. 
rewrite subst_fresh; auto.
eapply sig_has_ctr_fv; eauto.
eapply H1; eauto.
 eapply (proj1 (wf_tm_branches_lc_mutual S1 S2)); eauto. 

Case "wf_brn_constructor_arg".
intros. simpl.
gather_atoms' as L'. 
eapply wf_brn_constructor_arg with (L:=L') (result_typ:=[z~>v]result_typ); eauto.
intros. unfold L' in H2. 
simpl in H.
rewrite subst_open_commute; auto.
rewrite subst_open_commute; auto.

Case "wf_brn_datatype_param".
intros. simpl.
eapply wf_brn_datatype_param; eauto.
rewrite <- subst_open; auto.

Case "wf_env_nil".
intros. 
assert (J:= eq_empty_inv H).
contradiction.

Case "wf_env_cons_var".
intros.
destruct F. simpl in H1.
inversion H1. subst. simpl. auto.
inversion H1. destruct p as [y s]. inversion H5.
        destruct s; try congruence.
        inversion H8. subst. simpl_env.
eapply wf_env_cons_var with (E:= E0& subst_env z v F) (x:=y) (t:= [z~>v]t0); eauto.

rewrite dom_concat. unfold subst_env. 
rewrite dom_map; auto. 


Case "wf_env_cons_eq".
intros.
destruct F. simpl in H3.
inversion H3. 
inversion H3. destruct p as [y s]. inversion H7.
        destruct s; try congruence.
        inversion H10. subst. simpl_env. simpl.
eapply wf_env_cons_eq with (E:= E0& subst_env z v1 F) (x:=y) (e1:=[z~>v1]t0) (e2:=[z~>v1]t1) (t:= [z~>v1]t2); eauto.
rewrite subst_fresh; auto.
eapply atomic_type_fresh; eauto. 
rewrite dom_concat.  unfold subst_env.
rewrite dom_map; auto.
apply value_subst; auto.  apply value_subst; auto.  
simpl in H2; eapply H2; eauto.
Qed.


Corollary simple_tm_substitution:
 forall S1 S2 e t E z v u,
   wf_sig S1 -> 
   wf_tm S1 S2 (E & z ~: u)  e t ->
   wf_tm S1 S2 E v u ->
   value v ->
   wf_tm S1 S2 E  ([z~>v]e) ([z~>v]t).
Proof.
introv Hwf_sig  Hwf_tm Hwf_v Hvalue.
assert (J:= substitution S2 Hwf_sig).
destruct J as [J _].
  assert(J':=J (E & z ~: u) e t Hwf_tm E empty).
  simpl in J'.
eapply J'; eauto.
Qed.
 
Lemma defns_disjoint_weakening: forall d S, 
  disjoint (dom_defns d) (dom_sig S) -> 
  disjoint (dom_defns (get_tctr_defns_of d)) (dom_sig S).
 Proof.
 introv H.
 induction d; auto.
 destruct a; auto.
 destruct  p. simpl.
 unfold disjoint; intros.
 destruct (in_inv H0); subst; auto.
 apply IHd; auto. 
 unfold disjoint; intros; apply H; auto.
 unfold dom_defns; simpl.
 right ; apply in_or_app; auto.
Qed.


Lemma sig_has_ctr_weakening_sig_r: forall S1 S2 c t, 
 sig_has_ctr c t S2 -> disjoint (dom_sig S1)  (dom_sig S2) -> sig_has_ctr c t (S1 ++S2).
Proof. 
 intros.
 induction S1.
 Case "nil".
 replace (nil++S2) with S2; auto.
 Case "cons".
 destruct a.
   Case "bundle_defns".
   rewrite <- app_comm_cons.
   unfold sig_has_ctr. unfold sig_lookup_ctr.
   remember (bundle_defns_ctr_lookup c d) as B.
   destruct B.
     Case "B = Some t".
     symmetry in HeqB. 
     assert (J:= defns_lookup_some_in c d HeqB).
     assert (J1:= sig_lookup_ctr_some_in c S2 H).
     unfold disjoint in H0; simpl.
     assert (J2: ~ In c (dom_sig S2)).
     eapply H0. unfold dom_sig;  simpl. apply in_or_app. left; auto. contradiction.
 
     Case "B=None".
     fold sig_lookup_ctr. apply IHS1.
     unfold dom_sig in H0. simpl in H0.
     unfold disjoint; intros.
     unfold disjoint in H0.
     apply H0.  eapply in_or_app. right; auto.   

    Case "bundle_assn".
    rewrite <- app_comm_cons.
    unfold sig_has_ctr. unfold sig_lookup_ctr. apply IHS1. 
    unfold dom_sig in H0. simpl in H0.
    unfold disjoint; intros.
    unfold disjoint in H0.
    apply H0.  eapply in_cons; auto.   
Qed.


Lemma sig_has_assn_weakening_sig_r: forall S1 S2 c t, 
 sig_has_assn c t S2 -> disjoint (dom_sig S1)  (dom_sig S2) -> sig_has_assn c t (S1 ++S2).
Proof. 
 intros.
 induction S1.
 Case "nil".
 replace (nil++S2) with S2; auto.
 Case "cons".
 destruct a.
   Case "bundle_defns".
   rewrite <- app_comm_cons.
   unfold sig_has_assn. unfold sig_lookup_assn. 
   apply IHS1. 
    unfold dom_sig in H0. simpl in H0.
    unfold disjoint; intros.
    unfold disjoint in H0.
    apply H0.  eapply in_or_app; auto.   

   Case "bundle_assn".
   rewrite <- app_comm_cons.
   unfold sig_has_assn. unfold sig_lookup_assn.
   destruct (eq_ctr_dec c c0).
   subst.  assert (J1:= sig_lookup_assn_some_in c0 S2 H).
   unfold disjoint in H0; simpl.
   assert (J2: ~ In c0 (dom_sig S2)) .
   eapply H0. unfold dom_sig;  simpl. left; auto.  contradiction.
 
  fold sig_lookup_assn. 
  apply IHS1.
  unfold dom_sig in H0. simpl in H0.
  unfold disjoint; intros.
  unfold disjoint in H0.
  apply H0.  eapply in_cons; auto. 
Qed.

(*
Lemma sig_has_ctr_weakening_sig_r: forall d S2 c t, 
 sig_has_ctr c t S2 -> disjoint (dom_defns d)  (dom_sig S2) -> sig_has_ctr c t (bundle_defns d ::S2).
Proof. 
 intros.
   unfold sig_has_ctr. unfold sig_lookup_ctr.
   remember (bundle_defns_ctr_lookup c d) as B.
   destruct B.
     Case "B = Some t".
     symmetry in HeqB. 
     assert (J:= defns_lookup_some_in c d HeqB).
     assert (J1:= sig_lookup_ctr_some_in c S2 H).
     unfold disjoint in H0; unfold not in H0.
     destruct (H0 c J J1).

     Case "B=None".
     fold sig_lookup_ctr.  auto. 
Qed.


Lemma sig_has_assn_weakening_sig_r: forall d S2 c t, 
 sig_has_assn c t S2 -> disjoint (dom_defns d)  (dom_sig S2) -> sig_has_assn c t (bundle_defns d::S2).
Proof. 
 intros.
   unfold sig_has_assn. unfold sig_lookup_assn. simpl.
   auto.
Qed.


Lemma get_ctrs_weakening_sig_r: forall d S2 c t, 
 get_ctrs_of c S2 = Some t -> disjoint (dom_defns d)  (dom_sig S2) -> 
  get_ctrs_of c (bundle_defns d ::S2) = Some t  .
Proof. 
intros. 
   unfold get_ctrs_of. 
   remember (get_defns_ctrs_of c d) as G.
   destruct G.
   Case "G=Some".
   symmetry in HeqG.
   assert (J:= get_defns_ctr_some_in c d HeqG).
   assert (J2:= get_ctr_sig_in c S2 H).
   unfold disjoint in H0; unfold not in H0; destruct (H0 c J J2).
   Case "G=None".
    auto.
Qed.    
*)


Lemma get_ctrs_weakening_sig_r: forall S1 S2 c t, 
 get_ctrs_of c S2 = Some t -> disjoint (dom_sig S1)  (dom_sig S2) -> 
  get_ctrs_of c (S1 ++S2) = Some t  .
Proof. 
intros. 
 induction S1.
 Case "nil".   replace (nil++S2) with S2; auto.
 Case "cons". 
   rewrite <- app_comm_cons.
   unfold get_ctrs_of. 
   destruct a.
   remember (get_defns_ctrs_of c d) as G.
   destruct G.
   Case "G=Some".
   symmetry in HeqG.
   assert (J:= get_defns_ctr_some_in c d HeqG).
   assert (J2:= get_ctr_sig_in c S2 H).
   assert (J3: ~In c (dom_sig S2)).
   eapply H0; eauto. unfold dom_sig. simpl.
   apply in_or_app. left; auto. contradiction.
   Case "G=None".
   apply IHS1. 
   unfold dom_sig in H0. simpl in H0.
   unfold disjoint; intros.
   unfold disjoint in H0.
   apply H0.  eapply in_or_app; auto.   
 
  fold get_ctrs_of.
   apply IHS1. 
   unfold dom_sig in H0. simpl in H0.
   unfold disjoint; intros.
   unfold disjoint in H0.
   apply H0.  eapply in_cons; auto.
Qed.    


Lemma get_atomic_ctrs_sig_weakening:  forall S1 S2 L L1,
  get_atomic_ctrs_sig S2 L = L1 ->
  disjoint (dom_sig S1)  (dom_sig S2) -> 
  exists L2,   get_atomic_ctrs_sig (S1++S2) L = L1++L2.
Proof.
 introv Hget Hdisj.
 induction S1.
 exists (nil: list ctr); auto. replace (nil++S2) with S2; auto.
 rewrite <- app_nil_end; auto.

destruct a.
Case "bundle_defns".
  destruct IHS1; auto.
  unfold dom_sig in Hdisj; simpl in Hdisj.
  unfold disjoint; unfold not; intros. 
  unfold disjoint in Hdisj; unfold not in Hdisj; apply (Hdisj n); auto.
  eapply in_or_app; eauto.
  remember (all_atomic_decls_defns  ((dom_defns(get_tctr_defns_of d))++ L1++x) d) as A.
  rewrite <- app_comm_cons. 
  destruct A. 
  exists (x++(dom_defns(get_tctr_defns_of d))); auto.
  unfold get_atomic_ctrs_sig. fold get_atomic_ctrs_sig.
  rewrite H. rewrite <- HeqA. rewrite app_ass; auto.
  
 exists x. unfold get_atomic_ctrs_sig.
 fold get_atomic_ctrs_sig. rewrite H. rewrite <- HeqA; auto.
  
Case "bundle_assn".
 destruct IHS1; auto.
 unfold dom_sig in Hdisj; simpl in Hdisj.
 unfold disjoint; unfold not; intros.
 unfold disjoint in Hdisj; unfold not in Hdisj; apply (Hdisj n); auto.
 exists x; simpl; auto.
Qed.


Lemma simple_type_weakening:  forall L L1 t,
  simple_type L t ->
  simple_type (L++L1) t.
Proof. 
 introv J.
 unfold simple_type. unfold simple_type in J.
 induction t; simpl in J; simpl; auto.
 destruct (In_dec eq_ctr_dec c L); inversion J.
 destruct (In_dec eq_ctr_dec c (L++L1)); auto.
 destruct n. eapply in_or_app; eauto.
 Qed.
 

Lemma atomic_types_weakening_sig_r: forall S1 S2 t, 
           atomic_type S2 t ->
           disjoint (dom_sig S1)  (dom_sig S2) -> 
           atomic_type (S1++S2)  t.
Proof.
  introv Hfo Hdisj.
  unfold atomic_type. unfold atomic_type in Hfo.
  destruct Hfo as [L [Jctr Jsimpl]].
  assert (J: exists L1, get_atomic_ctrs_sig (S1++S2) nil = L ++L1). 
  eapply get_atomic_ctrs_sig_weakening; eauto.
  destruct J as [L1 Jctr1].
  exists (L++L1); split; auto.
  eapply simple_type_weakening; eauto.
Qed.

Definition sig_extend (S1:sig) (S2:sig) : Prop :=
  (S1 = S2) \/ exists d, exists S, S1 = (bundle_defns(get_tctr_defns_of d)::S) 
                                                    /\ S2 = (bundle_defns d::S) 
                                                    /\ disjoint (dom_defns d) (dom_sig S).

Lemma sig_extend_disjoint: forall S1 S2 S,
 sig_extend S1 S2 ->
 disjoint (dom_sig S) (dom_sig S2) ->
 disjoint (dom_sig S) (dom_sig S1).
Proof. 
 introv Hext Hdisj.
 inversion Hext; subst; auto.

 destruct H as [d [S3 [HS1 [HS2 IH3]]]]. subst.
 assert (J: disjoint (dom_sig S) (dom_sig S3)).
 unfold disjoint; unfold not; intros.
 unfold dom_sig in Hdisj; simpl in Hdisj. 
 apply (Hdisj n); auto. apply in_or_app; auto. 
 assert (J2: disjoint (dom_defns (get_tctr_defns_of d)) (dom_sig S)).
 apply defns_disjoint_weakening; auto. 
 unfold disjoint; unfold not; intros. 
  unfold dom_sig in Hdisj; simpl in Hdisj.  
 apply (Hdisj n); auto. apply in_or_app; auto.

 unfold dom_sig; simpl. replace (flat_map dom_bundle S3) with (dom_sig S3); auto.
 unfold disjoint; unfold not; intros.
 destruct (in_app_or (dom_defns(get_tctr_defns_of d) )(dom_sig S3) n H0).
 unfold disjoint in J2; unfold not in J2; apply (J2 n); auto.
 unfold disjoint in J; apply (J n); auto.

Qed.  

 
 
Lemma weakening_sig_r_mutual: forall S,
  wf_sig S ->
  (forall E e t, wf_tm S S E e t->
     forall d,
      disjoint (dom_defns d)  (dom_sig S)  ->
       wf_tm (bundle_defns(get_tctr_defns_of d)::S) (bundle_defns d::S) E e t )
/\
  (forall E t_e args t_b t_res, wf_branches S S E t_e args t_b t_res ->
      forall d,
      disjoint (dom_defns d)  (dom_sig S)  ->
      wf_branches  (bundle_defns(get_tctr_defns_of d)::S) (bundle_defns d::S)  E t_e args t_b t_res )
/\
  (forall s args t_ctr t_b t_res, wf_brn S S s args t_ctr t_b t_res -> 
     forall d,
      disjoint (dom_defns d)  (dom_sig S)  ->
      wf_brn  (bundle_defns(get_tctr_defns_of d)::S) (bundle_defns d::S)  s args t_ctr t_b t_res )
/\
  (forall E, wf_env S S E ->
     forall d,
      disjoint (dom_defns d)  (dom_sig S)  ->
      wf_env  (bundle_defns(get_tctr_defns_of d)::S) (bundle_defns d::S)  E)
 .
Proof. 

introv Hwf_sig.
apply mutual_wf_tm_branches_brn_env_ind; intros ; eauto. 

Case "ctr".

apply wf_tm_ctr; auto.
 replace (bundle_defns(get_tctr_defns_of d)::S) with ((bundle_defns(get_tctr_defns_of d)::nil)++S); auto.
  eapply sig_has_ctr_weakening_sig_r; eauto.
  unfold dom_sig; simpl. 
 rewrite <- app_nil_end. replace (flat_map dom_bundle S) with  (dom_sig S); auto.
  eapply defns_disjoint_weakening; eauto.

Case "pif".
eapply wf_tm_TM_PIF; eauto.
 replace (bundle_defns d::S) with ((bundle_defns d::nil)++S); auto.
eapply (atomic_types_weakening_sig_r); auto. 
  unfold dom_sig; simpl. 
 rewrite <- app_nil_end. replace (flat_map dom_bundle S) with  (dom_sig S); auto.

Case "tm_mat".
intros. eapply wf_tm_mat_TYPE with (k:= k); eauto.
 replace (bundle_defns(get_tctr_defns_of d)::S) with ((bundle_defns(get_tctr_defns_of d)::nil)++S); auto.
 apply sig_has_ctr_weakening_sig_r; auto.
   unfold dom_sig; simpl. 
   replace (flat_map dom_bundle S) with  (dom_sig S); auto.
   rewrite <- app_nil_end;  eapply defns_disjoint_weakening; eauto.
   destruct b as [l [Jctr Jaux]].
   unfold branches_cover; exists l; split; auto.
    replace (bundle_defns d::S) with ((bundle_defns d::nil)++S); auto.
   apply get_ctrs_weakening_sig_r; auto. 
     unfold dom_sig; simpl. 
   replace (flat_map dom_bundle S) with  (dom_sig S); auto.
   rewrite <- app_nil_end;  auto. 
 
Case "tm_mat".
intros. eapply wf_tm_mat_PROP with (k:= k); eauto.
 replace (bundle_defns(get_tctr_defns_of d)::S) with ((bundle_defns(get_tctr_defns_of d)::nil)++S); auto.
 apply sig_has_ctr_weakening_sig_r; auto.
   unfold dom_sig; simpl. 
   replace (flat_map dom_bundle S) with  (dom_sig S); auto.
   rewrite <- app_nil_end;  eapply defns_disjoint_weakening; eauto.
   destruct b as [l [Jctr Jaux]].
   unfold branches_cover; exists l; split; auto.
    replace (bundle_defns d::S) with ((bundle_defns d::nil)++S); auto.
   apply get_ctrs_weakening_sig_r; auto. 
     unfold dom_sig; simpl. 
   replace (flat_map dom_bundle S) with  (dom_sig S); auto.
   rewrite <- app_nil_end;  auto. 

Case "branches". 
  eapply wf_branches_Some; eauto.
   replace (bundle_defns(get_tctr_defns_of d)::S) with ((bundle_defns(get_tctr_defns_of d)::nil)++S); auto.
  eapply sig_has_ctr_weakening_sig_r; eauto.
  unfold dom_sig; simpl. 
 rewrite <- app_nil_end. replace (flat_map dom_bundle S) with  (dom_sig S); auto.
  eapply defns_disjoint_weakening; eauto.
 
Case "wf_env".
 eapply wf_env_cons_eq; eauto. 

 replace (bundle_defns d::S) with ((bundle_defns d::nil)++S); auto.
eapply (atomic_types_weakening_sig_r); auto. 
  unfold dom_sig; simpl. 
 rewrite <- app_nil_end. replace (flat_map dom_bundle S) with  (dom_sig S); auto.
Qed. 

Lemma wf_tm_weakening_sig_r: 
forall S E e t d, 
  wf_sig S ->
  wf_tm S S E e t->
      disjoint (dom_defns d)  (dom_sig S)  ->
       wf_tm (bundle_defns(get_tctr_defns_of d)::S) (bundle_defns d::S) E e t.
Proof. intros.
 Check weakening_sig_r_mutual.
apply (proj1(weakening_sig_r_mutual H )); auto.
Qed.

Lemma in_tctr_defns_in: forall c d, In c (dom_defns (get_tctr_defns_of d)) -> In c (dom_defns d).
Proof.
 intros.
 induction d; auto.
 destruct a. destruct p.
 simpl in H.
 destruct H.
 subst; auto.
 unfold dom_defns; simpl.
 right. fold dom_defns. eapply in_or_app; eauto.
Qed.

Inductive wf_dom_defns: defns -> Prop:=
 | wf_dom_defns_nil: wf_dom_defns nil
 | wf_dom_defns_cons:  forall c k cdl d,
                         wf_dom_defns d ->
                         disjoint (dom_ctr_decls cdl) (dom_defns d) ->
                         wf_dom_defns ((c,k,cdl)::d).

Inductive wf_dom_sig: sig -> Prop:=
 | wf_dom_sig_nil: wf_dom_sig nil
 | wf_dom_sig_cons_assn: forall c k S,
                      wf_dom_sig S ->
                      ~ In c (dom_sig S) ->
                      wf_dom_sig (bundle_assert c k::S)
 | wf_dom_sig_cons_defns: forall d S,
                      wf_dom_defns d ->
                      wf_dom_sig S ->
                      disjoint (dom_defns d) (dom_sig S) ->
                      wf_dom_sig (bundle_defns d::S).
                         
Hint Constructors wf_dom_defns wf_dom_sig.

Lemma wf_bundle_ctrs_gives_wf_dom_defns: forall P S1 S2 d,
  wf_bundle_ctrs P S1 S2 d ->
 wf_dom_defns d.
Proof.
 intros.
 induction d; auto.
 inversion H; subst.
 apply wf_dom_defns_cons; auto.
Qed.

Lemma wf_sig_gives_wf_dom_sig: forall S,
  wf_sig S -> wf_dom_sig S.
Proof.
 intros.
 induction S; auto.
 destruct a.
 inversion H; subst; apply wf_dom_sig_cons_defns; auto; eapply wf_bundle_ctrs_gives_wf_dom_defns; eauto.
 inversion H; auto.
Qed.
 
Lemma defns_has_ctr_weakening: forall d t c,
 wf_dom_defns d ->
 bundle_defns_ctr_lookup c (get_tctr_defns_of d) = Some t ->
 bundle_defns_ctr_lookup c d = Some t .
Proof.
introv Hwf Hb.
 induction d; auto.
 destruct a. destruct p.
 inversion Hwf; subst.
 simpl in Hb.  unfold bundle_defns_ctr_lookup. 
 destruct (eq_ctr_dec c c1).
 Case "c = c1".
 inversion Hb; subst; auto.
 Case "c<>c1".
 remember (ctr_decls_lookup c c0) as C.
 destruct C.  symmetry in HeqC.
 assert (Jin1: In c (dom_ctr_decls c0)). eapply ctr_decls_lookup_some_in; eauto.
 assert (Jin2: In c (dom_defns d)). apply in_tctr_defns_in; auto.
 eapply  defns_lookup_some_in; eauto.
 destruct (H4 c Jin1 Jin2).
 Case "C = None".
 fold bundle_defns_ctr_lookup.
 eapply IHd; eauto.
Qed.


Lemma sig_has_ctr_weakening_sig1_r: forall  d S2 c t, 
 wf_dom_defns d ->
 disjoint (dom_defns d) (dom_sig S2) -> 
 sig_has_ctr c t (bundle_defns (get_tctr_defns_of d)::S2)  -> sig_has_ctr c t (bundle_defns d ::S2).
Proof.  
 intros.
   unfold sig_has_ctr in H1. unfold sig_lookup_ctr in H1. 
   remember (bundle_defns_ctr_lookup c (get_tctr_defns_of d)) as B.
   destruct B.
     Case "B = Some t".
     symmetry in HeqB.
     inversion H1; subst.
     unfold sig_has_ctr. unfold sig_lookup_ctr.
     assert (J:= defns_has_ctr_weakening c H HeqB).
     remember (bundle_defns_ctr_lookup c d) as B1.
     destruct B1; auto. inversion J.
     
     fold sig_lookup_ctr in H1. symmetry in HeqB.
     
     unfold sig_has_ctr. unfold sig_lookup_ctr.
     remember (bundle_defns_ctr_lookup c d) as B1.
     destruct B1; auto.
     symmetry in HeqB1.
    assert (Jin1:= defns_lookup_some_in c d HeqB1).
    assert (Jin2:= sig_lookup_ctr_some_in c S2 H1).
    destruct (H0 c Jin1 Jin2).
Qed.

Lemma sig_has_assn_weakening_sig1_r: forall d S2 c t, 
 disjoint (dom_defns d) (dom_sig S2) -> 
 sig_has_assn c t (bundle_defns (get_tctr_defns_of d)::S2)  -> 
 sig_has_assn c t (bundle_defns d ::S2).
Proof. 
 intros.
   unfold sig_has_assn. unfold sig_lookup_assn. simpl.
   unfold sig_has_assn in H0. unfold sig_lookup_assn in H0. simpl in H0. auto.
Qed.


Lemma weakening_sig1_r_mutual: forall S1 S2,
  (forall E e t,  wf_tm S1 S2 E e t ->
     forall S1' S2' d, S1 = (bundle_defns (get_tctr_defns_of d)::S1') ->
                      S2 = (bundle_defns d::S2') ->
                      disjoint (dom_defns d) (dom_sig S1') ->
                      wf_dom_defns d ->
                      wf_tm (bundle_defns d::S1') (bundle_defns d::S2') E e t )
/\
  (forall E t_e args t_b t_res, 
   wf_branches S1 S2 E t_e args t_b t_res ->
   forall S1' S2' d, S1 = (bundle_defns (get_tctr_defns_of d)::S1') ->
                     S2 = (bundle_defns d::S2') ->
                      disjoint (dom_defns d) (dom_sig S1') ->
                      wf_dom_defns d->
                     wf_branches  (bundle_defns d::S1') (bundle_defns d::S2')  E t_e args t_b t_res )
/\
  (forall s args t_ctr t_b t_res, wf_brn S1 S2 s args t_ctr t_b t_res -> 
     forall S1' S2' d,
     S1 = (bundle_defns (get_tctr_defns_of d)::S1') ->
     S2 = (bundle_defns d::S2') ->
                      disjoint (dom_defns d) (dom_sig S1') ->
                      wf_dom_defns d ->
     wf_brn  (bundle_defns d::S1') (bundle_defns d::S2')  s args t_ctr t_b t_res )
/\
  (forall E, wf_env S1 S2 E ->
     forall S1' S2' d,
     S1 = (bundle_defns (get_tctr_defns_of d)::S1') ->
     S2 = (bundle_defns d::S2') ->
     disjoint (dom_defns d) (dom_sig S1') ->
     wf_dom_defns d ->
     wf_env  (bundle_defns d::S1') (bundle_defns d::S2')  E)
 .
Proof. 

introv.
apply mutual_wf_tm_branches_brn_env_ind; eauto.

Case "ctr".
intros. subst. 
apply wf_tm_ctr; auto. 
 eapply sig_has_ctr_weakening_sig1_r; eauto. 

Case "assn".
intros. subst. 
apply wf_tm_assn; auto.

Case "pif".
intros.
eapply wf_tm_TM_PIF; eauto. subst; auto.

Case "tm_mat".
intros; subst. eapply wf_tm_mat_TYPE with (k:= k); eauto.
 eapply sig_has_ctr_weakening_sig1_r; eauto. 

Case "tm_mat".
intros; subst. eapply wf_tm_mat_PROP with (k:= k); eauto.
 eapply sig_has_ctr_weakening_sig1_r; eauto. 

Case "branches". 
 intros; subst. eapply wf_branches_Some; eauto.
  eapply sig_has_ctr_weakening_sig1_r; eauto. 
 
Case "wf_env".
 intros; subst.
 eapply wf_env_cons_eq; eauto. 
Qed.

Lemma wf_tm_weakening_sig1_r: 
forall S1 S2  E e t, 
  wf_tm S1 S2 E e t->
     forall S1' S2' d, S1 = (bundle_defns (get_tctr_defns_of d)::S1') ->
     S2 = (bundle_defns d::S2') ->
     disjoint (dom_defns d) (dom_sig S1') ->
     wf_dom_defns d ->
     wf_tm (bundle_defns d::S1') (bundle_defns d::S2') E e t.
Proof. intros.
eapply (proj1(weakening_sig1_r_mutual  S1 S2)); eauto.
Qed.

 
Lemma weakening_sig2_r_mutual: forall S1 S2,
  wf_sig S1 ->
  sig_extend S1 S2 ->
  (forall E e t, wf_tm S1 S2 E e t->
     forall S,
      disjoint (dom_sig S)  (dom_sig S2)  ->
       wf_tm (S++S1) (S++S2) E e t )
/\
  (forall E t_e args t_b t_res, wf_branches S1 S2 E t_e args t_b t_res ->
      forall S,
      disjoint (dom_sig S)  (dom_sig S2)  ->
      wf_branches  (S ++ S1) (S ++ S2)  E t_e args t_b t_res )
/\
  (forall s args t_ctr t_b t_res, wf_brn S1 S2 s args t_ctr t_b t_res -> 
     forall S,
      disjoint (dom_sig S)  (dom_sig S2)  ->
      wf_brn  (S++S1) (S++S2)  s args t_ctr t_b t_res )
/\
  (forall E, wf_env S1 S2 E ->
     forall S,
      disjoint (dom_sig S)  (dom_sig S2)  ->
      wf_env  (S++S1) (S++S2)  E)
 .
Proof. 

introv Hwf_sig Hext.
apply mutual_wf_tm_branches_brn_env_ind; intros ; eauto. 

Case "ctr".
 
 assert (Jd: disjoint (dom_sig S) (dom_sig S1)).
  eapply sig_extend_disjoint; eauto.

apply wf_tm_ctr; auto.
  eapply sig_has_ctr_weakening_sig_r; eauto.

Case "assn".
 
 assert (Jd: disjoint (dom_sig S) (dom_sig S1)).
  eapply sig_extend_disjoint; eauto.

apply wf_tm_assn; auto.
  eapply sig_has_assn_weakening_sig_r; eauto.

Case "pif".
 assert (Jd: disjoint (dom_sig S) (dom_sig S1)).
  eapply sig_extend_disjoint; eauto.
eapply wf_tm_TM_PIF; eauto. 
eapply ( atomic_types_weakening_sig_r); auto. 

Case "tm_mat".
  assert (Jd: disjoint (dom_sig S) (dom_sig S1)).
  eapply sig_extend_disjoint; eauto.
  intros. eapply wf_tm_mat_TYPE with (k:= k); eauto.
  apply sig_has_ctr_weakening_sig_r; auto.
   destruct b as [l [Jctr Jaux]].
   unfold branches_cover; exists l; split; auto.
   apply get_ctrs_weakening_sig_r; auto. 
 
Case "tm_mat".
  assert (Jd: disjoint (dom_sig S) (dom_sig S1)).
  eapply sig_extend_disjoint; eauto.
  intros. eapply wf_tm_mat_PROP with (k:= k); eauto.
  apply sig_has_ctr_weakening_sig_r; auto.
   destruct b as [l [Jctr Jaux]].
   unfold branches_cover; exists l; split; auto.
   apply get_ctrs_weakening_sig_r; auto. 

Case "branches". 
  assert (Jd: disjoint (dom_sig S) (dom_sig S1)).
  eapply sig_extend_disjoint; eauto.
  eapply wf_branches_Some; eauto.
  eapply sig_has_ctr_weakening_sig_r; eauto.
 
Case "wf_env".
 assert (Jd: disjoint (dom_sig S) (dom_sig S1)).
 eapply sig_extend_disjoint; eauto.
 eapply wf_env_cons_eq; eauto.  
 eapply ( atomic_types_weakening_sig_r); auto. 
Qed.

Lemma wf_tm_weakening_sig2_r: 
forall S1 S2 E e t S, 
  wf_sig S1 ->
  sig_extend S1 S2 ->
  wf_tm S1 S2 E e t->
      disjoint (dom_sig S)  (dom_sig S2)  ->
       wf_tm (S++S1) (S++S2) E e t.
Proof. intros.
apply (proj1(weakening_sig2_r_mutual  H H0)); auto.
Qed.

Lemma ctr_decls_has_strong_type_inv1: forall P S1 S2 c cdl t0 t c1 ds d dsc,
 wf_ctr_decls P
  ((bundle_defns(get_tctr_defns_of(ds++(c1,t0,dsc++cdl)::d))::S1))
  ((bundle_defns(ds++(c1,t0,dsc++cdl)::d)::S2)) c1 t0 cdl ->
 ctr_decls_has c t cdl ->
 exists k, wf_tm  ((bundle_defns(get_tctr_defns_of(ds++(c1,t0,dsc++cdl)::d))::S1))
    (bundle_defns(ds++(c1,t0,dsc++cdl)::d)::S2) empty t k.
Proof. 
intros. gen ds dsc.
induction cdl; intros.
Case "nil".
 inversion H0.
Case "cons".
 destruct a. unfold ctr_decls_has in H0. unfold ctr_decls_lookup in H0.
  destruct (eq_ctr_dec c c0).
  Case "c=c0".
  inversion H0; subst.
  inversion H; subst. exists K; auto.  
 Case "c<>c0".
 fold ctr_decls_lookup in H0. 
 inversion H; subst.
 assert (J:wf_ctr_decls P 
  ((bundle_defns (get_tctr_defns_of (ds++(c1,t0,(dsc++(ctr0, length args, t1)::nil) ++ cdl)::d))) ::S1)
  ((bundle_defns (ds++(c1,t0,(dsc++(ctr0, length args, t1)::nil) ++ cdl)::d)) ::S2) c1 t0 cdl).
 replace  ((dsc++(ctr0,length args, t1)::nil)++cdl) with (dsc++(ctr0,length args, t1) ::cdl); auto.
  rewrite app_ass. rewrite <- app_comm_cons; auto.
 destruct ( IHcdl H0 ds (dsc++(ctr0, length args, t1)::nil) J) as [k Jwf].
 exists k. 
  replace  ((dsc++(ctr0,length args, t1)::nil)++cdl) with (dsc++(ctr0,length args, t1) ::cdl) in Jwf; auto.
  rewrite app_ass. rewrite <- app_comm_cons; auto.
Qed. 


Lemma bundle_defns_has_strong_typ_inv:  forall c d t S  K P ds,
 wf_sig S ->
 bundle_defns_ctr_lookup c d = Some t ->
 disjoint (dom_defns(ds++d)) (dom_sig S) ->
 wf_bundle_tctrs S d K ->
 wf_bundle_ctrs P (bundle_defns (get_tctr_defns_of(ds++d))::S)  (bundle_defns (ds++d)::S) d ->
 exists k, wf_tm (bundle_defns (get_tctr_defns_of(ds++d))::S) ((bundle_defns (ds++d))::S) empty t k.
Proof.
intros.
gen ds.
induction d; intros. 
Case "nil".
 inversion H0.
Case "cons". 
  destruct a.
  unfold bundle_defns_ctr_lookup in H0.
  destruct p.
  destruct (eq_ctr_dec c c1).
  Case "c=c1".
  inversion H0; subst.
  inversion H2; subst.
  exists S_KIND. 
  eapply wf_tm_weakening_sig_r; eauto.  
 Case "c<>c1". 
   remember (ctr_decls_lookup c c0) as C.
   destruct C.
   Case "C=Some t1".
   inversion H0; subst.   inversion H3; subst.
   replace c0 with (nil++c0); auto. symmetry in HeqC.
   eapply ctr_decls_has_strong_type_inv1 with (dsc:=(nil: ctr_decls)) (c:=c); eauto.

   Case "C=None".
   fold bundle_defns_ctr_lookup in H0.
   inversion H2; subst. inversion H3; subst.
   assert (J:= IHd H0 H8 (ds++(c1,t0,c0)::nil)).
   replace ((ds++(c1,t0,c0)::nil)++d) with (ds++(c1,t0,c0)::d) in J.
   destruct ( J H1 H15) as [k Jwf].
   exists k; auto.
    rewrite app_ass. rewrite <- app_comm_cons. auto.
Qed.

Lemma ctr_decls_has_strong_type_inv2: forall P S c cdl t0 t c1 ds d dsc,
 wf_ctr_decls P
  ((bundle_defns(get_tctr_defns_of(ds++(c1,t0,dsc++cdl)::d))::S))
  ((bundle_defns(ds++(c1,t0,dsc++cdl)::d)::S)) c1 t0 cdl ->
 ctr_decls_has c t cdl ->
  disjoint (dom_defns (ds++(c1,t0,dsc++cdl)::d)) (dom_sig S) ->
  wf_dom_defns(ds++(c1,t0,dsc++cdl)::d) ->
 exists k, wf_tm  ((bundle_defns(ds++(c1,t0,dsc++cdl)::d))::S)
    (bundle_defns(ds++(c1,t0,dsc++cdl)::d)::S) empty t k.
Proof.  
intros. gen ds dsc.
induction cdl; intros.
Case "nil".
 inversion H0.
Case "cons".
 destruct a. unfold ctr_decls_has in H0. unfold ctr_decls_lookup in H0.
  destruct (eq_ctr_dec c c0).
  Case "c=c0".
  inversion H0; subst.
  inversion H; subst. exists K; eapply wf_tm_weakening_sig1_r; eauto.  
 Case "c<>c0".
 fold ctr_decls_lookup in H0. 
 inversion H; subst.
 assert (J:wf_ctr_decls P 
  ((bundle_defns (get_tctr_defns_of(ds++(c1,t0,(dsc++(ctr0, length args, t1)::nil) ++ cdl)::d))) ::S)
  ((bundle_defns (ds++(c1,t0,(dsc++(ctr0, length args, t1)::nil) ++ cdl)::d)) ::S) c1 t0 cdl).
 replace  ((dsc++(ctr0,length args, t1)::nil)++cdl) with (dsc++(ctr0,length args, t1) ::cdl); auto.
  rewrite app_ass. rewrite <- app_comm_cons; auto.
 destruct ( IHcdl H0 ds (dsc++(ctr0, length args, t1)::nil) J) as [k Jwf]. 
 replace  ((dsc++(ctr0,length args, t1)::nil)++cdl) with (dsc++(ctr0,length args, t1) ::cdl) ; auto.
   rewrite app_ass. rewrite <- app_comm_cons; auto.
   replace  ((dsc++(ctr0,length args, t1)::nil)++cdl) with (dsc++(ctr0,length args, t1) ::cdl) ; auto. 
   rewrite app_ass. rewrite <- app_comm_cons; auto.
 exists k. 
  replace  ((dsc++(ctr0,length args, t1)::nil)++cdl) with (dsc++(ctr0,length args, t1) ::cdl) in Jwf; auto.
  rewrite app_ass. rewrite <- app_comm_cons; auto.
Qed. 

Lemma bundle_defns_has_strong_typ_inv1:  forall c d t S K P ds,
 wf_sig S ->
 bundle_defns_ctr_lookup c d = Some t ->
 disjoint (dom_defns(ds++d)) (dom_sig S) ->
 wf_bundle_tctrs S d K ->
 wf_bundle_ctrs P (bundle_defns (get_tctr_defns_of (ds++d))::S)  (bundle_defns (ds++d)::S) d ->
 wf_dom_defns (ds++d) ->
 exists k, wf_tm (bundle_defns (ds++d)::S) ((bundle_defns (ds++d))::S) empty t k.
Proof.
intros.
gen ds.
induction d; intros. 
Case "nil".
 inversion H0.
Case "cons". 
  destruct a.
  unfold bundle_defns_ctr_lookup in H0.
  destruct p.
  destruct (eq_ctr_dec c c1).
  Case "c=c1".
  inversion H0; subst.
  inversion H2; subst.
  exists S_KIND.
  assert (J: wf_tm   (bundle_defns(ds++(c1,t,c0)::d)::nil++S)   (bundle_defns(ds++(c1,t,c0)::d)::nil++S) empty t S_KIND).
  eapply wf_tm_weakening_sig2_r with (S1:=S) (S2:=S) (S:= bundle_defns(ds++(c1,t,c0)::d)::nil ); eauto.
  unfold sig_extend; auto.
  unfold dom_sig; simpl; rewrite <- app_nil_end; auto.
  replace (bundle_defns(ds++(c1,t,c0)::d)::nil++S) with (bundle_defns(ds++(c1,t,c0)::d)::S) in J; auto.
 Case "c<>c1". 
   remember (ctr_decls_lookup c c0) as C.
   destruct C.
   Case "C=Some t1".
   inversion H0; subst.   inversion H3; subst.
   replace c0 with (nil++c0); auto. symmetry in HeqC. 
    eapply ctr_decls_has_strong_type_inv2 with (dsc:=(nil: ctr_decls)) (c:=c); eauto.

   Case "C=None".
   fold bundle_defns_ctr_lookup in H0.
   inversion H2; subst. inversion H3; subst.
   assert (J:= IHd H0 H9 (ds++(c1,t0,c0)::nil)).
   replace ((ds++(c1,t0,c0)::nil)++d) with (ds++(c1,t0,c0)::d) in J.
   destruct ( J H1 H16 H4) as [k Jwf].
   exists k; auto.
    rewrite app_ass. rewrite <- app_comm_cons. auto.
Qed.

Lemma   wf_bundle_tctrs_has_inv: forall S c d K t,
  wf_bundle_tctrs S (get_tctr_defns_of d) K -> 
  bundle_defns_ctr_lookup c (get_tctr_defns_of d) = Some t 
     -> wf_tm S S empty t S_KIND.
Proof.
 introv Htctr Hctr.
 induction d.
 simpl in Hctr; auto. inversion Hctr.
 destruct a. destruct p.
 simpl in Htctr. 
 inversion Htctr; subst.
 simpl in Hctr.
 destruct (eq_ctr_dec c c1).
 inversion Hctr; subst; auto.
 
 apply IHd; auto.
Qed.

Lemma sig_has_ctr_has_typ_inv: forall S1 S2 c t,
 wf_sig S1 -> sig_extend S1 S2 -> sig_has_ctr c t S1 -> exists k, wf_tm S1 S2 empty t k.
Proof. 
introv Hwf_sig HS Hctr.
gen S2.
induction S1; intros; auto.
Case "nil".
 inversion Hctr.
Case "cons".
 destruct a. 
 Case "bundle_defns".
 unfold sig_has_ctr in Hctr. unfold sig_lookup_ctr in Hctr. 
 remember (bundle_defns_ctr_lookup c d) as B.
 destruct B.
 Case "B = some". 
  inversion Hctr; subst.   inversion Hwf_sig; subst.
   Case "TYPE".
   inversion HS; subst.
   Case "S1 = S2".
   replace d with (nil++d); auto. 
   eapply  bundle_defns_has_strong_typ_inv1 with (ds:=nil:defns); eauto.
   replace (nil++d) with d; auto.  eapply wf_bundle_ctrs_gives_wf_dom_defns; eauto. 
   Case "S2 extends S1".
   destruct H as [d1 [S' [JS1 [JS2 JS3]]]].
   inversion JS1; subst. 
   exists S_KIND.  eapply wf_tm_weakening_sig_r; eauto.
   eapply wf_bundle_tctrs_has_inv; eauto.
   
   Case "PROP". 
        inversion HS; subst.
   Case "S1 = S2".
   replace d with (nil++d); auto. 
   eapply  bundle_defns_has_strong_typ_inv1 with (ds:=nil:defns); eauto.
      replace (nil++d) with d; auto.  eapply wf_bundle_ctrs_gives_wf_dom_defns; eauto. 
   Case "S2 extends S1".
   destruct H as [d1 [S' [JS1 [JS2 JS3]]]].
   inversion JS1; subst. 
   exists S_KIND.  eapply wf_tm_weakening_sig_r; eauto.
   eapply wf_bundle_tctrs_has_inv; eauto.

 Case "B=none".
 fold sig_lookup_ctr in Hctr. 
 inversion Hwf_sig; subst.
 Case "TYPE".
    inversion HS; subst.
   Case "S1 = S2".
   assert (Jext: sig_extend S1 S1). unfold sig_extend; auto.
  destruct (IHS1 H1 Hctr S1 Jext) as [k Jwf].
 exists k.  replace (bundle_defns d ::S1) with ((bundle_defns d ::nil)++S1); auto. 
 eapply wf_tm_weakening_sig2_r; auto. 
  unfold dom_sig; simpl.  rewrite <- app_nil_end; auto. 

  Case "S2 extends S1".
   destruct H as [d1 [S' [JS1 [JS2 JS3]]]].
   inversion JS1; subst.
   assert (Jext: sig_extend S' S'). unfold sig_extend; auto.
   destruct (IHS1 H1 Hctr S' Jext) as [k Jwf].
   exists k.  eapply wf_tm_weakening_sig_r; auto. 

 Case "Prop".  
      inversion HS; subst.
   Case "S1 = S2".
   assert (Jext: sig_extend S1 S1). unfold sig_extend; auto.
  destruct (IHS1 H1 Hctr S1 Jext) as [k Jwf].
 exists k.  replace (bundle_defns d ::S1) with ((bundle_defns d ::nil)++S1); auto. 
 eapply wf_tm_weakening_sig2_r; auto. 
  unfold dom_sig; simpl.  rewrite <- app_nil_end; auto. 

  Case "S2 extends S1".
   destruct H as [d1 [S' [JS1 [JS2 JS3]]]].
   inversion JS1; subst.
   assert (Jext: sig_extend S' S'). unfold sig_extend; auto.
   destruct (IHS1 H1 Hctr S' Jext) as [k Jwf].
   exists k.  eapply wf_tm_weakening_sig_r; auto. 

 Case "bundle_assn".
 unfold sig_has_ctr in Hctr. unfold sig_lookup_ctr in Hctr.
 fold sig_lookup_ctr in Hctr. 
  inversion Hwf_sig; subst.  
  inversion HS; subst.
  Case "S1 = S2".
  assert (Jext: sig_extend S1 S1). unfold sig_extend; auto.
  destruct (IHS1 H2 Hctr S1 Jext) as [k Jwf].
  exists k.  replace (bundle_assert c0 t0 ::S1) with ((bundle_assert c0 t0 ::nil)++S1); auto. 
  eapply wf_tm_weakening_sig2_r; auto. 
  unfold dom_sig; simpl.  unfold disjoint; unfold not; intros.
  destruct (in_inv H). subst. apply H5; auto. inversion H1.
  
  Case "S2 extends S1".
  destruct H as [d1 [S' [JS1 JS2]]].
  inversion JS1; subst.
Qed. 

Lemma sig_has_assn_has_typ_inv: forall S1 S2 c t,
 wf_sig S1 -> sig_extend S1 S2 ->  sig_has_assn c t S1 ->  (constructs t K_PROP /\wf_tm S1 S2 empty t S_KIND).
Proof. 
introv Hwf_sig HS Hassn.
gen S2.
induction S1; intros. 
Case "nil".
 inversion Hassn.
Case "cons".
 destruct a. 
 Case "bundle_defns".
 unfold sig_has_assn in Hassn. unfold sig_lookup_assn in Hassn. fold sig_lookup_assn in Hassn.
 inversion Hwf_sig; subst.
  Case "TYPE".
  inversion HS; subst.
  Case "S1 = S2".
  assert (Jext: sig_extend S1 S1). unfold sig_extend; auto.
  assert (Jwf:= IHS1 H1 Hassn S1 Jext). destruct Jwf; split; auto.
   replace (bundle_defns d ::S1) with ((bundle_defns d ::nil)++S1); auto. 
  eapply wf_tm_weakening_sig2_r; auto. 
  unfold dom_sig; simpl.  rewrite <- app_nil_end; auto. 

  Case "S2 extends S1".
   destruct H as [d1 [S' [JS1 [JS2 JS3]]]].
   inversion JS1; subst.
   assert (Jext: sig_extend S' S'). unfold sig_extend; auto.
   assert (Jwf := IHS1 H1 Hassn S' Jext).  destruct Jwf; split; auto.
   eapply wf_tm_weakening_sig_r; auto.
  Case "PROP".
     inversion HS; subst.
  Case "S1 = S2".
  assert (Jext: sig_extend S1 S1). unfold sig_extend; auto.
  assert (Jwf:= IHS1 H1 Hassn S1 Jext).  destruct Jwf; split; auto.
   replace (bundle_defns d ::S1) with ((bundle_defns d ::nil)++S1); auto. 
  eapply wf_tm_weakening_sig2_r; auto. 
  unfold dom_sig; simpl.  rewrite <- app_nil_end; auto. 

  Case "S2 extends S1".
   destruct H as [d1 [S' [JS1 [JS2 JS3]]]].
   inversion JS1; subst.
   assert (Jext: sig_extend S' S'). unfold sig_extend; auto.
   assert (Jwf := IHS1 H1 Hassn S' Jext). destruct Jwf; split; auto.
   eapply wf_tm_weakening_sig_r; auto.

 Case "bundle_assn".
  unfold sig_has_assn in Hassn. unfold sig_lookup_assn in Hassn. 
  destruct (eq_ctr_dec c c0).
  Case "c = c0".
    inversion Hassn; subst.
    inversion Hwf_sig; subst; split; auto.
     inversion HS; subst.
     Case "S1 = S2".
     assert (Jext: sig_extend S1 S1). unfold sig_extend; auto.
     replace (bundle_assert c0 t ::S1) with ((bundle_assert c0 t ::nil)++S1); auto. 
     eapply wf_tm_weakening_sig2_r; auto. 
     unfold dom_sig; simpl.  unfold disjoint; unfold not; intros.
     destruct (in_inv H). subst. apply H5; auto. inversion H1.
   
   Case "S2 extends S1".
   destruct H as [d1 [S' [JS1 JS2]]].
   inversion JS1. 

  Case "c<>c0".
   fold sig_lookup_assn in Hassn.
   inversion Hwf_sig; subst.
     inversion HS; subst.
     Case "S1 = S2".
     assert (Jext: sig_extend S1 S1). unfold sig_extend; auto.
       assert (Jwf:=IHS1 H2 Hassn S1 Jext). destruct Jwf; split; auto.
     replace (bundle_assert c0 t0 ::S1) with ((bundle_assert c0 t0 ::nil)++S1); auto. 
     eapply wf_tm_weakening_sig2_r; auto. 
     unfold dom_sig; simpl.  unfold disjoint; unfold not; intros. 
     destruct (in_inv H1). subst. apply H5; auto. inversion H7.
   
   Case "S2 extends S1".
   destruct H as [d1 [S' [JS1 JS2]]].
   inversion JS1. 
Qed.

Lemma wf_tm_tm: forall S1 S2 E e t,
  wf_sig S1 -> wf_tm S1 S2 E e t ->  sig_extend S1 S2 -> t = S_KIND \/ exists k, wf_tm S1 S2 E t k.
Proof. 
intros.
induction H0; subst; auto. 

Case "arr".
 pick_fresh z. 
 assert (Jz: z \notin L); auto.
 assert (J:= H3 z Jz ).
 destruct J; auto. 
 right. destruct H5 as [k Jwf]. exists k.   rewrite_env (E & empty).
  replace (E&z~:u1) with  (E&z~:u1&empty) in Jwf; auto.
  eapply (proj1 (typing_strengthening_var_mutual S2 H)); eauto.
 rewrite_env E. eapply wf_tm_gives_wf_env; eauto.

Case "ctr".
 right.  destruct (sig_has_ctr_has_typ_inv H H1 H2) as [k Jwf].
 exists k. rewrite_env (empty & E & empty).
 replace (empty:env) with ((empty:env)&empty) in Jwf; auto.
 apply typing_weakening; auto. rewrite_env E; auto.

 
Case "assn".
 right.  destruct (sig_has_assn_has_typ_inv H H1 H2).
 exists S_KIND.  rewrite_env (empty & E & empty).
 replace (empty:env) with ((empty:env)&empty) in H4; auto.
 apply typing_weakening; auto. rewrite_env E; auto.

Case "fv".
  destruct (binds_var_inversion H2) as [E1 [E2  Jeq]].
  rewrite <- Jeq in H0.  right. 
 destruct (wf_env_strong_inv_var E1 x t E2 H0) as [k J].
 exists k. rewrite <- Jeq.
 rewrite_env (E1&(x~:t&E2)).
 eapply typing_weakening'; eauto. 
 simpl_env; auto.

right; exists k; auto.

Case "app". 
 destruct IHwf_tm1. inversion H2.
 destruct H2 as [ku Jwfu].
 right. inversion Jwfu; subst.
 exists ku.
 destruct H0.
 pick_fresh z.
 rewrite (@subst_intro z); auto. rewrite <- (@subst_fresh z e2 ku); auto.
 eapply simple_tm_substitution; eauto. 
 eapply(proj1 (wf_tm_branches_lc_mutual S1 S2)); eauto.
 unfold open_tm; rewrite <- open_lc_id; auto.
 pick_fresh z.
 assert (J: wf_tm S1 S2 (E&z~:t2) (u^^ tm_fv z) ku). eapply H6; eauto.
  unfold open_tm in J.  rewrite <- open_lc_id in J; auto.
  rewrite_env (E&empty).
  replace (E& z~:t2) with (E& z~:t2& empty) in J; auto.
 eapply (proj1 (typing_strengthening_var_mutual S2 H)); eauto.
 rewrite_env E; auto. eapply wf_tm_gives_wf_env;eauto.

Case "P_SAYS".
right; exists K_PROP; auto. eapply wf_tm_P_SAYS; eauto. 
  rewrite_env (empty&E); eapply typing_weakening'; eauto.
  rewrite_env E;eapply wf_tm_gives_wf_env; eauto .
  rewrite_env (empty&E); eapply typing_weakening'; eauto.
  rewrite_env E;eapply wf_tm_gives_wf_env; eauto .
right ;exists S_KIND; auto.
right; exists S_KIND; auto.
apply wf_tm_K_TYPE; auto. eapply wf_tm_gives_wf_env; eauto.
right; exists K_TYPE; auto.


Case "T_PF". 
 right.
 destruct IHwf_tm2.  inversion H2.
  destruct H2 as [k Jwf]. 
  inversion Jwf; subst. 
  pick_fresh z. assert (Jz: z\notin L); auto. assert (J1:= H6 z Jz).
  unfold open_tm in J1. rewrite <- open_lc_id in J1. 
  inversion J1; subst. inversion H5. 
  exists K_TYPE; auto. 
  rewrite_env (E & empty).
  replace (E&z~:P) with  (E&z~:P&empty) in J1; auto.
  eapply (proj1 (typing_strengthening_var_mutual S2 H)); eauto.
  simpl; auto. rewrite_env E. eapply wf_tm_gives_wf_env; eauto.
  apply lc_app;  auto. unfold T_PF; auto.

right; exists K_TYPE; auto.
  apply wf_tm_T_PF; auto. apply wf_tm_P_SAYS; auto. apply wf_tm_TM_SELF. eapply wf_tm_gives_wf_env; eauto.
right; exists K_TYPE; auto.
right; exists K_PROP; auto.

Case "pf_bind".
right; exists K_PROP; auto. 
  destruct IHwf_tm2.  inversion H2.
  destruct H2 as [k Jwf]. 
  inversion Jwf; subst. 
  pick_fresh z. assert (Jz: z\notin L); auto. assert (J1:= H6 z Jz).
  unfold open_tm in J1. rewrite <- open_lc_id in J1.
  assert (J2: wf_tm S1 S2 E (tm_app(tm_app P_SAYS a ) Q ) k).
  rewrite_env (E & empty).
  replace (E&z~:P) with  (E&z~:P&empty) in J1; auto.
  eapply (proj1 (typing_strengthening_var_mutual S2 H)); eauto.
  simpl; auto. rewrite_env E. eapply wf_tm_gives_wf_env; eauto.
 inversion J2; subst. inversion H5. inversion H12.
   apply wf_tm_P_SAYS; auto. 
  repeat (apply lc_app); auto. unfold P_SAYS; auto.
    destruct IHwf_tm1.  inversion H2.
    destruct H2 as [ka Jwf1]. 
    assert (Ja: lc(tm_app(tm_app P_SAYS a) P)). 
    eapply (proj1(wf_tm_branches_lc_mutual S1 S2)); eauto.
    inversion Ja; subst. inversion H5; subst; auto.

right; exists K_TYPE; auto.
right; exists K_PROP; auto.
right; exists K_TYPE; auto.
Qed.


Lemma  fully_appled_ctr_in_bundle_tctr: forall P S1 S2 d c t c_inst ctr0 args S' K,
  wf_bundle_tctrs S' d K  ->
  (K =  K_TYPE \/ K=K_PROP) -> 
  wf_bundle_ctrs  P S1 S2 d ->
  bundle_defns_ctr_lookup c d = Some t -> 
  constructs t c_inst ->
  fully_applied_args c_inst ctr0 args ->
  In ctr0 (dom_defns (get_tctr_defns_of d)) .
Proof.
introv Hwf_tctrs Heq Hwf_bundle Hc Hcon. intro Hfully_app. 
induction d.
Case "nil". inversion Hc.
Case "cons".
  destruct a.
  inversion Hc; subst.
  destruct p.
  destruct (eq_ctr_dec c c1).
  Case "c=c1".
   inversion H0; subst. 
   inversion Hwf_bundle; subst.
   inversion Hwf_tctrs; subst.
   destruct Heq; subst.
   assert (Jeq:= construct_unique Hcon H13).  
   rewrite Jeq in Hfully_app; inversion Hfully_app.
   assert (Jeq:= construct_unique Hcon H13). 
   rewrite Jeq in Hfully_app; inversion Hfully_app.
  Case "c<>c1".
    remember (ctr_decls_lookup c c0) as C.
    destruct C.
    Case "C = Some t".
     inversion H0; subst. 
     inversion Hwf_bundle; subst.
     symmetry in HeqC.
     destruct c.
     assert (J:= ctr_decls_has_strong_type_inv H8 HeqC).
     destruct J as [c_inst' [args' [nt [ Jlevel [Jcon [Jfully_app J1]]]]]].
      assert (Jeq1:= construct_unique Hcon Jcon).  
      rewrite Jeq1 in Hfully_app. 
      destruct (fully_applied_args_unique Hfully_app Jfully_app).  
      subst. auto.
    Case "C = None".
      inversion Hwf_tctrs; subst.
      inversion Hwf_bundle; subst.
      simpl. right. (*eapply in_or_app; *) eauto.
Qed.

Lemma  fully_appled_ctr_in_bundle: forall P S1 S2 d c t c_inst ctr0 args S' K,
  wf_bundle_tctrs S' d K  ->
  (K =  K_TYPE \/ K=K_PROP) -> 
  wf_bundle_ctrs  P S1 S2 d ->
  bundle_defns_ctr_lookup c d = Some t -> 
  constructs t c_inst ->
  fully_applied_args c_inst ctr0 args ->
  In ctr0 (dom_defns  d).
Proof.  
 intros.
 eapply  in_tctr_defns_in; eauto.
 eapply fully_appled_ctr_in_bundle_tctr; eauto.
Qed.

Lemma  fully_applied_args_ctr_in_sig: forall S c t c_inst ctr0 args,
  wf_sig S ->
  sig_has_ctr c t S -> 
  constructs t c_inst ->
  fully_applied_args c_inst ctr0 args ->
  In ctr0 (dom_sig S).
Proof.
 introv Hwf_sig Hwf_c Hcon Hfully_app.
 induction S.
Case "nil".
 inversion Hwf_c.
Case "cons".
 destruct a.
 Case "bundle_defns".
 inversion Hwf_c; subst.
 remember (bundle_defns_ctr_lookup c d) as B.
 destruct B.
    Case "B = Some t".
    inversion H0; subst.
    inversion Hwf_sig; subst.
    Case "wf_bundle TYPE".
      symmetry in HeqB.
      unfold dom_sig; simpl.
      assert (J: In ctr0 (dom_defns d)).
      eapply fully_appled_ctr_in_bundle; eauto.
      eapply in_or_app; eauto.
    Case "wf_bundle PROP".
      symmetry in HeqB.
      unfold dom_sig; simpl.
      assert (J: In ctr0 (dom_defns d)).
      eapply fully_appled_ctr_in_bundle; eauto.
      eapply in_or_app; eauto.
   Case "B = None".
     unfold dom_sig; simpl. eapply in_or_app; eauto.
     inversion Hwf_sig; subst; right; eapply IHS; eauto.
 Case "bundle_assert".
 simpl. right.
 inversion Hwf_sig; subst.
 inversion Hwf_c; subst.
 apply IHS; auto.
Qed.
  
Lemma  fully_applied_args_ctr_in_ext_sig: forall S1 S2 c t c_inst ctr0 args,
  wf_sig S1 ->
  sig_extend S1 S2 ->
  sig_has_ctr c t S1 -> 
  constructs t c_inst ->
  fully_applied_args c_inst ctr0 args ->
  In ctr0 (dom_sig S2).
Proof.
 intros.
 destruct H0; subst; auto.
 eapply fully_applied_args_ctr_in_sig; eauto.

 destruct H0 as [d [S [HS1 [HS2 HS3]]]].
 subst.
 assert (J: In ctr0 (dom_sig(bundle_defns (get_tctr_defns_of d)::S))).
 eapply fully_applied_args_ctr_in_sig; eauto.
 unfold dom_sig in J; simpl in J.
 unfold dom_sig; simpl.
 apply in_or_app.
 destruct (in_app_or (dom_defns (get_tctr_defns_of d)) (flat_map dom_bundle S)  ctr0 J); auto.
 left; apply in_tctr_defns_in; auto.
Qed.
 
 
Lemma  fully_applied_ctr_defns_some: forall P S1 S2 d c t c_inst ctr0 args S' K,
  wf_bundle_tctrs S' d K  ->
  (K =  K_TYPE \/ K=K_PROP) -> 
  wf_bundle_ctrs  P S1 S2 d ->
  bundle_defns_ctr_lookup c d = Some t -> 
  constructs t c_inst ->
  fully_applied_args c_inst ctr0 args ->
  exists l, get_defns_ctrs_of ctr0 d = Some l.
Proof.
introv Hwf_tctrs Heq Hwf_bundle Hc Hcon. intro Hfully_app. 
induction d.
Case "nil". inversion Hc.
Case "cons".
  destruct a.
  inversion Hc; subst.
  destruct p.
  destruct (eq_ctr_dec c c1).
  Case "c=c1".
   inversion H0; subst. 
   inversion Hwf_bundle; subst.
   inversion Hwf_tctrs; subst.
   destruct Heq; subst.
   assert (Jeq:= construct_unique Hcon H13).  
   rewrite Jeq in Hfully_app; inversion Hfully_app.
   assert (Jeq:= construct_unique Hcon H13). 
   rewrite Jeq in Hfully_app; inversion Hfully_app.
  Case "c<>c1".
    remember (ctr_decls_lookup c c0) as C.
    destruct C.
    Case "C = Some t".
     inversion H0; subst. 
     inversion Hwf_bundle; subst.
     symmetry in HeqC.
     destruct c.
     assert (J:= ctr_decls_has_strong_type_inv H8 HeqC).
     destruct J as [c_inst' [args' [nt [ Jlevel [Jcon [Jfully_app J1]]]]]].
      assert (Jeq1:= construct_unique Hcon Jcon).  
      rewrite Jeq1 in Hfully_app. 
      destruct (fully_applied_args_unique Hfully_app Jfully_app).  
      subst. exists (dom_ctr_decls c0). simpl;  eauto.
      destruct (eq_ctr_dec c1 c1); auto.
      destruct n2; auto.
    Case "C = None".
      inversion Hwf_tctrs; subst.
      inversion Hwf_bundle; subst.
      simpl. destruct (eq_ctr_dec ctr0 c1); subst; eauto.
Qed.

Lemma  fully_applied_args_in_sig_some: forall S c t c_inst ctr0 args,
  wf_sig S ->
  sig_has_ctr c t S -> 
  constructs t c_inst ->
  fully_applied_args c_inst ctr0 args ->
  exists l, get_ctrs_of ctr0 S = Some l.
Proof.
 introv Hwf_sig Hwf_c Hcon Hfully_app.
 induction S.
Case "nil".
 inversion Hwf_c.
Case "cons".
 destruct a.
 Case "bundle_defns".
 inversion Hwf_c; subst.
 remember (bundle_defns_ctr_lookup c d) as B.
 destruct B.
    Case "B = Some t".
    inversion H0; subst.
    inversion Hwf_sig; subst.
    Case "wf_bundle TYPE".
      symmetry in HeqB.
      unfold dom_sig; simpl.
      assert (J: exists l, get_defns_ctrs_of ctr0 d = Some l).
      eapply  fully_applied_ctr_defns_some; eauto.
      destruct J as [l Jget].
      exists l. remember (get_defns_ctrs_of ctr0 d) as G.
      destruct G; subst; auto. inversion Jget.

    Case "wf_bundle PROP".
      symmetry in HeqB.
      unfold dom_sig; simpl.
      assert (J: exists l, get_defns_ctrs_of ctr0 d = Some l).
      eapply  fully_applied_ctr_defns_some; eauto.
      destruct J as [l Jget].
      exists l. remember (get_defns_ctrs_of ctr0 d) as G.
      destruct G; subst; auto. inversion Jget.

   Case "B = None".
     unfold dom_sig; simpl. 
     destruct (get_defns_ctrs_of ctr0 d); eauto.
     inversion Hwf_sig; subst; eapply IHS; eauto.
 Case "bundle_assert".
 simpl. 
 inversion Hwf_sig; subst.
 inversion Hwf_c; subst.
 apply IHS; auto.
Qed.
  

(*Lemma get_defns_ctrs_of_weakening: forall .*)

Lemma get_ctrs_of_weakening: forall c d l,
  get_defns_ctrs_of c (get_tctr_defns_of d) = Some l ->
  exists l,  get_defns_ctrs_of c d = Some l .
Proof.
 intros.
 induction d.
 inversion H.
 destruct a. destruct p.
 simpl in H.
 simpl.
 destruct (eq_ctr_dec c c1).
 subst; exists (dom_ctr_decls c0); simpl; auto.
 eapply IHd; eauto.
Qed.
 
Lemma  fully_applied_ctr_in_ext_sig: forall S1 S2 c t c_inst ctr0 args,
  wf_sig S1 ->
  sig_extend S1 S2 ->
  sig_has_ctr c t S1 -> 
  constructs t c_inst ->
  fully_applied_args c_inst ctr0 args ->
  exists l, get_ctrs_of ctr0 S2 = Some l.
Proof.
 intros.
 destruct H0; subst; auto.
 eapply fully_applied_args_in_sig_some; eauto.

 destruct H0 as [d [S [HS1 [HS2 HS3]]]].
 subst.
 assert (J: exists l, get_ctrs_of ctr0 (bundle_defns (get_tctr_defns_of d)::S) = Some l).
 eapply fully_applied_args_in_sig_some; eauto.
 destruct J as [l Jin].
 simpl in Jin. simpl.
 remember (get_defns_ctrs_of ctr0 (get_tctr_defns_of d)) as G.
 destruct G.
 inversion Jin; subst; symmetry in HeqG.
 assert (J: exists l1, get_defns_ctrs_of ctr0 d = Some l1).
 eapply get_ctrs_of_weakening; eauto.
 destruct J as [l1 J1].
 exists l1; auto.
 destruct (get_defns_ctrs_of ctr0 d); auto. inversion J1.
 
 destruct (get_defns_ctrs_of ctr0 d); eauto.
Qed.
 
 Lemma fully_applied_args_ctr_in_bundle: forall S1 S2  S'  P  K  ctr0 d l c t c_inst args,
  wf_bundle_tctrs S' d K  ->
  (K =  K_TYPE \/ K=K_PROP) -> 
  wf_bundle_ctrs  P S1 S2 d ->
  get_defns_ctrs_of ctr0 d = Some l ->
  bundle_defns_ctr_lookup c d = Some t ->
  constructs t c_inst ->
  fully_applied_args c_inst ctr0 args->
  In c l.
Proof.
 introv Hwf_tctr HeqK Hwf_bundle Hget_ctr Hc. introv Hcon Hfully_app.
 induction d.
 Case "nil". inversion Hget_ctr.
 Case "cons".
  destruct a. destruct p.
  inversion Hc.
  destruct (eq_ctr_dec c c1).
  Case "c=c1".
   inversion H0; subst.
   clear IHd. 
   inversion Hwf_tctr; subst.
   destruct HeqK; subst.
   assert (Jeq:= construct_unique Hcon H10).  
   rewrite Jeq in Hfully_app; inversion Hfully_app.
   assert (Jeq:= construct_unique Hcon H10). 
   rewrite Jeq in Hfully_app; inversion Hfully_app.
  Case "c<>c1".
    remember (ctr_decls_lookup c c0) as C.
    destruct C.
    Case "C = Some t".
     inversion H0; subst. 
     inversion Hwf_bundle; subst.
     symmetry in HeqC.
     inversion Hget_ctr; subst.
     destruct (eq_ctr_dec ctr0 c1).
     Case "ctr0 = c1".
     inversion H1; subst. 
     eapply ctr_decls_lookup_some_in; eauto.
     Case "ctr0 <> c1".
     inversion Hwf_tctr; subst.
     destruct c. 
     assert (J:= ctr_decls_has_strong_type_inv H8 HeqC).
     destruct J as [c_inst' [args' [nt [ Jlevel [Jcon [Jfully_app J1]]]]]].
     assert (Jeq1:= construct_unique Hcon Jcon). 
     rewrite Jeq1 in Hfully_app. 
     destruct ( fully_applied_args_unique Hfully_app Jfully_app); contradiction. 

    Case "C = None".
      inversion Hwf_tctr; subst.
      inversion Hwf_bundle; subst.
      eapply IHd; eauto.
      inversion Hget_ctr.
      destruct (eq_ctr_dec ctr0 c1); auto.
      Case "ctr0 = c1".
      inversion H1;  subst. clear IHd H1.
      assert (J: In c1 (dom_defns d)).
      eapply fully_appled_ctr_in_bundle; eauto. 
      contradiction.
Qed.
  

Lemma fully_applied_args_ctr_in: forall S ctr0 l c t c_inst args,
     wf_sig S ->
     get_ctrs_of ctr0 S = Some l  ->
     sig_has_ctr c t S ->
     constructs t c_inst ->
     fully_applied_args c_inst ctr0 args->
     In c l.
Proof.
  introv Hwf_sig Hget_ctrs Hwf_c Hcon Hfully_app.
  induction S.
  Case "nil". inversion Hwf_c.
  Case "cons".
   destruct a.
   Case "bundle_defns".
    inversion Hwf_c; subst.
    remember (bundle_defns_ctr_lookup c d) as B.
    destruct B.
     Case "B = Some".
      inversion H0; subst.
      inversion Hget_ctrs; subst.
      remember (get_defns_ctrs_of ctr0 d) as G.
      destruct G.
       Case "G = Some l".
        inversion H1; subst.
        symmetry in HeqG. symmetry in HeqB.
        inversion Hwf_sig; subst.
        eapply fully_applied_args_ctr_in_bundle with (d:=d); eauto.
         eapply fully_applied_args_ctr_in_bundle with (d:=d); eauto.

        Case "G=None".
         inversion Hwf_sig; subst.
          assert (J: In ctr0 (dom_defns d)).
          eapply fully_appled_ctr_in_bundle; eauto.
          assert (Jin:=  get_ctr_sig_in ctr0 S H1).
           destruct (H5 ctr0 J Jin).
          assert (J: In ctr0 (dom_defns d)).
          eapply fully_appled_ctr_in_bundle; eauto.
          assert (Jin:=  get_ctr_sig_in ctr0 S H1).
           destruct (H5 ctr0 J Jin).
 
      Case "B=None".
        inversion Hget_ctrs; subst.
        remember (get_defns_ctrs_of ctr0 d) as G.
        destruct G.
        Case "G = Some l".
        inversion H1; subst.  
        inversion Hwf_sig; subst.
         assert (J: In ctr0 (dom_sig S)).
         eapply fully_applied_args_ctr_in_sig with (c:=c); eauto. apply H0.
         symmetry in HeqG.
          assert (Jin:=  get_defns_ctr_some_in ctr0 d HeqG).
           destruct (H5 ctr0 Jin J).
           assert (J: In ctr0 (dom_sig S)).
         eapply fully_applied_args_ctr_in_sig with (c:=c); eauto. apply H0.
         symmetry in HeqG.
          assert (Jin:=  get_defns_ctr_some_in ctr0 d HeqG).
           destruct (H5 ctr0 Jin J).
         
        Case "G=None".
        inversion Hwf_sig; apply IHS; auto. 
  Case "bundle_assert".
    inversion Hget_ctrs. inversion Hwf_c. inversion Hwf_sig.
    apply IHS;auto.
Qed.
                                   
Lemma tm_T_PF_inv:
forall S E P Q,
wf_sig S
-> lc Q 
->  (exists k:tm, wf_tm S S E (tm_arr P (tm_app T_PF Q)) k )
-> wf_tm S S E Q K_PROP.
Proof. 
introv Hwf_sig (*Hwf_P*) Hlc_Q H.
destruct H as [k Hwf].
inversion Hwf; subst. 
pick_fresh x.
assert (Jx : x\notin L). auto. assert (J:= H3 x Jx).
inversion J; subst. 
  inversion H2. 
rewrite <- (open_lc_id) in H2.
replace (concat E (x~:P)) with (E&(x~:P)& empty) in H2.
rewrite_env (E&@empty bnd).
eapply (proj1 (typing_strengthening_var_mutual S Hwf_sig)); eauto.
auto. rewrite_env E. 
rewrite_env E. eapply wf_tm_gives_wf_env;  eauto. auto. auto.
Qed.

Lemma open_eq: forall t1 t2 x k,
 x \notin fv t1 -> x \notin fv t2 ->
open_tm_aux k (tm_fv x) t1 = open_tm_aux k (tm_fv x) t2 -> t1 = t2.
Proof. 
intros.
gen t2 k. induction t1; intros. 
simpl in H. simpl in H1. 
 destruct t2; subst; inversion H1; subst; auto.
 simpl in H1. destruct (b===k). 
 inversion H1; subst. destruct (notin_same H).
 inversion H1.
 
 simpl in H1. destruct t2; subst; try solve [ destruct (b===k); subst; simpl in H1; inversion H1].
 destruct (b===k). inversion H1; subst; simpl in H0; destruct (notin_same H0).
 inversion H1.
 simpl in H1.  destruct (b===k).  destruct (b0 ===k). subst; auto. inversion H1.
  
 destruct (b0 ===k). inversion H1; subst; contradiction. auto.

 destruct t2; subst; inversion H1.
 destruct (b===k); inversion H3. subst; auto.

 destruct t2; subst; inversion H1.
 destruct (b===k); inversion H3. subst; auto.

 destruct t2; subst; simpl in H1; try solve [inversion H1].
 destruct (b===k); subst; inversion H1. 
 inversion H1; subst. 
 assert (J1: t1_1 = t2_1). simpl in H. simpl in H0.  eapply IHt1_1; eauto.
assert (J2: t1_2 = t2_2). simpl in H. simpl in H0. eapply IHt1_2; eauto.
subst; auto.

  destruct t2; subst; simpl in H1; try solve [inversion H1].
 destruct (b===k); subst; inversion H1. 
 inversion H1; subst. 
 assert (J1: t1_1 = t2_1). simpl in H. simpl in H0.  eapply IHt1_1; eauto.
assert (J2: t1_2 = t2_2). simpl in H. simpl in H0. eapply IHt1_2; eauto.
subst; auto.
 
  destruct t2; subst; simpl in H1; try solve [inversion H1].
 destruct (b===k); subst; inversion H1. 
 inversion H1; subst. 
 assert (J1: t1_1 = t2_1). simpl in H. simpl in H0.  eapply IHt1_1; eauto.
assert (J2: t1_2 = t2_2). simpl in H. simpl in H0. eapply IHt1_2; eauto.
subst; auto.

destruct t2; subst; simpl in H1; try solve [inversion H1].
 destruct (b===k); subst; inversion H1. 
 inversion H1; subst. 
 assert (J1: t1_1 = t2_1). simpl in H. simpl in H0.  eapply IHt1_1; eauto.
 assert (J2: t1_2 = t2_2). simpl in H. simpl in H0. eapply IHt1_2; eauto.
 assert (J3: t1_3 = t2_3). simpl in H. simpl in H0. eapply IHt1_3; eauto.
subst; auto.

destruct t2; subst; simpl in H1; try solve [inversion H1].
 destruct (b===k); subst; inversion H1. 
 inversion H1; subst. 
 assert (J1: t1_1 = t2_1). simpl in H. simpl in H0.  eapply IHt1_1; eauto.
assert (J2: t1_2 = t2_2). simpl in H. simpl in H0. eapply IHt1_2; eauto.
subst; auto.

destruct t2; subst; simpl in H1; try solve [inversion H1].
 destruct (b===k); subst; inversion H1. 
 inversion H1; subst. 
 assert (J1: t1_1 = t2_1). simpl in H. simpl in H0.  eapply IHt1_1; eauto.
assert (J2: t1_2 = t2_2). simpl in H. simpl in H0. eapply IHt1_2; eauto.
subst; auto.
Qed.

(* NOT TRUE with beta conversion.*)
Lemma typing_conv_eq:
forall t1 t2,
converts empty t1 t2 -> (t1 = t2).
Proof.
intros t1 t2 Hconv.
remember (empty:env) as E'.
induction Hconv ; auto. 
rewrite (IHHconv HeqE'); auto.
rewrite (IHHconv1 HeqE'); auto.
rewrite HeqE' in H. inversion H.
rewrite (IHHconv1 HeqE'); rewrite (IHHconv2 HeqE'); auto.

pick_fresh x. 
assert (J: t2 = u2).  eapply open_eq with (x:= x) (k:=0) ; eauto.
rewrite (IHHconv HeqE'); rewrite J; auto.

pick_fresh x. 
assert (J: t2 = u2).  eapply open_eq with (x:= x)  (k:=0); eauto.
rewrite (IHHconv HeqE'); rewrite J; auto.
(*
rewrite (IHHconv1 HeqE'); rewrite (IHHconv2 HeqE'); rewrite (IHHconv3 HeqE'); auto.
rewrite (IHHconv1 HeqE'); rewrite (IHHconv2 HeqE'); auto.
rewrite (IHHconv1 HeqE'); rewrite (IHHconv2 HeqE'); auto. *)
Qed.


Inductive step_mat_ctr_type_inv: tm -> tm -> list tm -> nat -> Prop :=
 | step_mat_ctr_type_inv_base: forall  v_tp e_tp ctr args,
    fully_applied_args v_tp ctr args ->
   step_mat_ctr_type_inv v_tp e_tp nil 0

| step_mat_ctr_type_inv_rec_arg: forall L v_tp e_tp T,
  (forall x, x\notin L -> 
     step_mat_ctr_type_inv (v_tp^^(tm_fv x)) (e_tp^^(tm_fv x)) nil 0) ->
  step_mat_ctr_type_inv (tm_arr T v_tp) (tm_arr T e_tp) nil 0

| step_mat_ctr_type_inv_rec_param: forall n v_tp e_tp T a args,
  step_mat_ctr_type_inv (v_tp^^ a)  e_tp args n  ->
  ~(n < 0) ->
  step_mat_ctr_type_inv (tm_arr T v_tp) e_tp (a::args) (n+1).


Inductive wf_brn_p (S:sig) : tm -> (list tm) -> (list tm) -> tm -> tm -> tm -> Prop :=
| wf_brn_p_base :             
     forall  s result_type (*result_type'*),
     wf_brn_p S s nil nil s result_type result_type
| wf_brn_p_constructor_arg :
(* can only be used when the distance between s and result_type > 0*)
     forall s T1 T U a args result_typ,
     wf_brn_p S s nil args (T^^a) (U^^a) result_typ ->
     wf_brn_p S s nil (a::args) (tm_arr T1 T) (tm_arr T1 U) result_typ
| wf_brn_p_datatype_param:
     forall s arg args args1 T1 T U result_type,
       wf_brn_p S s args args1 (T ^^ arg) U result_type ->
       wf_brn_p S s (arg::args) args1 (tm_arr T1 T) U result_type.

Lemma subst_step_mat_ctr_type_inv: forall v_tp e_tp args n x v,
  lc(v) ->  value v ->
  step_mat_ctr_type_inv v_tp e_tp args n ->
  step_mat_ctr_type_inv ([x~>v]v_tp) ([x~>v]e_tp) 
  (List.map(subst_tm x v) args) n.
Proof. 
intros v_tp e_tp args n x v Hlc Hvalue Hinv.
induction Hinv; subst. simpl.
eapply step_mat_ctr_type_inv_base.
eapply subst_fully_applied_args; eauto.

simpl.
gather_atoms' as L'.
eapply step_mat_ctr_type_inv_rec_arg with (L:=L'). 
intros.
assert (J: x0 \notin L). unfold L' in H1; auto.
assert (J1:= H0 x0 J).
assert (J2: ~(x=x0)).  unfold L' in H1; auto.
rewrite subst_open_commute.
rewrite subst_open_commute.

 auto. auto.
auto.  auto. auto.

simpl. 
eapply step_mat_ctr_type_inv_rec_param; eauto.
rewrite subst_open in IHHinv. 
auto. auto.
Qed.


Lemma app_nil_eq: forall  A l,
(@nil A)++l = l.
intros; induction l; auto.
Qed.

Lemma wf_brn_p_inv1: forall S s T U a args  args1 b_tp c_tp e_tp,
  wf_brn_p S s args args1 c_tp b_tp e_tp ->
  s = tm_arr T U ->
  args1 = nil ->
  wf_brn_p S (U^^a) (args++(a::nil)) nil c_tp b_tp e_tp.
Proof. 
intros S s T U a args args1 b_tp c_tp e_tp Hwf Heqs Heqa.

induction Hwf; subst.

assert (J: wf_brn_p S (U^^a) nil nil (U^^a) result_type result_type). apply wf_brn_p_base.
rewrite app_nil_eq.
eapply wf_brn_p_datatype_param; eauto.
inversion Heqa.

rewrite <- app_comm_cons.
apply wf_brn_p_datatype_param; auto.
Qed.

Lemma wf_brn_p_inv2: forall S T U K c_tp b_tp  v args args1,
  lc(v) ->
  wf_brn_p S (tm_arr T U) args args1 c_tp b_tp (tm_arr T K) ->
  wf_brn_p S (U^^v) args (args1++v::nil) c_tp b_tp (K^^v).
Proof.
intros S T U K c_tp b_tp  v args args1 Hlc_v (*Hvalue*) Hwf.
remember (tm_arr T U) as s.
remember (tm_arr T K) as r_tp.
induction Hwf; subst.

rewrite app_nil_eq. 
apply wf_brn_p_constructor_arg.
apply wf_brn_p_base.

rewrite <- app_comm_cons.
eapply wf_brn_p_constructor_arg.
apply IHHwf; auto.

apply wf_brn_p_datatype_param; auto.
Qed.


Inductive all_tm_value: list tm ->Prop :=
| all_tm_value_nil : all_tm_value nil
| all_tm_value_cons: forall v l, value v -> all_tm_value l -> all_tm_value (v::l).

Inductive all_tm_wf: sig -> list tm ->Prop :=
| all_tm_wf_nil : forall S, all_tm_wf S nil
| all_tm_wf_cons: forall S E e t l, 
  wf_tm S S E e t -> all_tm_wf S l -> all_tm_wf S (e::l).

Lemma all_tm_value_app: forall e l,
all_tm_value l -> value e -> all_tm_value (l++e::nil).
Proof. 
intros e l Hl He.
induction l.
rewrite app_nil_eq; apply all_tm_value_cons; auto.
inversion Hl; subst.
rewrite <- app_comm_cons.
apply all_tm_value_cons; auto.
Qed.

Lemma all_tm_wf_app: forall S E e t l,
all_tm_wf S l -> wf_tm S S E e t -> all_tm_wf S (l++e::nil).
Proof. 
intros S E e t l Hl He.
induction l.
rewrite app_nil_eq; eapply all_tm_wf_cons; eauto.
inversion Hl; subst.
rewrite <- app_comm_cons.
eapply all_tm_wf_cons; eauto.
Qed.

Lemma ctr_decl_unique: forall S c t1 t2,
wf_sig S -> sig_has_assn c t1 S -> sig_has_ctr c t2 S -> False.
Proof. 
 intros.
 induction H.
 Case "nil". inversion H0.
 Case "assert".  
   unfold sig_has_assn in H0. simpl in H0. 
   destruct (eq_ctr_dec c c0). 
   Case "c=c0". subst.
     unfold sig_has_ctr in H1.  simpl in H1. 
     apply H4.  eapply sig_lookup_ctr_some_in; eauto.
   Case "c<>c0". 
     unfold sig_has_ctr in H1.  simpl in H1.  apply IHwf_sig; auto.
 Case "defn_type". 
   unfold sig_has_assn in H0; simpl in H0.
   unfold sig_has_ctr in H1; simpl in H1.
   remember (bundle_defns_ctr_lookup c d) as B.
   destruct B.
   Case "B = Some t".
      symmetry in HeqB.
      eapply (H3 c ( defns_lookup_some_in c d HeqB) (sig_lookup_assn_some_in c S H0) ); eauto.
   Case "B = None". 
     apply IHwf_sig; eauto.
 Case "defn_prop". 
   unfold sig_has_assn in H0; simpl in H0.
   unfold sig_has_ctr in H1; simpl in H1.
   remember (bundle_defns_ctr_lookup c d) as B.
   destruct B.
   Case "B = Some t".
      symmetry in HeqB.
      eapply (H3 c ( defns_lookup_some_in c d HeqB) (sig_lookup_assn_some_in c S H0) ); eauto.
   Case "B = None". 
     apply IHwf_sig; eauto.
Qed.
 
Lemma step_mat_ctr_preserve_type: forall S E v v_tp b b_tp c n c_tp e m args args1 args2 args3,
 wf_sig S ->
 wf_tm S S E v v_tp -> 
 wf_tm S S E b b_tp ->
 sig_has_ctr  (c, n) c_tp S ->
 step_matches_ctr v (c,n) b (e,m, args1, args3) ->
 step_mat_ctr_type_inv c_tp b_tp args n ->
 args = args1++args2 ->
 (all_tm_value args3 /\ 
  all_tm_wf S args3 /\
 exists e_tp, (wf_brn_p S v_tp args1 args3 c_tp b_tp e_tp 
               /\ wf_tm S S E e e_tp 
               /\ step_mat_ctr_type_inv v_tp e_tp args2 m)).
Proof.
intros S E v v_tp b b_tp c n c_tp e m args args1 args2 args3 
         Hwf_sig Hwf_v Hwf_b Hwf_c Hstep  Hinv_cb Hargs.
gen m e args1 args2 args3.
induction Hwf_v; intros m res ags1 ags2 Hargs  ags3 Hstep; try solve [inversion Hstep]...

Case "base".
inversion Hstep; subst.
split. apply all_tm_value_nil.
split. apply all_tm_wf_nil.
exists b_tp.
assert (J1:= wf_ctr_unique Hwf_sig Hwf_c H0). subst. 
split. apply wf_brn_p_base. 
split. auto.
replace (nil++ags2) with ags2 in Hinv_cb. 
auto. 
rewrite app_nil_eq; auto.

Case "assn".
inversion Hstep; subst.
assert(J:= ctr_decl_unique Hwf_sig H0 Hwf_c); contradiction.

Case "ind". 
clear IHHwf_v2.
inversion Hstep; subst.
Case "param".
split. apply all_tm_value_nil.
split. apply all_tm_wf_nil.
assert(Jl1 : (args0 ++ e2::nil) ++ ags2 = (args0 ++ (e2::ags2))). 
rewrite app_ass;
replace ((e2::nil)++ags2) with (e2::ags2); auto.

assert(J:= IHHwf_v1 Hwf_b m0 res args0 (e2::ags2) Jl1 nil H9).
destruct J as [Hv [Hwf_args [e_tp [ Hwf_e [ Hwf_res Htinv]]]]].
exists e_tp. split.
eapply wf_brn_p_inv1;  eauto.

split. auto.

inversion Htinv; subst.
replace (n0 +1 -1)  with n0.
apply H7. 
omega.

Case "arg".
assert(J1: (ags1++ags2 = ags1++ags2)). auto.
assert(J:= IHHwf_v1 Hwf_b 0 e ags1 ags2 J1 args1 H9).
destruct J as [Hv [Hwf_args [e_tp [ Hwf_e [ Hwf_res Htinv]]]]].
clear J1 IHHwf_v1.
split. apply (all_tm_value_app Hv H5).
split. apply (all_tm_wf_app Hwf_args Hwf_v2).


(* need to know that e_tp = t2 -> ? *)
inversion Htinv; subst.  inversion H0.

exists (e_tp0 ^^e2).
split. 
eapply wf_brn_p_inv2; eauto.
eapply (proj1 (wf_tm_branches_lc_mutual S S)); eauto.
split. 

(* Todo*)
assert (Jext: sig_extend S S). unfold sig_extend; auto.
assert(J:= wf_tm_tm Hwf_sig Hwf_res Jext).
destruct J.  inversion H0. destruct H0 as [k Hwf].

inversion Hwf; subst.

pick_fresh z.
eapply wf_tm_app; eauto. (*
rewrite (@subst_intro z e_tp0 e2).
eapply simple_tm_substitution; eauto. auto.
auto. eapply (proj1 (wf_tm_branches_lc_mutual S)); eauto.*)

pick_fresh z.
rewrite (@subst_intro z u e2).
rewrite (@subst_intro z e_tp0 e2).
assert (J: nil = (List.map(subst_tm z e2) nil)). auto.
rewrite J.
eapply subst_step_mat_ctr_type_inv; eauto.
eapply (proj1 (wf_tm_branches_lc_mutual S S)); eauto.
auto.  
eapply (proj1 (wf_tm_branches_lc_mutual S S)); eauto.
auto.
eapply (proj1 (wf_tm_branches_lc_mutual S S)); eauto.
assert (J: n0<n0+1). omega.
contradiction. 


Case "inversions". 
inversion Hstep. inversion H9. inversion H18. inversion H20. 

inversion H9. inversion H19. inversion H18. 

inversion Hstep. inversion H10. inversion H19. inversion H21.
inversion H10. inversion H20. inversion H19. 

inversion Hstep. inversion H8. inversion H8. 
inversion Hstep. inversion H8. inversion H8. 

inversion Hstep. inversion H9. inversion H18. inversion H20.
inversion H9. inversion H19. inversion H18. 

inversion Hstep. inversion H8. inversion H8. 
inversion Hstep. inversion H13. inversion H22. inversion H31. inversion H40. inversion H42. inversion H33.
inversion H43. inversion H42. inversion H24. inversion H34. inversion H43. inversion H46. inversion H33. inversion H43.
inversion H42. inversion H13. inversion H23. inversion H32. inversion H41. inversion H43. inversion H34. inversion H44. inversion H43.
inversion H22. inversion H32. inversion H41. inversion  H43. inversion H31. inversion H41. inversion H40. 

inversion Hstep. inversion H9. inversion H18. inversion H20. inversion H9. inversion H19. inversion H18. 

inversion Hstep. inversion H9. inversion H18. inversion H20. inversion H9. inversion H19. inversion H18. 

Qed.


Lemma is_tm_arr_open_aux: forall u  k e,
is_tm_arr u -> is_tm_arr (open_tm_aux k e u).
Proof.
intros; induction H; simpl; apply is_tm_arr_arr.
Qed. 

Lemma is_tm_arr_open: forall u e,
is_tm_arr u -> is_tm_arr (u^^e).
Proof.
intros. apply (is_tm_arr_open_aux 0 e H).
Qed. 

Lemma open_fully_applied_args_aux: forall c ctr args v m,
fully_applied_args c ctr args  ->
fully_applied_args (open_tm_aux m v c) ctr (List.map (open_tm_aux m v) args).
Proof.
introv H.
induction H; subst; try solve [simpl; auto].
Qed.


Lemma fully_applied_args_not_is_tm_arr: forall s c args,
 fully_applied_args s c args -> ~is_tm_arr s.
Proof.
intros. induction H; unfold not; intro.
inversion H.  inversion H0.
Qed.

Lemma open_level1: forall v t  n  k c_inst ctr0 args,
has_level t n ->
constructs t c_inst ->
fully_applied_args c_inst ctr0 args ->
exists c_inst', exists args', 
 constructs (open_tm_aux k v t) c_inst' /\ 
 fully_applied_args c_inst' ctr0 args' /\ has_level (open_tm_aux k v t)  n.
Proof.
 intros.
gen n k .
induction H0; subst; intros.
Case "base".
inversion H0; subst.  
exists (open_tm_aux k v s) (List.map (open_tm_aux k v) args).
assert (J1:= open_fully_applied_args_aux v k  H1).
assert (J2:= fully_applied_args_not_is_tm_arr J1).
repeat split; auto.

assert (J:= is_tm_arr_arr t u); contradiction.

Case "arr".
inversion H; subst.
assert (J:= is_tm_arr_arr t1 t2); contradiction.
destruct (IHconstructs H1 n0 H5 (1+k) ) as  [c_inst'[args' [Jcon [Jfully_app Jlevel]]]].
exists c_inst' args'; repeat split; simpl; auto.
replace (S n0) with (1+n0); auto.
Qed.

Lemma not_is_tm_arr_arr_false: forall x y,
~(is_tm_arr (tm_arr x y)) -> False.
Proof.
intros; assert (J:= is_tm_arr_arr x y); contradiction.
Qed.

Lemma has_level_unique: forall t n m,
  has_level t n -> has_level t m -> n=m.
Proof.
 intros. 
 gen m. 
 induction H. 
  intros.
  inversion H0; auto. rewrite <- H2 in H. assert (J:= is_tm_arr_arr t u0); contradiction.
 intros. 
  inversion H0.   assert (J:= is_tm_arr_arr t u); contradiction.
  assert (J:= IHhas_level n0 H4). omega.
Qed.

Lemma step_mat_ctr_preserve_level: forall S E v v_tp b c n c_tp ctr0 k l e m args1 args2 nc,
 wf_sig S ->
 sig_has_ctr ctr0 k S ->
 get_ctrs_of ctr0 S  = Some l ->
 In (c,n) l -> 
 wf_tm S S E v v_tp -> 
 sig_has_ctr (c, n) c_tp S ->
 step_matches_ctr  v (c,n) b (e,m, args1, args2) ->
 m>0 ->
 has_level c_tp nc ->
 (exists t_inst, exists ctr, exists args3,
  constructs v_tp t_inst /\ fully_applied_args t_inst ctr args3 /\has_level v_tp (nc  -n + m)).
Proof.
intros S E v v_tp b c n c_tp ctr0 k l e m args1 args2 nc Hwf_sig Hctr Hget_ctr Hin 
        Hwf_v Hwf_c Hstep Hm Hlevel.
gen e m args1 args2.
induction Hwf_v; intros res m Hm ags1 ags2 Hstep; try solve [inversion Hstep].

Case "base".
inversion Hstep; subst.
assert (J:= wf_ctr_unique Hwf_sig H0 Hwf_c).
assert (J1:= sig_has_ctr_strong_typing_inv Hwf_sig Hctr  Hget_ctr Hin  H0).
subst. firstorder. assert (Jeq:= has_level_unique H5 Hlevel).
subst.  
replace (nc - (length x0) + (length x0)) with nc; auto. 
exists x ctr0 x0; repeat split; auto.
omega.

Case "assn". 
inversion Hstep; subst.
assert (J:= ctr_decl_unique Hwf_sig H0 Hwf_c); contradiction.

Case "inductive".
inversion Hstep; subst.
Case "param".
clear IHHwf_v2.

assert (J:= IHHwf_v1 res m0 H10 args nil H9).
destruct J as [t_inst' [ctr' [args3' [Jcon [Jfully_app Jlevel]]]]].
inversion Jcon; subst.
assert (J1:= not_is_tm_arr_arr_false H0); contradiction.
inversion Jlevel; subst.  assert (J1:= not_is_tm_arr_arr_false H2); contradiction.
assert (J1:= open_level1 e2 0 H1 H3 Jfully_app).
destruct J1 as [t_inst1 [args31 [Jcon1 [Jfully_app1 Jlevel1]]]].
replace (nc -n + (m0-1)) with n0; auto.
exists t_inst1 ctr' args31; repeat split; auto.
omega.

inversion Hm.

(* many inversions *)

inversion Hstep. 
inversion H9. inversion H18. inversion H20. 

inversion H9. inversion H19. inversion H18. 

inversion Hstep. inversion H10. inversion H19. inversion H21.
inversion H10. inversion H20. inversion H19. 

inversion Hstep. inversion H8. inversion H8. 
inversion Hstep. inversion H8. inversion H8. 

inversion Hstep. inversion H9. inversion H18. inversion H20.
inversion H9. inversion H19. inversion H18. 

inversion Hstep. inversion H8. inversion H8. 
inversion Hstep.   inversion H13. inversion H22. inversion H31. inversion H40. inversion H42. inversion H33.
inversion H43. inversion H42. inversion H24. inversion H34. inversion H43. inversion H46. inversion H33. inversion H43.
inversion H42. inversion H13. inversion H23. inversion H32. inversion H41. inversion H43. inversion H34. inversion H44. inversion H43.
inversion H22. inversion H32. inversion H41. inversion  H43. inversion H31. inversion H41. inversion H40. 

inversion Hstep. inversion H9. inversion H18. inversion H20. inversion H9. inversion H19. inversion H18. 

inversion Hstep. inversion H9. inversion H18. inversion H20. inversion H9. inversion H19. inversion H18. 

Qed.

Lemma wf_brn_gives_ctr_type_inv: forall S E v v_tp args c_tp b_tp r_tp n ctr0 args1,
  wf_sig S ->
  wf_tm S S E v v_tp ->
  wf_brn S S v_tp args c_tp b_tp r_tp ->
  fully_applied_args v_tp ctr0 args1 ->
  n = List.length args ->
  step_mat_ctr_type_inv c_tp b_tp args n.
Proof. 
introv Hwf_sig Hwf_v Hwf_brn  Harr Hn.
gen n.
induction Hwf_brn.
intros. simpl in Hn.
subst. eapply step_mat_ctr_type_inv_base; eauto.

intros. simpl in Hn; subst.
gather_atoms' as L'.
eapply step_mat_ctr_type_inv_rec_arg with (L:=L'); auto.
intros. 
apply (H0 x).
unfold L' in H1; auto.
unfold L' in H1; auto. auto. auto. auto.

intros. 
replace n with (length args + 1).
apply step_mat_ctr_type_inv_rec_param; auto.
intro. inversion H.
simpl in Hn.  rewrite Hn.  
omega.
Qed.

Lemma wf_brn_p_gives_wf_brn: forall S s args args1 c_tp   
  b_tp  r_tp r_tp1 ctr0 args2 ,
  wf_sig S ->
  wf_brn_p S s args args1 c_tp b_tp r_tp1 ->
  fully_applied_args s ctr0 args2 ->
  all_tm_value args1 ->
  all_tm_wf S args1 ->
  wf_brn S S s args c_tp b_tp r_tp ->
  r_tp = r_tp1.
Proof.
introv Hwf_sig Hwf_p Harr Hv Hwf_args. intro Hwf.
induction Hwf_p; subst.

inversion Hwf; subst.
auto.
inversion Harr.

inversion Hwf. 
rewrite H1 in Harr.  inversion Harr. 
subst.
pick_fresh z.
assert(J1: z \notin L). auto.
assert(J2: z \notin fv r_tp). auto.
assert (Jwf:= H3 z J1 J2).
inversion Hwf_args.
inversion Hv.

assert(Jwf' : wf_brn S S s nil (T^^a) (U ^^a) r_tp).

rewrite (@subst_intro z).
rewrite (@subst_intro z U).
replace (@nil tm) with (List.map (subst_tm z a) nil).
replace s with ([z~>a]s).
replace r_tp with ([z~> a]r_tp).
eapply (proj1 (proj2 (proj2 (substitution S Hwf_sig)))); eauto. 
 eapply (proj1 (wf_tm_branches_lc_mutual S S)); eauto.
apply subst_fresh; auto. apply subst_fresh; auto.
auto. auto.
 eapply (proj1 (wf_tm_branches_lc_mutual S S)); eauto.
auto. 
 eapply (proj1 (wf_tm_branches_lc_mutual S S)); eauto.

apply(IHHwf_p Harr  H8 H4 Jwf').

inversion Hwf; subst.
apply(IHHwf_p Harr Hv Hwf_args H6).
Qed.

Lemma constructs_lc: forall u c n e,
constructs u c ->
lc c ->
constructs (open_tm_aux n e u) c.
Proof.
intros u c n e Hcon Hlc.
gen n.
induction Hcon; subst.
intros. rewrite <- (open_lc_id e Hlc). auto.

intros.
simpl;  apply constructs_arr.
apply IHHcon; auto.
Qed.

Lemma lc_non_bv_fully_applied_args: forall c ctr args,
fully_applied_args c ctr args ->
all_tm_lc args ->
lc c.
Proof.
introv Hfully_app Hnon_bv.
induction Hfully_app; auto.
inversion Hnon_bv; subst.
apply lc_app; auto.
Qed.

Lemma open_construct1: forall u c ctr args e,
constructs u c ->
fully_applied_args c ctr args ->
all_tm_lc args ->
(*value e ->*)
constructs (u^^e) c.
Proof. 
introv Hcon Hfully_app Hnon_bv.

assert(Jlc:= lc_non_bv_fully_applied_args Hfully_app Hnon_bv).
unfold open_tm.
apply (constructs_lc 0 e  Hcon Jlc).
Qed.
Lemma open_id_lc_list: forall args k e,
all_tm_lc args ->
List.map (open_tm_aux k e) args = args.
Proof.
intros args k e H.
induction args; subst.
simpl; auto.
simpl.
inversion H; subst.
assert(J:= IHargs H2). rewrite J.
rewrite <- (open_lc_id); auto.
Qed.

Lemma all_tm_bv_eq: forall n b l,
all_tm_bv n (b::l) -> b = tm_bv n.
Proof.
intros n b l H.
remember (b::l) as l1.
destruct H; subst; inversion Heql1; auto.
Qed.

Lemma open_bv_list_aux: forall n l b e,
all_tm_bv n (l++b::nil) ->
List.map (open_tm_aux (n+length l) e) (l++b::nil) 
  = l++e::nil.
Proof.
intros n l b e Hb.
gen n.
induction l; subst.
intros n H.
rewrite app_nil_eq in H; rewrite app_nil_eq; simpl.
rewrite (all_tm_bv_eq H).
replace (n+0) with n; simpl ; auto.
destruct (n===n); auto. 
assert(J:n=n); trivial; contradiction.

intros n H.
rewrite <- app_comm_cons in H.
inversion H; subst.
assert(J1:= IHl (n+1) H2).
simpl. destruct(n===n+S(length l)). 
assert (J:~(n=n+S(length l))). omega. contradiction. 
replace  (n+ S(length l)) with (n+ 1+length l).
rewrite J1; auto. omega.
Qed.


Lemma open_construct2: forall u c ctr args  nu na bargs args0 m bargs0 barg e,
constructs u c ->
fully_applied_args c ctr args ->
has_level u nu ->
args = bargs ++ args0 ->
all_tm_bv na bargs ->
List.length bargs = m ->
bargs = bargs0++barg::nil ->
(* value e ->*)
all_tm_lc args0 ->
~(na + m < nu+1) ->
exists c', exists args',
(constructs (open_tm_aux (na + m - nu-1) e u) c' /\
 fully_applied_args c' ctr args' /\
 args' = bargs0++ e::args0).
Proof.
intros u c ctr0 args nu na bargs args0 m bargs0 barg e
          Hcon Hfully_app Hlevel Harg Hbv Hlen Hbarg (* Hvalue*) Hlc Homega.
gen nu.
induction Hcon.

Case "~arr s".
inversion Hfully_app; subst.
replace ((bargs0 ++ barg::nil) ++args0)  with (bargs0++barg::args0) in H2.
assert (J:= app_cons_not_nil bargs0 args0 barg); contradiction.
rewrite (app_ass). rewrite <- app_comm_cons.
rewrite app_nil_eq; auto.

destruct bargs0; subst.
rewrite app_nil_eq in Hbv.
rewrite app_nil_eq in H3.
rewrite <- app_comm_cons in H3.
rewrite app_nil_eq in H3.
inversion H3; subst.
rewrite app_nil_eq in Hfully_app.
rewrite <- app_comm_cons in Hfully_app.
rewrite app_nil_eq in Hfully_app.
clear H3.
intros nu Hlevel Homega.
inversion Hlevel; subst.
exists (open_tm_aux na e (tm_app t barg)).
exists (List.map (open_tm_aux na e) (barg::args0)).
split; auto. rewrite app_nil_eq. replace (na +length(barg::nil) - 0 - 1) with na.
apply constructs_base. 
unfold not; intro. inversion H2. simpl. omega. 
split.  apply open_fully_applied_args_aux; auto.
rewrite app_nil_eq.
simpl. 
assert(J: List.map (open_tm_aux na e) args0 = args0).
apply open_id_lc_list; auto.
assert(J1: barg = tm_bv na).
eapply all_tm_bv_eq; eauto.
rewrite J1; simpl. 
destruct (na===na).
rewrite J; auto.
assert(J2:na = na); auto.
contradiction.
 
intros nu Hlevel Homega.
inversion Hlevel; subst. clear H1.
exists (open_tm_aux (na + length(t0::bargs0++barg::nil)-1) e (tm_app t tn)).
exists (List.map (open_tm_aux (na + length(t0::bargs0++barg::nil)-1) e ) 
           (t0::bargs0++barg::args0)).
assert( J: na +  length((t0::bargs0)++barg::nil) -0 -1 = na + length(t0::bargs0++barg::nil) -1).
rewrite app_comm_cons.
omega.
rewrite J.
split. apply constructs_base. simpl.  intro. inversion H1. 

split.
replace (((t0::bargs0) ++barg::nil)++args0) 
 with (t0::bargs0++barg::args0) in Hfully_app. 
apply open_fully_applied_args_aux; auto. 
rewrite  (app_ass (t0::bargs0) (barg::nil) args0).
auto.

replace (t0::bargs0++barg::args0) 
 with ((t0::bargs0 ++barg::nil)++args0).


replace (List.map (open_tm_aux (na+length(t0::bargs0++barg::nil)-1) e)
              ((t0::bargs0++barg::nil)++args0))
with ( (List.map (open_tm_aux (na+length(t0::bargs0++barg::nil)-1) e)
              ((t0::bargs0++barg::nil)) 
              ++   (List.map (open_tm_aux (na+length(t0::bargs0++barg::nil)-1) e) 
                      args0))).

replace (List.map (open_tm_aux (na+length(t0::bargs0++barg::nil)-1) e)
              (t0::bargs0++barg::nil))
with (t0::bargs0++e::nil).
replace (List.map (open_tm_aux (na+length(t0::bargs0++barg::nil)-1) e)
              args0) with args0.
rewrite app_comm_cons;
rewrite  (app_ass (t0::bargs0) (e::nil) args0); auto.
rewrite open_id_lc_list; auto.
replace (na + length(t0::bargs0++barg::nil) -1) with
        (na + length(t0::bargs0)).
rewrite app_comm_cons.
rewrite app_comm_cons.
rewrite (open_bv_list_aux (t0::bargs0) barg e Hbv).
auto. simpl. rewrite app_length. simpl. omega. 
rewrite map_app; auto.
rewrite app_comm_cons;
rewrite  (app_ass (t0::bargs0) (barg::nil) args0); auto.

intros nu Hlevel Homega.
inversion Hlevel; subst.
assert(J:= not_is_tm_arr_arr_false H); contradiction.
assert (Jomega: ~ na+length(bargs0++barg::nil) < n + 1).
omega.
assert (J:= IHHcon Hfully_app n H2 Jomega ).
destruct J as [c' [args' [Hc Happ]]].
exists c' args'. 
split; auto. simpl; apply constructs_arr.
replace (S(na+length (bargs0 ++barg::nil) -S n -1))
 with (na+length(bargs0++barg::nil) - n -1); auto.
remember (length(bargs0++barg::nil)) as m.
omega.
Qed.

Lemma all_tm_lc_app: forall l a,
all_tm_lc l -> lc a -> all_tm_lc (l++a::nil).
Proof.
intros l a Hlc Ha.
induction l; subst.
rewrite app_nil_eq. apply all_tm_lc_cons; auto.
inversion Hlc; subst.
rewrite <- app_comm_cons.
apply all_tm_lc_cons; auto.
Qed.

Lemma all_tm_lc_list_rev: forall l,
all_tm_lc l -> all_tm_lc (rev l).
Proof.  
intros l H. 
induction H; subst.
simpl; apply all_tm_lc_nil.
simpl. 
apply all_tm_lc_app; auto.
Qed.

Lemma list_non_empty_inv: forall A l,
(@length A l) > 0 -> exists l0, exists b, l = l0++b::nil.
Proof. 
intros.
induction l; subst.
simpl in H; inversion H.
destruct l; subst.
exists  (@nil A) a; auto.
assert (J: length(a0::l) >0).
simpl. omega. 
destruct (IHl J) as [l1 [b J1]].
exists (a::l1) b. rewrite J1; auto.
Qed.

Lemma all_tm_bv_app: forall n l b,
all_tm_bv n (l++b::nil) -> all_tm_bv n l.
Proof.
intros n l b Hbv.
gen n.
induction l; subst.
intros n H.
apply all_tm_bv_nil.
intros n H.
rewrite <- app_comm_cons in H.
inversion H; subst.
apply all_tm_bv_cons; auto.
Qed.



Lemma wf_brn_arg_unique: forall S E v t c n c_tp ctr0 k l nc na body e m params ags,
wf_sig S ->
wf_tm S S E v t ->
sig_has_ctr ctr0 k S ->
get_ctrs_of  ctr0 S = Some l ->
 In (c, n) l ->
sig_has_ctr (c,n) c_tp S ->
step_matches_ctr v (c, n) body (e, m, params, ags) ->
has_level c_tp nc ->
 na = nc - n ->
exists c_inst, exists args, exists bargs, 
  (constructs t c_inst /\      (* t constructs an instance c_inst *)
  fully_applied_args c_inst ctr0 args  /\  (*   that is fully applied [c] of kind [k] *)
  args = bargs++(List.rev params)  /\ 
  List.length bargs = m /\ 
  all_tm_bv na bargs /\ 
  all_tm_lc params).

Proof. 
intros S E v t c n c_tp ctr0 k l nc na body e m params ags
          Hwf_sig  Hwf Hctr Hget_ctr Hin Hc Hstep H_level Hna.
gen e m params ags.
induction Hwf; intros res m params ags Hstep; try solve [inversion Hstep]...

Case "base".
inversion Hstep; subst.
assert(J:= sig_has_ctr_strong_typing_inv Hwf_sig Hctr Hget_ctr Hin H0).
firstorder.
assert (J:= wf_ctr_unique Hwf_sig Hc H0).  rewrite J in H_level. 
assert (J1:= has_level_unique H_level  H5).   rewrite J1.
exists x x0 x0. repeat split; auto. simpl. apply app_nil_end.


Case "assn".
inversion Hstep; subst.
assert (J:= ctr_decl_unique Hwf_sig H0 Hc); contradiction.

Case "inductive".
inversion Hstep; subst.
Case "datatype parameter".
clear IHHwf2.
assert (J := IHHwf1 res m0 args nil H9).
destruct J as [c_inst [args0 [bargs J1]]].
destruct J1 as [Hcon [Hfully_ap [Heq [Hlen [Hbv Hlc]]]]].
clear IHHwf1.
assert(J:= step_mat_ctr_preserve_level Hwf_sig  Hctr Hget_ctr Hin Hwf1 Hc H9 H10 H_level). 
destruct J as [t_inst [ctr1 [args3  [Jcon1 [Jfully_app1 Jlevel]]]]].
inversion Jlevel; subst. 
assert(J:= not_is_tm_arr_arr_false H2); contradiction.
inversion Hcon; subst.
assert(J:= not_is_tm_arr_arr_false H0); contradiction.
clear Hstep.
assert (Jeq : bargs++rev args = bargs ++rev args); auto.
assert (Jlc:= all_tm_lc_list_rev Hlc).
assert (Jlen: length bargs = length bargs); auto.
assert (Jb:= @ list_non_empty_inv tm bargs H10).
destruct Jb as [bargs0 [barg Jb]].
assert (Jomega: ~nc- n + length bargs < n0 + 1). omega.
assert (J:= open_construct2 bargs0 barg e2 H5 Hfully_ap H1 Jeq Hbv Jlen Jb Jlc Jomega).
destruct J as [c' [args' [Jcon [Jfully_app Jeq1]]]].
rewrite <- H3 in Jcon. replace (Datatypes.S n0 -n0 -1) with 0 in Jcon.
exists c' args' bargs0; repeat split; auto.
rewrite rev_unit; auto.
rewrite Jb; rewrite app_length; simpl. omega.  
rewrite Jb in Hbv. 
eapply all_tm_bv_app; eauto.
apply all_tm_lc_app; auto.
eapply (proj1 (wf_tm_branches_lc_mutual S S)); eauto.
omega.

Case "arguments".
clear IHHwf2.
assert (J := IHHwf1 e 0 params args1 H9).
destruct J as [c_inst [args0 [bargs J1]]].
destruct J1 as [Hcon [Hfully_ap [Heq [Hlen [Hbv Hlc]]]]].
clear IHHwf1.
exists c_inst  args0 (nil:list tm).
destruct bargs; subst.
repeat split; auto. 
unfold open_tm. 
eapply open_construct1; eauto.
inversion Hcon; auto.
assert (J:= not_is_tm_arr_arr_false H0); contradiction.
rewrite app_nil_eq. apply all_tm_lc_list_rev; auto.
simpl in Hlen; inversion Hlen.

inversion Hstep. inversion H9. inversion H18. inversion H20. 

inversion H9. inversion H19. inversion H18. 

inversion Hstep. inversion H10. inversion H19. inversion H21.
inversion H10. inversion H20. inversion H19. 

inversion Hstep. inversion H8. inversion H8. 
inversion Hstep. inversion H8. inversion H8. 

inversion Hstep. inversion H9. inversion H18. inversion H20.
inversion H9. inversion H19. inversion H18. 

inversion Hstep. inversion H8. inversion H8. 
inversion Hstep. inversion H12.  inversion H13. inversion H23. inversion H32. inversion H41. inversion H43. inversion H34.
inversion H44. inversion H43. inversion H25. inversion H35. inversion H44. inversion H47. inversion H34. inversion H44.
inversion H43. inversion H13. inversion H23. inversion H32. inversion H41. inversion H43. inversion H34. inversion H44. inversion H43.
inversion H22. inversion H32. inversion H41. inversion  H43. inversion H31. inversion H41. inversion H40. 

inversion Hstep. inversion H9. inversion H18. inversion H20. inversion H9. inversion H19. inversion H18. 

inversion Hstep. inversion H9. inversion H18. inversion H20. inversion H9. inversion H19. inversion H18. 

Qed.


Lemma list_rev_eq: forall  A l1 l2, @List.rev A l1 = List.rev l2 -> l1 = l2.
Proof.
intros. gen l2.
induction l1.
intros. simpl in H. destruct l2. auto. simpl in H.
assert (J:= app_cons_not_nil (rev l2) nil a); contradiction.
intros. destruct l2. 
simpl in H.  
assert (J:= app_cons_not_nil (rev l1) nil a).  rewrite  H in J. assert (J1: @nil A = nil); auto. contradiction.
simpl in H.  assert(J:= app_inj_tail (rev l1) (rev l2) a a0 H).
destruct J. assert (J:= IHl1 l2 H0). subst; auto.
Qed.


Lemma eq_arg_length: forall s1 s2 ctr0 args1 args2 k,
 fully_applied_args s1 ctr0 args1 ->
 fully_applied_typ s1 ctr0 k ->
 fully_applied_args s2 ctr0 args2 ->
 fully_applied_typ s2 ctr0 k ->
 length args1 = length args2.
Proof.
introv Hfully_app_args1  Hfully_app_typ1 Hfully_app_args2  Hfully_app_typ2.
gen s2 args2 k.
induction Hfully_app_args1; subst; intros.
inversion Hfully_app_typ1; subst.
inversion Hfully_app_typ2; subst.
inversion Hfully_app_args2; subst; auto.

inversion Hfully_app_typ1; subst.
inversion Hfully_app_typ2; subst.
inversion Hfully_app_args2; subst; auto.

inversion Hfully_app_typ1; subst.
inversion Hfully_app_typ2; subst.
inversion Hfully_app_args2; subst.
assert (J: length args = length args0).
eapply IHHfully_app_args1 ; eauto.
simpl; subst; auto.
Qed.


Lemma step_mat_brn_preserve_type: forall S E s args branches result_typ e v ctr0 k ctrl l l0,
  wf_sig S ->
  wf_tm S S E v s ->
  wf_branches S S E s args branches result_typ ->
  sig_has_ctr ctr0 k S ->
  get_ctrs_of ctr0 S  = Some l ->
  l = l0++ ctrl -> 
  branches_cover_aux ctrl branches ->
  fully_applied_args  s ctr0 (List.rev args) ->
  fully_applied_typ s ctr0 k ->
 step_matches_branches v branches e ->
 wf_tm S S E e result_typ.
Proof.  
intros S E s args branches result_typ e v ctr0 k ctrl l l0 Hwf_sig 
         Hwf_v (*Harr*) Hwf_branches Hsig_has Hctr HctrL Hbrn_cover Hfully_app_args Hfully_app_typ H_step. 
gen e l0 ctrl. induction Hwf_branches; intros e H_step.

 inversion H_step.

inversion H_step; subst.
Case "step_mathces_here".
 intros l0 ctrl HctrL Hbrn_cover.
assert(Jcon: constructs s s). 
apply constructs_base. intro. inversion H2; subst. inversion Hfully_app_args. 
destruct ectr; subst. 
inversion Hbrn_cover; subst.
assert (Jin : In (n,n0) (l0++(n,n0)::l)).  assert (J1:= in_eq (n,n0) l). assert (J3: In (n,n0) l0 \/  In (n,n0) ((n,n0)::l)). right. auto. apply (in_or_app l0 ((n,n0)::l) (n,n0)  J3).
assert (J:= sig_has_ctr_strong_typing_inv Hwf_sig Hsig_has Hctr Jin H).
destruct J as [c_inst [args' [nt  [Jconc [ Jfully_app_args [Jfully_app_typ [Jlen[Jlevel [ Jeq Jbvc ]]]]]]]]].
assert (Jeq_len:= eq_arg_length Jfully_app_args Jfully_app_typ Hfully_app_args Hfully_app_typ).
rewrite  Jlen in Jeq_len. rewrite rev_length in Jeq_len.
assert(Jstep_inv:=  wf_brn_gives_ctr_type_inv Hwf_sig Hwf_v H1 Hfully_app_args Jeq_len).
assert (Jwf_env:= wf_tm_gives_wf_env Hwf_v).
assert (Jwf_c:= wf_tm_ctr Jwf_env H). 
assert (J2: nt - n0 = nt -n0). auto.
assert(Jarg:= wf_brn_arg_unique Hwf_sig Hwf_v Hsig_has Hctr Jin  H H7 Jlevel J2).
destruct Jarg as [ctr0' [args1'  [bargs[ Jcon' [ Jfully_app1 [ Jeq_args [Jlen_bargs [Jbv Jlc]]]]]]]].
destruct bargs;  inversion Jlen_bargs.
rewrite app_nil_eq in Jeq_args. 
assert(Jeqs := construct_unique Jcon Jcon'). 
rewrite  <- Jeqs in Jfully_app1.
assert(Jeq2:= proj2(fully_applied_args_unique Hfully_app_args Jfully_app1)).
rewrite Jeq_args in Jeq2.  
assert (Jarg:= list_rev_eq args args0 Jeq2).  
assert (Jarg' : args = args0++nil).
rewrite Jarg; rewrite <- app_nil_end; auto. 

assert (J:= step_mat_ctr_preserve_type nil Hwf_sig Hwf_v H0 H H7 Jstep_inv Jarg').
destruct J as [Hv [Harg [e_tp [Hwf_brn_p [Hwf_e Hctr_inv]]]]].
rewrite Jarg in H1.
assert(Jeq_restp := wf_brn_p_gives_wf_brn Hwf_sig Hwf_brn_p Hfully_app_args Hv Harg H1).
rewrite <- Jeq_restp in Hwf_e; auto.

Case "step_matches_earlier".
intros l0 ctrl HctrL Hbrn_cover.
inversion Hbrn_cover; subst.
apply IHHwf_branches with (l2:= l0++ectr::nil) (ctrl :=l1); auto.
rewrite app_ass. rewrite <- app_comm_cons. auto.
Qed.


Lemma values_step_false: 
( forall e, value e  -> forall e', step e e' -> False) 
/\ (forall e, applied_value e  -> forall e', step e e' -> False).
Proof. 
apply mutual_tm_ctr_value_ind; intros; try solve [inversion H].
 
eapply H; eauto.
inversion H0. inversion H4. inversion H8. 
  inversion H7. inversion H3. inversion H8.
inversion H. inversion H3. inversion H7. inversion H6.
 inversion H2. inversion H7.
inversion H. inversion H3. inversion H2. 
inversion H0. inversion H4. inversion H3. eapply H; eauto.   
inversion H. inversion H3. inversion H7. inversion H6. 
  inversion H2. inversion H7.

inversion H. inversion H3. inversion H7. inversion H6. 
 inversion H2. inversion H7.

inversion H1; subst; try solve [inversion a].
inversion a. inversion H4.
inversion a. inversion H4.  inversion H8. inversion H12.
inversion a. inversion H4. inversion H9. inversion H13.
eapply H; eauto.
eapply H0; eauto.
inversion a. inversion H4.
inversion a. inversion H5.
Qed.                


(* TODO: consider beta conversions*)
Lemma preservation: forall S  e t e',
  wf_sig S ->  wf_tm S S empty e t -> 
  step e e' -> wf_tm S S empty e' t.
Proof with simpl_env; eauto.
intros S e t e' Hwf_sig Hwf_e  H_step.
gen e'.
remember (empty:env) as E.
induction Hwf_e; intros e' H_step; subst; try solve [inversion H_step]...

Case "step_app".

inversion H_step; subst. 
inversion Hwf_e1; subst.
pick_fresh x.
 rewrite (@subst_intro x) ; trivial; eauto.
    rewrite (@subst_intro x u) ; trivial; eauto.
 eapply simple_tm_substitution with (u:=t2); eauto.

eapply (proj1 (wf_tm_branches_lc_mutual S S)); eauto.
eapply (proj1 (wf_tm_branches_lc_mutual S S)); eauto.

inversion Hwf_e1; subst.  inversion H5.
inversion Hwf_e1; subst.
inversion Hwf_e1; subst.  inversion H4. inversion H8. inversion H15. 

inversion Hwf_e1.  inversion H5. inversion H12.  inversion H19.

Case "step_app_cong1". 
eapply wf_tm_app; eauto.

Case "step_app_cong2".
destruct H.  
destruct ( (proj1 values_step_false) e2 H e2' H3).

unfold open_tm. rewrite <- open_lc_id; auto.
replace u with (u^^e2').
eapply wf_tm_app; eauto.
(* unfold open_tm in Hwf_e4; rewrite <- open_lc_id in Hwf_e4; auto.*)
unfold open_tm; rewrite <- open_lc_id  ; eauto. 
(*unfold open_tm; rewrite <- open_lc_id; auto.*)

destruct H. destruct ( (proj1 values_step_false) e2 H e2' H4).
unfold open_tm. rewrite <- open_lc_id; auto.
replace u with (u^^e2').
eapply wf_tm_app; eauto.
(*unfold open_tm in Hwf_e4; rewrite <- open_lc_id in Hwf_e4; auto.*)
unfold open_tm; rewrite <- open_lc_id  ; eauto. 
(*unfold open_tm; rewrite <- open_lc_id; auto.*)

inversion Hwf_e1. 
inversion Hwf_e1. inversion H5.

inversion H_step. inversion H1. inversion H7. inversion H11.
(*  inversion H1. inversion H8. inversion H10.  inversion H5. inversion H9.*)

inversion Hwf_e1. inversion H11. inversion H5. inversion H9.
 inversion Hwf_e1. inversion H11.
inversion Hwf_e1. inversion H11. 

Case "PF_SAY".
 inversion H_step. inversion H3. inversion H7. inversion H6. 
 inversion H2. inversion H7.

Case "PF_SIGN".
inversion H_step. inversion H4. inversion H8.
inversion H_step. inversion H7.
 inversion H12. inversion H17.
 inversion H3. inversion H8.

Case "T_PF".
inversion H_step. inversion H2. inversion H1.
inversion H_step. inversion H2. inversion H1.

Case "PF_RETURN".
inversion H_step; subst.
inversion H5. inversion H4.
eapply wf_tm_TM_PF_RETURN; eauto. 

Case "step_pf_bind".
inversion H_step; subst.
 inversion Hwf_e1; subst. inversion H5.

assert (J1:lc (tm_app T_PF Q)).
eapply lc_app; eauto. unfold T_PF; apply lc_con.
rewrite (@open_lc_id (tm_app T_PF Q) p J1 0).
eapply wf_tm_app; eauto.
(*
unfold open_tm. rewrite <- open_lc_id.
eapply wf_tm_T_PF.

assert (H_Q := wf_tm_tm Hwf_sig Hwf_e2).
destruct H_Q. inversion H0.

eapply  tm_T_PF_inv with (P:=P); auto.
eapply lc_app; eauto. unfold T_PF; apply lc_con.
*)
inversion H3. inversion H4. inversion H2.
inversion H2. inversion H3.

Case "step_pf_bind_cong1".
assert (Jeq: @empty bnd = empty); auto.
eapply wf_tm_TM_PF_BIND; eauto.

Case "step_pf_bind_cong2".
assert (Jeq: @empty bnd = empty); auto.
eapply wf_tm_TM_PF_BIND; eauto.

Case "step_say".
inversion H_step; subst.
eapply wf_tm_TM_PF_RETURN; eauto.  

inversion H2. inversion H1.

Case "step_pif".
inversion H_step; subst.

pick_fresh x.
 assert (J:= proj1 (typing_strengthening_eq_mutual S Hwf_sig )).
 rewrite_env ((empty :env)& empty).
 eapply J; eauto. 
 assert (Jx: x\notin L); auto. assert (J2:= H2 x Jx).
 replace (empty & x~bnd_eq v2 v2 T1) with (empty & x~bnd_eq v2 v2 T1& empty) in J2; eauto. 

 auto.
 inversion H7. inversion H8.  inversion H12. inversion H16. inversion H15.
  inversion H11. inversion H16. inversion H6. inversion H12. inversion H16.
  inversion H6. inversion H7. inversion H12. inversion H16.

Case "step_pf_says_bind".
inversion H_step; subst.
inversion H3. inversion H4. inversion H2.
inversion H2. inversion H3.

inversion H_step; subst. inversion H3. inversion H4.
 inversion H2. inversion H2. inversion H3.  

Case "branches_type".
inversion H_step; subst.
assert(Jarg:= rev_involutive args). 
rewrite <- Jarg in H.
unfold branches_cover in H3. 
destruct H3 as [ctrl [Hget_ctrs Hbrn_cover]].
assert(Jctrl: ctrl = nil++ctrl). auto.  
apply (step_mat_brn_preserve_type (@nil ctr) Hwf_sig Hwf_e1 H2 H0 Hget_ctrs Jctrl Hbrn_cover H H1 H8); auto.

Case "step_mat_cong".
eapply wf_tm_mat_TYPE; eauto.

Case "branches_prop".
inversion H_step; subst.

assert(Jarg:= rev_involutive args). 
rewrite <- Jarg in H.
unfold branches_cover in H3. 
destruct H3 as [ctrl [Hget_ctrs Hbrn_cover]].
assert(Jctrl: ctrl = nil++ctrl). auto.  
apply (step_mat_brn_preserve_type (@nil ctr) Hwf_sig Hwf_e1 H2 H0 Hget_ctrs Jctrl Hbrn_cover H H1 H8); auto.

Case "step_mat_cong".
eapply wf_tm_mat_PROP; eauto.

Case "step_cast".
inversion H_step; subst.
rewrite (@typing_conv_eq  t2 t1); auto.
apply conv_sym; auto.

Case "step_cast_cong". 
 eapply wf_tm_cast; eauto.

Qed.

Lemma open_construct: forall v t   k c_inst ctr0 args,
constructs t c_inst ->
fully_applied_args c_inst ctr0 args ->
exists c_inst', exists args', 
 constructs (open_tm_aux k v t) c_inst' /\ 
 fully_applied_args c_inst' ctr0 args' .
Proof.
 intros.
gen k .
induction H; subst; intros.
Case "base". 
exists (open_tm_aux k v s) (List.map (open_tm_aux k v) args).
assert (J1:= open_fully_applied_args_aux v k  H0).
assert (J2:= fully_applied_args_not_is_tm_arr J1).
split; auto.
Case "arr".
destruct (IHconstructs H0 (1+k)) as  [c_inst'[args' [Jcon Jfully_app]]].
exists c_inst' args'; split; simpl; auto.
Qed.

Lemma bundle_has_type_strong_inv: forall P S S' S1 d K c t,
  wf_bundle_tctrs S d K ->
  (K=K_TYPE \/ K=K_PROP) ->
  wf_bundle_ctrs P S' S1 d ->
  defns_has_ctr c t d ->
  wf_tm S S empty t S_KIND 
  \/ exists c_inst,  exists ctr0, exists args,  
      constructs t c_inst
      /\ fully_applied_args c_inst ctr0 args.
Proof. 
introv Hwf_tctr HK Hwf_bundle Hc.
induction d.
Case "nil". inversion Hc.
Case "cons".
  destruct a. destruct p.
  inversion Hc. destruct (eq_ctr_dec c c1).
  Case "c=c1".
  inversion H0; subst.
  inversion Hwf_tctr; auto.
  Case "c<>c1".
   remember (ctr_decls_lookup c c0) as C.
   destruct C.
    Case "C = some t1".
    inversion H0; subst.
     inversion Hwf_bundle; subst.
     symmetry in HeqC.
     right. destruct c. assert (J:= ctr_decls_has_strong_type_inv  H8 HeqC).
     destruct J as [c_inst' [args' [nt [Jlv [Jcon [Jfully_app Jx]]]]]].
     exists c_inst'  c1 args'; repeat split; auto.
    Case "C=None".
      inversion Hwf_tctr; subst.
      inversion Hwf_bundle; subst.
      eapply IHd; eauto.
Qed.

Lemma sig_has_ctr_type_strong_inv: forall S c u,
  wf_sig S -> 
  sig_has_ctr c u S -> 
  wf_tm S S empty u S_KIND 
  \/ exists c_inst, 
       exists ctr0, 
         exists args, 
            constructs u c_inst
           /\ fully_applied_args c_inst ctr0 args .
Proof.
 intros S c u Hwf_sig Hc.
 induction S.
 Case "nil".  inversion Hc.
 Case "cons".
 destruct a.
 Case "bundle_defns".
 inversion Hc; subst.
 remember (bundle_defns_ctr_lookup c d) as B.
 destruct B.
    Case "B = Some t".
    inversion H0; subst.
    inversion Hwf_sig; subst.
    Case "wf_bundle TYPE". 
    symmetry in HeqB. 
     assert (J:   wf_tm S S empty u S_KIND 
     \/ exists c_inst, exists ctr0, exists args, 
           constructs u c_inst  /\ fully_applied_args c_inst ctr0 args).
    eapply bundle_has_type_strong_inv with (c:=c) (S:=S); eauto. 
    destruct J; auto.
    left.  replace (bundle_defns d::S) with (bundle_defns d :: nil++S); auto. 
   eapply wf_tm_weakening_sig2_r with (S:=bundle_defns d::nil) (S1:=S) (S2:=S) ; eauto.
   unfold sig_extend; auto.
    unfold dom_sig; simpl. rewrite <- app_nil_end; auto. 

    Case "wf_bundle PROP".
      symmetry in HeqB.
      assert (J:   wf_tm S S empty u S_KIND 
      \/ exists c_inst, exists ctr0, exists args, 
      constructs u c_inst  /\ fully_applied_args c_inst ctr0 args).
      eapply bundle_has_type_strong_inv with (c:=c) (S:=S); eauto. 
      destruct J; auto.
      left.  replace (bundle_defns d::S) with (bundle_defns d :: nil++S); auto. 
   eapply wf_tm_weakening_sig2_r with (S:=bundle_defns d::nil) (S1:=S) (S2:=S) ; eauto.
   unfold sig_extend; auto.
    unfold dom_sig; simpl. rewrite <- app_nil_end; auto. 

   Case "B = None".
    inversion Hwf_sig; subst.
    Case "TYPE".
    destruct IHS; auto.
    left.      eapply wf_tm_weakening_sig2_r with (S:=bundle_defns d::nil) (S1:=S) (S2:=S) ; eauto.
   unfold sig_extend; auto.
    unfold dom_sig; simpl. rewrite <- app_nil_end; auto. 

    Case "PROP".
    destruct IHS; auto.
    left.  eapply wf_tm_weakening_sig2_r with (S:=bundle_defns d::nil) (S1:=S) (S2:=S) ; eauto.
   unfold sig_extend; auto.
    unfold dom_sig; simpl. rewrite <- app_nil_end; auto. 

 Case "bundle_assert".
 inversion Hwf_sig; subst.
 inversion Hc; subst.
 destruct IHS; auto.
 left.   replace (bundle_assert c0 t ::S) with (bundle_assert c0 t :: nil++S); auto. 
eapply wf_tm_weakening_sig2_r with (S2 :=S) (S1:= S) (S:=bundle_assert c0 t::nil) ; eauto.
unfold sig_extend; auto.
 unfold dom_sig; simpl.  unfold disjoint.  intros.
 destruct (in_inv H1). subst. apply H5; auto. inversion H6.
Qed.

Lemma applied_value_inv: forall S E e u,
  wf_sig S -> 
  wf_env S S E ->
  applied_value  e -> wf_tm S S E e u ->
  (wf_tm S S E u S_KIND 
   \/ exists c_inst, 
       exists ctr0, 
         exists args, 
           constructs u c_inst
           /\ fully_applied_args c_inst ctr0 args ).
Proof.
 introv Hwf_sig Hwf_env Hvalue Hwf_e.
 gen u.
 induction Hvalue; intros; subst.
 Case "ctr".
  inversion Hwf_e; subst. 
  assert (J: wf_tm S S empty u S_KIND \/ 
                     exists c_inst, 
                     exists ctr0, 
                     exists args, 
                     constructs u c_inst
                     /\ fully_applied_args c_inst ctr0 args ).
  eapply sig_has_ctr_type_strong_inv; eauto.
  destruct J; eauto.
  left. rewrite_env(empty&E).
  eapply typing_weakening'; eauto.
  rewrite_env E ;auto.

   Case "assn".
   left.  assert (J: constructs u K_PROP /\ wf_tm S S empty u S_KIND). eapply sig_has_assn_has_typ_inv; eauto.
   unfold sig_extend; auto. destruct J; auto.
   rewrite_env(empty&E).
   eapply typing_weakening'; eauto.
   rewrite_env E ;auto.

 Case "app".
  inversion Hwf_e; subst; try solve [inversion Hvalue; auto].
  destruct (IHHvalue (tm_arr t0 u0)); auto.
    Case "t0->u0 S_KIND".
    left. inversion H0; subst.
    pick_fresh z. 
    assert (J: wf_tm S S (E & z~:t0) (u0^^ tm_fv z) S_KIND).
    eapply (H8  z); eauto.
    rewrite (@subst_intro z); auto.
    replace S_KIND  with ([z~>t2]S_KIND); auto.
    eapply simple_tm_substitution; eauto.
    eapply (proj1 (wf_tm_branches_lc_mutual S S)); eauto.
    Case "inv".
    right. destruct H0 as [c_inst [ctr0 [args [Hcon Hfully_app]]]].
    inversion Hcon; subst. 
    assert (J:= is_tm_arr_arr t0 u0); contradiction. 
    assert (J:= open_construct t2 0 H5 Hfully_app).
    destruct J as [c_inst' [args'  [Jcon Jfully_app]]].
    exists c_inst' ctr0 args'; split; auto.

 Case "inversion ".
  inversion Hvalue. inversion H4.
  inversion Hvalue. inversion H3.
  inversion Hvalue. inversion H8. inversion H14. inversion H18.
  inversion Hvalue. inversion H4.
  inversion Hvalue. inversion H3.
Qed.


Lemma open_tm_eq : forall x n t1 t2,
x \notin (fv t1 \u fv t2) -> 
open_tm_aux n (tm_fv x) t1 = open_tm_aux n (tm_fv x) t2 -> 
t1 = t2.
Proof.
 intros x n t1 t2 Hin Heq.
 gen n t2.
 induction t1; subst; intros; auto.

Case "fv".
  simpl in Heq. 
  destruct t2; subst; inversion Heq; auto.
  destruct (b===n);auto.
  inversion H0; subst. simpl in Hin.
  assert (J:= in_same x).
  destruct (in_union x {{x}} {}).
  assert (J1: x\in {{x}}\u {}). apply H1; auto.
  contradiction.

Case "bv".
  simpl in Heq.
  destruct (b===n); auto.
  Case "b===n".
  destruct t2; subst; inversion Heq.
  simpl in Hin.   subst. 
  assert (J:= in_same a).
  destruct (in_union a {} {{a}}).
  assert (J1:  a \in {}\u {{a}}). apply H0; auto.
  contradiction.
  destruct (b0===n).
  subst; auto.
  inversion H0.

  Case "b<>n".
  destruct t2; subst; inversion Heq.
  destruct (b0===n). inversion H0.
  inversion H0; subst; auto.

Case "con".
   simpl in Heq.
   destruct t2; subst; inversion Heq.
   destruct (b===n); auto. inversion H0.
   subst; auto.

Case "ctr".
   simpl in Heq.
   destruct t2; subst; inversion Heq.
   destruct (b===n); auto. inversion H0.
   subst; auto.

Case "app".
  simpl in Heq.
  destruct t2; subst; inversion Heq.
  destruct (b===n). inversion H0. inversion H0.
  assert (Jeq1: t1_1 = t2_1).
  eapply IHt1_1; eauto. simpl in Hin; auto.
  assert (Jeq2: t1_2 = t2_2).
  eapply IHt1_2; eauto. simpl in Hin; auto.
  subst; auto.
   
Case "abs". 
  simpl in Heq.
  destruct t2; subst; inversion Heq.
  destruct (b===n). inversion H0. inversion H0.
  assert (Jeq1: t1_1 = t2_1).
  eapply IHt1_1; eauto. simpl in Hin; auto.
  assert (Jeq2: t1_2 = t2_2).
  eapply IHt1_2 with (n:= S n); eauto. simpl in Hin; auto.
  subst; auto.
   

Case "arr".
   simpl in Heq.
  destruct t2; subst; inversion Heq.
  destruct (b===n). inversion H0. inversion H0.
  assert (Jeq1: t1_1 = t2_1).
  eapply IHt1_1; eauto. simpl in Hin; auto.
  assert (Jeq2: t1_2 = t2_2).
  eapply IHt1_2 with (n:= S n); eauto. simpl in Hin; auto.
  subst; auto.

Case "tm_mat".
  simpl in Heq.
  destruct t2; subst; inversion Heq.
  destruct (b===n). inversion H0. inversion H0.
  assert (Jeq1: t1_1 = t2_1).
  eapply IHt1_1; eauto. simpl in Hin; auto.
  assert (Jeq2: t1_2 = t2_2).
  eapply IHt1_2; eauto. simpl in Hin; auto.
  assert (Jeq3: t1_3 = t2_3).
  eapply IHt1_3; eauto. simpl in Hin; auto.
  subst; auto.
   

Case "tm_brn".
   simpl in Heq.
   destruct t2; subst; inversion Heq.
   destruct (b===n). inversion H0. inversion H0.
   assert (Jeq1: t1_1 = t2_1).
   eapply IHt1_1; eauto. simpl in Hin; auto.
   assert (Jeq2: t1_2 = t2_2).
   eapply IHt1_2; eauto. simpl in Hin; auto.
   subst; auto.
   

Case "tm_cast".
   simpl in Heq.
   destruct t2; subst; inversion Heq.
   destruct (b===n). inversion H0. inversion H0.
   assert (Jeq1: t1_1 = t2_1).
   eapply IHt1_1; eauto. simpl in Hin; auto.
   assert (Jeq2: t1_2 = t2_2).
   eapply IHt1_2; eauto. simpl in Hin; auto.
   subst; auto.
Qed.

Lemma wf_tm_unique : forall S1 S2 S3 E e t1 t2,
   wf_sig S1
 -> wf_tm S1 S2 E e t1
 -> wf_tm S1 S3 E e t2
 -> t1 = t2.
Proof.
 introv Hwf_sig Hwf1 Hwf2.
 gen t2.
 induction Hwf1;  subst; intros; try solve [inversion Hwf2; subst; auto].
 Case "arr".
  inversion Hwf2; subst.
  pick_fresh z.
  eapply H0 with (x:=z); eauto.
Case "ctr". 
 inversion Hwf2; subst.
 eapply wf_ctr_unique; eauto.
 destruct (ctr_decl_unique Hwf_sig H4 H0).
 Case "assn". 
   inversion Hwf2; subst.
   destruct (ctr_decl_unique Hwf_sig H0 H4).
   eapply wf_assn_unique; eauto.

 Case "fv".
   inversion Hwf2; subst.
   unfold binds_var in H0. unfold binds_var in H4.
   assert (Jeq:= binds_func H0 H4).
  inversion Jeq; auto.   

 Case "abs".
   inversion Hwf2; subst. 
   pick_fresh z.
   assert (Jeq : u2 ^^ tm_fv z = u3 ^^ tm_fv z).
   eapply H0 with (x:=z); eauto.
   assert (Jeq2: u2 = u3).
   eapply open_tm_eq with (n:=0) (x :=z); eauto.
   subst; auto. 

 Case "app". 
 inversion Hwf2; subst; try solve [inversion Hwf1_1].
 assert (Jeq: tm_arr t2 u = tm_arr t1 u0). eapply IHHwf1_1; eauto.
 inversion Jeq; subst; auto.
 inversion Hwf1_1.  inversion H7.
 inversion Hwf1_1. inversion H8.
 inversion Hwf1_1. inversion H7. 
 inversion Hwf1_1.  inversion H12. inversion H19. inversion H26.
 inversion Hwf1_1. inversion H8. 
 inversion Hwf1_1. inversion H7. 
 
 Case "P_SAYS". 
 inversion Hwf2; subst; auto.
  inversion H2. inversion H7.

Case "PF_SIGN".
 inversion Hwf2; subst; auto.
 inversion H3.  inversion H8.

Case "T_PF".
 inversion Hwf2; subst; auto.
 inversion H1.  

Case "TM_PF_RETURN".
 inversion Hwf2; subst; auto.
 inversion H1. assert (Jeq:= IHHwf1_1 P0 H0). subst; auto.

Case "TM_PF_BIND".   
 inversion Hwf2; subst; auto.
 inversion H2. inversion H7.
 assert (Jeq:= IHHwf1_2 (tm_arr P0 (tm_app T_PF Q0)) H4).
inversion Jeq; subst; auto.

Case "TM_SAY".
 inversion Hwf2; subst; auto.
 inversion H1.
 
Case "PIF".
 inversion Hwf2; subst; auto.
 inversion H6. inversion H11. inversion H18. inversion H25.

Case "P_SAYS".
 inversion Hwf2; subst; auto.
 inversion H2. inversion H7. 
 assert (Jeq:= IHHwf1_1 P0 H2). subst; auto.

Case "PF_SAYS_BIND".
 inversion Hwf2; subst; auto.
 inversion H2. inversion H7.
 assert (Jeq:= IHHwf1_2 (tm_arr P0 (tm_app(tm_app P_SAYS a0) Q0)) H4).
inversion Jeq; subst; auto.
Qed.

Lemma bundle_has_type_strong_inv1: forall S d K c t,
  wf_bundle_tctrs S (get_tctr_defns_of d) K ->
  (K=K_TYPE \/ K=K_PROP) ->
  defns_has_ctr c t (get_tctr_defns_of d) ->
  (constructs t K /\ wf_tm S S empty t S_KIND).
Proof. 
introv Hwf_tctr HK Hc.
induction d.
Case "nil". 
  inversion Hc.
Case "cons".
  destruct a. destruct p.
  simpl in Hc. inversion Hc. destruct (eq_ctr_dec c c1).
  Case "c=c1".
  inversion H0; subst.
  inversion Hwf_tctr; auto.
  Case "c<>c1".
      inversion Hwf_tctr; subst.
      eapply IHd; eauto.
Qed.

Lemma applied_value_inv1: forall S S1 S2 E d e u K,
  wf_sig S1 -> 
  wf_env S1 S2 E ->
  disjoint (dom_defns d) (dom_sig S) ->
  S1 = (bundle_defns (get_tctr_defns_of d)::S) ->
  S2 = (bundle_defns d::S) ->
  applied_value  e ->
  wf_tm S1 S2 E e u ->
  wf_tm S1 S2 E u K ->
  (K =  S_KIND 
   \/ exists c_inst, 
       exists ctr0, 
         exists args, 
           constructs u c_inst
           /\ fully_applied_args c_inst ctr0 args ).
Proof.
 introv Hwf_sig Hwf_env Hdisj HS1 HS2. intros Hvalue Hwf_e Hwf_u.
 gen u K.
 induction Hvalue; intros; subst.
 Case "ctr".
  inversion Hwf_e; subst.
  inversion H2.
  inversion Hwf_sig; subst.
  Case "TYPE".
  remember (bundle_defns_ctr_lookup c (get_tctr_defns_of d)) as B.
  destruct B.
  Case "c in d".
  inversion H1; subst. symmetry in HeqB.
  left. 
  assert (J: wf_tm (bundle_defns(get_tctr_defns_of d)::S)
                              (bundle_defns d::S) (empty&E) u S_KIND).
  eapply typing_weakening'; eauto.
  eapply wf_tm_weakening_sig_r; eauto.
  assert (J1: constructs u K_TYPE /\ wf_tm S S empty u S_KIND); auto. 
  eapply bundle_has_type_strong_inv1 with (c:=c); eauto.
  destruct J1; auto.
  rewrite_env E; auto.
  replace (empty& E) with E in J; auto.
  eapply wf_tm_unique with 
     (S1:= bundle_defns(get_tctr_defns_of d)::S)
     (S2:= bundle_defns d::S)
     (S3:= bundle_defns d::S); eauto. rewrite empty_concat; auto.
  Case "c in S". 
  assert (J:  wf_tm S S empty u S_KIND 
   \/ exists c_inst, 
       exists ctr0, 
         exists args, 
           constructs u c_inst
           /\ fully_applied_args c_inst ctr0 args ).
  eapply sig_has_ctr_type_strong_inv with (c:=c); eauto.
   destruct J; eauto.
   left.  assert (J: wf_tm (bundle_defns(get_tctr_defns_of d)::S)
                              (bundle_defns d::S) (empty&E) u S_KIND).
  eapply typing_weakening'; eauto.
  eapply wf_tm_weakening_sig_r; eauto.
  rewrite_env E; auto.
  replace (empty&E) with E in J; auto.
  eapply wf_tm_unique with 
     (S1:= bundle_defns(get_tctr_defns_of d)::S)
     (S2:= bundle_defns d::S)
     (S3:= bundle_defns d::S); eauto. rewrite empty_concat; auto.

 Case "PROP".
 remember (bundle_defns_ctr_lookup c (get_tctr_defns_of d)) as B.
  destruct B.
  Case "c in d".
  inversion H1; subst.
  symmetry in HeqB.
  left.  assert (J: wf_tm (bundle_defns(get_tctr_defns_of d)::S)
                              (bundle_defns d::S) (empty&E) u S_KIND).
  eapply typing_weakening'; eauto.
  eapply wf_tm_weakening_sig_r; eauto.
    assert (J1: constructs u K_PROP /\ wf_tm S S empty u S_KIND); auto. 
  eapply bundle_has_type_strong_inv1 with (c:=c); eauto.
  destruct J1; auto.
  rewrite_env E; auto.
  replace (empty& E) with E in J; auto.
  eapply wf_tm_unique with 
     (S1:= bundle_defns(get_tctr_defns_of d)::S)
     (S2:= bundle_defns d::S)
     (S3:= bundle_defns d::S); eauto. rewrite empty_concat; auto.
  Case "c in S". 
  assert (J:  wf_tm S S empty u S_KIND 
   \/ exists c_inst, 
       exists ctr0, 
         exists args, 
           constructs u c_inst
           /\ fully_applied_args c_inst ctr0 args ).
  eapply sig_has_ctr_type_strong_inv with (c:=c); eauto.
   destruct J; eauto.
   left.  assert (J: wf_tm (bundle_defns(get_tctr_defns_of d)::S)
                              (bundle_defns d::S) (empty&E) u S_KIND).
  eapply typing_weakening'; eauto.
  eapply wf_tm_weakening_sig_r; eauto.
  rewrite_env E; auto.
  replace (empty&E) with E in J; auto.
  eapply wf_tm_unique with 
     (S1:= bundle_defns(get_tctr_defns_of d)::S)
     (S2:= bundle_defns d::S)
     (S3:= bundle_defns d::S); eauto. rewrite empty_concat; auto.


   Case "assn".
   left.   assert (J: wf_tm (bundle_defns(get_tctr_defns_of d)::S)
                              (bundle_defns d::S) (empty&E) u S_KIND).
  eapply typing_weakening'; eauto.
  assert (J1:  constructs u K_PROP /\ wf_tm (bundle_defns(get_tctr_defns_of d)::S)
                              (bundle_defns d::S) empty u S_KIND).
 eapply sig_has_assn_has_typ_inv with (c:=c); eauto.
   unfold sig_extend; right;eauto. destruct J1; auto. 
   rewrite_env E; auto.
   replace (empty & E) with E in J.
    eapply wf_tm_unique with 
     (S1:= bundle_defns(get_tctr_defns_of d)::S)
     (S2:= bundle_defns d::S)
     (S3:= bundle_defns d::S); eauto. rewrite empty_concat; auto.

 Case "app".
 inversion Hwf_e; subst; try solve [inversion Hvalue; auto].
  assert (Jext: sig_extend (bundle_defns(get_tctr_defns_of d)::S)
                                             (bundle_defns d::S)); eauto. 
                                             unfold sig_extend; right; eauto.
                                            
  assert (J:= wf_tm_tm Hwf_sig H2 Jext).
  destruct J. inversion  H0.
  destruct H0 as [K' Jwfarr].
  destruct (IHHvalue (tm_arr t0 u0) H2 K'); auto.

    Case "t0->u0 S_KIND".
    clear IHHvalue.
    left.  inversion Jwfarr; subst. clear H10.
    pick_fresh z. 
    assert (J: wf_tm (bundle_defns(get_tctr_defns_of d)::S)
                         (bundle_defns d::S) (E & z~:t0) (u0^^ tm_fv z) S_KIND).
    eapply (H8  z); eauto.
    assert (J1: wf_tm  (bundle_defns(get_tctr_defns_of d)::S)
                         (bundle_defns d::S) E (u0^^t2) S_KIND).
    rewrite (@subst_intro z); auto.
    replace S_KIND  with ([z~>t2]S_KIND); auto.
    eapply simple_tm_substitution; eauto.
    eapply (proj1 (wf_tm_branches_lc_mutual  (bundle_defns(get_tctr_defns_of d)::S)
                         (bundle_defns d::S))); eauto.
    (*eapply wf_tm_unique with  (S1:=bundle_defns(get_tctr_defns_of d)::S)
                         (S2:=bundle_defns d::S) ; eauto. *) skip.
    Case "inv".
    right. destruct H0 as [c_inst [ctr0 [args [Hcon Hfully_app]]]].
    inversion Hcon; subst. 
    assert (J:= is_tm_arr_arr t0 u0); contradiction. 
    assert (J:= open_construct t2 0 H5 Hfully_app).
    destruct J as [c_inst' [args'  [Jcon Jfully_app]]].
    exists c_inst' ctr0 args'; split; auto.

 Case "inversion ".
  inversion Hvalue. inversion H3.
  inversion Hvalue. inversion H4.
  inversion Hvalue. inversion H3. 
  inversion Hvalue. inversion H8. inversion H14. inversion H18.
  inversion Hvalue. inversion H4.
  inversion Hvalue. inversion H3.
Qed.



Lemma cononical_forms_T_PF: forall S e P,
 wf_sig S -> wf_tm S S empty e (tm_app T_PF P) -> value e  ->
 exists e1, e = tm_app TM_PF_RETURN e1 .
Proof.
intros S e P Hwf_sig Hwf_e Hvalue.
remember (tm_app T_PF P) as K. remember (empty:env) as E.
induction Hwf_e; try solve [subst; inversion HeqK; auto].

Case "arr".
rewrite HeqK in H1. destruct H1. inversion H1. 
 destruct H1. inversion H1. inversion H1.

Case "ctr".
 destruct (sig_has_ctr_type_strong_inv Hwf_sig H0); eauto.
 rewrite HeqK in H1. inversion H1. inversion H6. 
 destruct H1 as [x [y [z [Jcon Japp]]]].
 rewrite HeqK in Jcon.
 inversion Jcon; subst. inversion Japp. inversion H6.

Case "assn".
 assert (Jext: sig_extend S S). unfold sig_extend; auto.
assert (J:= sig_has_assn_has_typ_inv  Hwf_sig  Jext H0 ).  
 rewrite HeqK in J. destruct J. inversion H2. inversion H7. 

Case "fv".
rewrite HeqE in H0.  inversion H0.

Case "app".    (*need stronger inversion properties on ctr.*) 
clear IHHwf_e1 IHHwf_e2.
 inversion Hvalue; subst. 
 inversion H0; subst.
 assert (Jenv: wf_env S S empty); auto.
 assert (J:= applied_value_inv Hwf_sig Jenv H3 Hwf_e1).
 destruct J.
 Case "S_KIND".
   inversion H1; subst.
   pick_fresh z. 
   assert (J: wf_tm S S (empty & z~:t2) (u^^tm_fv z) S_KIND).
   eapply H8; eauto.
   assert (J1: wf_tm S S empty ([z~>e2](u^^ tm_fv z)) ([z~>e2]S_KIND)).
  eapply simple_tm_substitution; eauto.
  rewrite <- subst_intro in J1; auto. rewrite subst_fresh in J1; auto.
  rewrite HeqK in J1.  inversion J1. inversion H11. 
  unfold S_KIND; simpl; auto. 
  eapply (proj1 (wf_tm_branches_lc_mutual S S)); eauto.

 Case "inv".
 destruct H1 as [c_inst [ctr0 [args [Hcon Hfully_app]]]].
 inversion Hcon; subst. 
 assert (J:= is_tm_arr_arr t2 u); contradiction. 
 assert (J1:= open_construct e2 0 H6 Hfully_app).
 destruct J1 as [c_inst' [args' [Jcon Jfully_app]]].
 fold open_tm in Jcon. rewrite HeqK in Jcon.  inversion Jcon; subst.
 inversion Jfully_app. inversion H9.

  Case "PF_SIGN".
   inversion Hwf_e1. inversion H5.
 Case "P_SAYS".
 inversion Hwf_e1. inversion H4.

   Case "PF_RETURN".
   inversion Hwf_e1.

Case "PF_RETURN". 
 exists e2; auto.
 
 inversion Hwf_e1. inversion H4.
 inversion Hwf_e1. inversion H4. 

 exists p; auto.
Case "PF_BIND".
inversion Hvalue.
 inversion H0. inversion H4. inversion H8. 

Case "TM_SAY".
 inversion Hvalue; subst. inversion H. inversion H2.

Case "TM_PIF".
inversion Hvalue. 
 inversion H4. inversion H8. inversion H12. inversion H16. inversion H20.

 Case "tm_mat_TYPE".
  clear IHHwf_e1 IHHwf_e2 IHHwf_e3.
 inversion Hvalue; subst. inversion H4.

 Case "tm_mat_PROP".
  clear IHHwf_e1 IHHwf_e2 IHHwf_e3.
 inversion Hvalue; subst. inversion H4.

Case "tm_cast".
   clear IHHwf_e1 IHHwf_e2.
   inversion Hvalue; subst. inversion H0.
Qed.


Lemma caonnical_forms_arr:  forall S e t1 t2,
 wf_sig S -> wf_tm S S empty e (tm_arr t1 t2) ->  value e  ->
(  (exists e2, e = tm_abs t1 e2) \/ applied_value  e).
Proof. 
 intros S e t1 t2 Hwf_sig Hwf_e Hvalue.
 inversion Hwf_e; subst; try solve [inversion Hvalue; auto].

Case "tm_arr".
 destruct H1. inversion H1. destruct H1. inversion H1. inversion H1.

Case "fv".
 inversion H0.

Case "abs".
 left. exists f; auto.

Case "app_value". 
  right. 
  inversion Hvalue; subst; auto.
  inversion Hwf_e. inversion H8.  inversion H15.
  inversion H0. inversion H7.
  inversion H0. inversion H0.
  inversion H0. inversion H7. 
  inversion H0. inversion H7. 
Qed.


Inductive fully_applied_args_ctr : tm -> ctr -> list tm -> Prop :=
| fully_applied_args_ctr_base : forall c, fully_applied_args_ctr (tm_ctr c) c nil
| fully_applied_args_ctr_app  : forall t tn c args,
    fully_applied_args_ctr t c args  ->
    fully_applied_args_ctr (tm_app t tn) c (tn::args).

Hint Constructors fully_applied_args_ctr.

Lemma fully_app_type_gives_fully_app_ctr: forall S e s ctr0 args l c_inst u,
  wf_sig S ->
(*  sig_extend S1 S2 ->*)
  wf_tm S S empty e s ->
  wf_tm S S empty s u ->
  ( u = K_TYPE \/ u = K_PROP) ->
  value e ->
  constructs s c_inst ->
  fully_applied_args c_inst ctr0 args  ->
  get_ctrs_of ctr0 S = Some l ->
 ( (exists t, exists f, e = tm_abs t f )
  \/ (exists c, exists n, exists t, exists args', 
         fully_applied_args_ctr e (c,n) args'  
         /\ applied_value e 
          /\ In (c,n) l /\ sig_has_ctr (c,n) t S)).
Proof. 
 introv  Hwf_sig (* Hext*) Hwf_e Hwf_s Hu. intros Hvalue. intros Hcon Hfully_app (*Hwf_ctr0*) Hget_ctrs.
 remember (empty:env) as E.
gen c_inst ctr0 args.
 induction Hwf_e; subst; intros; try solve [inversion Hfully_app; auto].

   assert (J: c_inst = S_KIND).
  eapply construct_unique; eauto.
  apply constructs_base; auto.
  unfold not. intro. inversion H0.
  rewrite J in Hfully_app. inversion Hfully_app. 

  assert (J: c_inst = S_KIND).
  eapply construct_unique; eauto.
  apply constructs_base; auto.
  unfold not. intro. inversion H0.
  rewrite J in Hfully_app. inversion Hfully_app. 

Case "arr".
 clear IHHwf_e H0. 
 destruct H1; subst.
     assert (J: c_inst = K_TYPE).
  eapply construct_unique; eauto.
  apply constructs_base; auto.
  unfold not. intro. inversion H0.
  rewrite J in Hfully_app. inversion Hfully_app. 
  destruct H0; subst.
     assert (J: c_inst = K_PROP).
  eapply construct_unique; eauto.
  apply constructs_base; auto.
  unfold not. intro. inversion H0.
  rewrite J in Hfully_app. inversion Hfully_app. 
    assert (J: c_inst = S_KIND).
  eapply construct_unique; eauto.
  apply constructs_base; auto.
  unfold not. intro. inversion H0.
  rewrite J in Hfully_app. inversion Hfully_app. 

Case "ctr". 
destruct c as (c1, n). 
 right. exists c1 n t (@nil tm). repeat split; auto.
eapply fully_applied_args_ctr_in; eauto.

Case "assn".
assert (J: constructs t K_PROP /\ wf_tm S S empty t S_KIND).
eapply sig_has_assn_has_typ_inv; eauto.
unfold sig_extend; auto. destruct J.
destruct Hu; subst;
assert (J1:= wf_tm_unique Hwf_sig Hwf_s H2 ); inversion J1.

Case "fv". inversion H0.

Case "abs".  clear IHHwf_e1 IHHwf_e2 H0. 
 left.  exists u1 f; auto. 

Case "app".
 clear IHHwf_e2.
 inversion Hvalue; subst. inversion H0; subst.
   right.
   assert (Jevn: wf_env S S empty); auto.
 assert (J:= applied_value_inv Hwf_sig Jevn H3 Hwf_e1).
 destruct J.
  Case "S_KIND".
   inversion H1; subst. 
   pick_fresh z.
   assert (J: wf_tm S S (empty & z~:t2) (u0^^ tm_fv z) S_KIND).
   eapply H8; eauto.
   assert (J1: wf_tm S S empty (u0^^e2)  S_KIND).
   rewrite (@subst_intro z).
   replace S_KIND  with ([z~>e2]S_KIND); auto.
   eapply simple_tm_substitution; eauto.  auto.
   eapply (proj1 (wf_tm_branches_lc_mutual S S)); eauto.
   assert (Jeq:= wf_tm_unique Hwf_sig J1 Hwf_s); subst.
   destruct Hu. inversion H2. inversion H2.
  Case "inv".
  destruct H1 as [c_inst1 [ctr1 [args1 [Jcon Jfully_app]]]].
  inversion Jcon; subst.
  assert (J:= is_tm_arr_arr t2 u0); contradiction.
  assert (J1:= open_construct e2 0 H6 Jfully_app).
  destruct J1 as [c_inst2 [args2  [Jcon1 Jfully_app1]]].
  assert (Jeq:= construct_unique Jcon1 Hcon).
  rewrite Jeq in Jfully_app1.
  destruct (fully_applied_args_unique Jfully_app1 Hfully_app); subst.
  assert (Jval:= value_app H3).
  assert (Jinv: (exists t:tm, exists f:tm, e1 = tm_abs t f) 
  \/ (exists c, exists n,   exists t,    exists args', fully_applied_args_ctr e1 (c,n) args'  /\ applied_value e1   
             /\ In (c,n) l /\ sig_has_ctr (c,n) t S)).
 eapply IHHwf_e1; eauto.
 clear IHHwf_e1.
 assert (Jext : sig_extend S S). unfold sig_extend; auto.
 destruct ( wf_tm_tm Hwf_sig Hwf_e1 Jext ). inversion H1.
 destruct H1 as [k_arr Jwf].
 assert (Jeq: u = k_arr). 
 inversion Jwf; subst. 
 assert (J: wf_tm S S empty (u0^^e2) k_arr).
 pick_fresh z.
 rewrite (@subst_intro z); auto.
 replace k_arr with ([z~>e2]k_arr); auto.
 eapply simple_tm_substitution; eauto.
 rewrite subst_fresh; auto. 
 eapply (proj1 (wf_tm_branches_lc_mutual S S)); eauto.
  eapply wf_tm_unique; eauto.
 subst; auto.
 destruct Jinv. destruct H1 as [tu [f J10]]; subst; inversion H3.
 destruct H1 as [c1 [n1 [t1 [argsx [Jfully_app_e1 [Jx [Jin Jc]]]]]]].
 exists c1 n1 t1 (e2::argsx); repeat split; auto.

Case "inversion". 
 clear IHHwf_e1.
  inversion Hwf_e1. inversion H5.
 clear IHHwf_e1. inversion Hwf_e1. inversion H4. 
 clear IHHwf_e1. inversion Hwf_e1. 
  clear IHHwf_e1. inversion Hwf_e1. 
 clear IHHwf_e1. inversion Hwf_e1. inversion H4.
  clear IHHwf_e1. inversion Hwf_e1. inversion H4.
  clear IHHwf_e1. inversion Hwf_s; subst.
         destruct Hu. inversion H1. inversion H1.
 clear IHHwf_e1 IHHwf_e2. 

  assert (J: (tm_app (tm_app P_SAYS v) p) = c_inst).
  eapply construct_unique; eauto.
  apply constructs_base; auto.
  unfold not. intro. inversion H1.
  rewrite <- J in Hfully_app. inversion Hfully_app. inversion H5. inversion H10.

  assert (J: c_inst = K_TYPE).
  eapply construct_unique; eauto.
  apply constructs_base; auto.
  unfold not. intro. inversion H0.
  rewrite J in Hfully_app. inversion Hfully_app. 

  assert (J: c_inst = K_TYPE).
  eapply construct_unique; eauto.
  apply constructs_base; auto.
  unfold not. intro. inversion H.
  rewrite J in Hfully_app. inversion Hfully_app. 

  assert (J: c_inst = tm_app T_PF P).
  eapply construct_unique; eauto.
  apply constructs_base; auto.
  unfold not. intro. inversion H.
  rewrite J in Hfully_app. inversion Hfully_app. inversion H3.
 
  assert (J: c_inst = tm_app T_PF Q).
  eapply construct_unique; eauto.
  apply constructs_base; auto.
  unfold not. intro. inversion H0.
  rewrite J in Hfully_app. inversion Hfully_app. inversion H4.

 assert (J: c_inst = (tm_app T_PF (tm_app (tm_app P_SAYS TM_SELF) P))).
  eapply construct_unique; eauto.
  apply constructs_base; auto.
  unfold not. intro. inversion H.
  rewrite J in Hfully_app. inversion Hfully_app. inversion H3.

 inversion Hvalue. inversion H4. inversion H8. inversion H12. inversion H16. inversion H20.

  assert (J: c_inst = T_PRIN).
  eapply construct_unique; eauto.
  apply constructs_base; auto.
  unfold not. intro. inversion H0.
  rewrite J in Hfully_app. inversion Hfully_app. 

  assert (J: c_inst = tm_app (tm_app P_SAYS a) P).
  eapply construct_unique; eauto.
  apply constructs_base; auto.
  unfold not. intro. inversion H0.
  rewrite J in Hfully_app. inversion Hfully_app. inversion H4. inversion H9.
 
  assert (J: c_inst = tm_app (tm_app P_SAYS a) Q).
  eapply construct_unique; eauto.
  apply constructs_base; auto.
  unfold not. intro. inversion H0.
  rewrite J in Hfully_app. inversion Hfully_app. inversion H4. inversion H9.

 inversion Hvalue. inversion H4. 
inversion Hvalue. inversion H4.
inversion Hvalue. inversion H0.
Qed.

Lemma applied_value_inv': forall  S1 S2 E  e u K,
  wf_sig S1 -> 
  wf_env S1 S2 E ->
  sig_extend S1 S2 ->
  applied_value  e ->
  wf_tm S1 S2 E e u ->
  wf_tm S1 S2 E u K ->
  (K =  S_KIND 
   \/ exists c_inst, 
       exists ctr0, 
         exists args, 
           constructs u c_inst
           /\ fully_applied_args c_inst ctr0 args ).
Proof.
 introv Hwf_sig Hwf_env Hexd Hvalue Hwf_e. intros Hwf_u.
 destruct Hexd; subst; auto.
 assert (J: wf_tm S2 S2 E u S_KIND \/ exists c_inst, 
                                                                  exists ctr0, 
                                                                  exists args, 
                                                                  constructs u c_inst
                                                                  /\ fully_applied_args c_inst ctr0 args ). 
                                                                  eapply applied_value_inv; eauto.
 destruct J; eauto.
 left. eapply wf_tm_unique with (S1:=S2); eauto.

 destruct H as [d [S [H1 [H2 H3]]]].
 eapply applied_value_inv1; eauto.
Qed.

Lemma ctr_type_gives_fully_app_ctr: forall S1 S2 E e u s ctr0 args  c_inst,
  wf_sig S1 ->
  sig_extend S1 S2 ->
  wf_tm S1 S2 E e s ->
  wf_tm S1 S2 E s u ->
  ( u = K_TYPE \/ u = K_PROP) ->
  applied_value e ->
  constructs s c_inst ->
  fully_applied_args c_inst ctr0 args  ->
  exists l, 
  get_ctrs_of ctr0 S2 = Some l /\
  (exists c, exists args', 
         fully_applied_args_ctr e c args'
         /\  In c l ).
Proof. 
 introv  Hwf_sig  Hext Hwf_e  Hwf_s Hu. intros Hvalue Hcon Hfully_app (*Hwf_ctr0  Hget_ctrs*).
gen c_inst ctr0 args.
 induction Hwf_e; subst; intros; try solve [inversion Hvalue; eauto].

Case "ctr".
  remember (get_ctrs_of ctr0 S2) as GG.
  destruct GG. symmetry in HeqGG. exists l; split; auto.
  exists c (@nil tm); split; auto.
  eapply fully_applied_args_ctr_in; eauto.
  destruct Hext; subst; auto.
  destruct H1 as [d [S [HS1 [HS2 HS3]]]]; subst.
  inversion H0.
  remember (bundle_defns_ctr_lookup c (get_tctr_defns_of d)) as B.
  destruct B.
  symmetry in HeqB.
  inversion H2; subst.
  inversion Hwf_sig; subst. 
  assert (J: constructs t K_TYPE /\ wf_tm S S empty t S_KIND).
 eapply   bundle_has_type_strong_inv1 with (c:=c); eauto.
 destruct J.
  assert (J:= construct_unique H1 Hcon); subst. inversion Hfully_app.
    assert (J: constructs t K_PROP /\ wf_tm S S empty t S_KIND).
 eapply   bundle_has_type_strong_inv1 with (c:=c); eauto.
 destruct J.
  assert (J:= construct_unique H1 Hcon); subst. inversion Hfully_app.
  Case "c in S".
   inversion Hwf_sig; subst.
   assert (J: In ctr0 (dom_sig S)).
   eapply fully_applied_args_ctr_in_sig with (c:=c); eauto. apply H2.
   inversion HeqGG.
   remember (get_defns_ctrs_of ctr0 d) as G.
   destruct G. 
   inversion H3; subst. symmetry in HeqG.
   assert (J1:= get_defns_ctr_some_in ctr0 d HeqG).
   destruct (HS3 ctr0 J1 J).
   unfold get_ctrs_of. 
   remember (get_defns_ctrs_of ctr0 (get_tctr_defns_of d)) as G1.
   destruct G1; auto. symmetry in HeqG1.
   assert (J1:= get_defns_ctr_some_in ctr0 (get_tctr_defns_of d) HeqG1).
   assert (J2:= in_tctr_defns_in ctr0 d J1).
   destruct (HS3 ctr0 J2 J).

   assert (J: In ctr0 (dom_sig S)).
   eapply fully_applied_args_ctr_in_sig with (c:=c); eauto. apply H2.
   inversion HeqGG.
   remember (get_defns_ctrs_of ctr0 d) as G.
   destruct G. 
   inversion H3; subst. symmetry in HeqG.
   assert (J1:= get_defns_ctr_some_in ctr0 d HeqG).
   destruct (HS3 ctr0 J1 J).
   unfold get_ctrs_of. 
   remember (get_defns_ctrs_of ctr0 (get_tctr_defns_of d)) as G1.
   destruct G1; auto. symmetry in HeqG1.
   assert (J1:= get_defns_ctr_some_in ctr0 (get_tctr_defns_of d) HeqG1).
   assert (J2:= in_tctr_defns_in ctr0 d J1).
   destruct (HS3 ctr0 J2 J).

 Case "get_ctrs_of ctr0 S2 = None".
  assert (J: exists l, get_ctrs_of ctr0 S2 = Some l).
 eapply fully_applied_ctr_in_ext_sig; eauto.
 destruct J; subst. rewrite H1 in HeqGG; inversion HeqGG.
  
 Case "assn".
 assert (J: constructs t K_PROP /\ wf_tm S1 S2 empty t S_KIND).
 eapply sig_has_assn_has_typ_inv; eauto. destruct J.
 assert (J1:= construct_unique Hcon H1 ); subst. inversion Hfully_app.

Case "app".
 clear IHHwf_e2.
 inversion Hvalue; subst. 
 assert (Jenv:= wf_tm_gives_wf_env Hwf_e1).
  assert (Jarr:= wf_tm_tm Hwf_sig Hwf_e1).
  destruct Jarr; auto. 
   inversion H0. destruct H0 as [K Jwf_arr].
 assert (J:= applied_value_inv' Hwf_sig Jenv Hext H2 Hwf_e1 Jwf_arr).
 destruct J.
  Case "S_KIND".
   inversion Jwf_arr; subst. 
   pick_fresh z.
   assert (J: wf_tm S1 S2 (E & z~:t2) (u0^^ tm_fv z) S_KIND).
   eapply (H7 z); eauto.
   assert (J1: wf_tm S1 S2 E (u0^^e2)  S_KIND).
   rewrite (@subst_intro z); auto.
   replace S_KIND  with ([z~>e2]S_KIND); auto.
   eapply simple_tm_substitution; eauto.  
   eapply (proj1 (wf_tm_branches_lc_mutual S1 S2)); eauto.
   assert (Jeq:= wf_tm_unique Hwf_sig J1 Hwf_s); subst.
   destruct Hu ; inversion H0. 
  Case "inv".
  destruct H0 as [c_inst1 [ctr1 [args1 [Jcon Jfully_app]]]].
  inversion Jcon; subst.
  assert (J:= is_tm_arr_arr t2 u0); contradiction.
  assert (J1:= open_construct e2 0 H5 Jfully_app).
  destruct J1 as [c_inst2 [args2  [Jcon1 Jfully_app1]]].
  assert (Jeq:= construct_unique Jcon1 Hcon).
  rewrite Jeq in Jfully_app1.
  destruct (fully_applied_args_unique Jfully_app1 Hfully_app); subst.
  assert (Jinv: exists l, get_ctrs_of ctr0 S2 = Some l /\ exists c, exists args', fully_applied_args_ctr e1 c args'  
             /\ In c l ).
 eapply IHHwf_e1; eauto.
 clear IHHwf_e1.
 assert (Jeq: u = K). 
 inversion Jwf_arr; subst. 
 assert (J: wf_tm S1 S2 E (u0^^e2) K).
 pick_fresh z.
 rewrite (@subst_intro z); auto.
 replace K with ([z~>e2]K); auto.
 eapply simple_tm_substitution; eauto.
 rewrite subst_fresh; auto. 
 eapply (proj1 (wf_tm_branches_lc_mutual S1 S2)); eauto.
  eapply wf_tm_unique; eauto.
 subst; auto.
 destruct Jinv as [l [Jget [c1 [argsx [Jfully_app_e1 Jin]]]]].
 exists l; split ;auto; exists c1 (e2::argsx); repeat split; auto.

Case "inversion". 
 clear IHHwf_e1.
 inversion Hvalue. inversion H2. inversion H6.
 inversion Hvalue. inversion H3. inversion H7.
 inversion Hvalue. inversion H1. 
inversion Hvalue.  inversion H1. 
inversion Hvalue. inversion H2. inversion H6.
inversion Hvalue.  inversion H1. 
inversion Hvalue. inversion H6. inversion H10. inversion H14. inversion H18.
inversion Hvalue. inversion H2. inversion H6. 
inversion Hvalue. inversion H2. inversion H6. 
Qed.

Lemma step_matches_ctr_progress: forall S e c n nc (* na *) args body t ctr0 k l  c_tp,
  wf_sig S ->
  fully_applied_args_ctr e (c,n) args ->
  applied_value  e ->
  wf_tm S S empty e t ->
  sig_has_ctr ctr0 k S ->
  get_ctrs_of  ctr0 S = Some l ->
  In (c, n) l ->
  sig_has_ctr (c,n) c_tp S ->
  has_level c_tp nc ->
  exists res, exists m, exists args1, exists args2, 
  step_matches_ctr e (c,n) body  (res, m, args1, args2)
  /\ (m >0 -> (args2 = nil /\  res = body))
  /\ exists c_inst, exists args, exists bargs, 
  (constructs t c_inst /\     
  fully_applied_args c_inst ctr0 args /\  
  args = bargs++(List.rev args1)  /\ 
  List.length bargs = m /\ 
  all_tm_bv (nc - n)  bargs /\ 
  all_tm_lc args1).
Proof.
 intros S e c n nc (* na *)args body t ctr0 k l c_tp Hwf_sig Hfully_app_ctr Hvalue Hwf_e Hwf_ctr0 Hget_ctr Hin Hwf_c Hlevel (* Heq*).
 gen t args.
 remember (empty:env) as E.
 induction Hvalue; intros; subst.
Case "ctr".
  inversion Hfully_app_ctr; subst.
  exists body n (@nil tm) (@nil tm).
  assert (J: step_matches_ctr (tm_ctr(c,n)) (c,n) body (body, n, nil,nil)).
  apply step_matches_ctr_base; auto.
  repeat split; auto.
  eapply wf_brn_arg_unique; eauto.

Case "app".
  inversion Hfully_app_ctr; subst.
  inversion Hwf_e; subst. 
  Case "app_value".
  assert (Jeq: empty = (empty :env)); auto.
  assert  (J:= IHHvalue  (tm_arr t0 u) H2 args0 H4). 
  destruct J as [res [m [args1 [args2 [Jstep [Jarg Jinv]]]]]].
  assert (Jm: (m>0) \/ (m=0)).  omega.
  destruct Jm.
  Case "m>0".
  destruct (Jarg H0) as [J1 J2]; subst.
  assert (Jstep': step_matches_ctr (tm_app t1 t2) (c,n) body (body, m -1, args1++t2::nil, nil)).
  eapply step_matches_ctr_strip_type_arg; eauto.
  exists body (m-1) (args1++t2::nil)  (@nil tm); repeat split; auto.
  eapply wf_brn_arg_unique; eauto.
 
  Case "m=0". subst.
  assert (Jstep': step_matches_ctr (tm_app t1 t2) (c,n) body (tm_app res t2, 0, args1, args2++t2::nil)).
  eapply step_matches_ctr_rec; eauto.
  exists (tm_app res t2)  0 args1  (args2++t2::nil); repeat split; auto.
  inversion H0. inversion H0.
  eapply wf_brn_arg_unique; eauto.


  Case "inversion".
   inversion H4. inversion H8.
   inversion H4. inversion H9.
   inversion H4. inversion H4.
   inversion H4. inversion H8.
   inversion H4. 
   inversion H4. inversion H13. inversion H18. inversion H23. 
   inversion H4. inversion H9. 
   inversion H4. inversion H8. 
Qed.

Lemma lc_fully_applied_args_args_lc: forall t ctr0 args0,
 lc t -> fully_applied_args t ctr0 args0  -> all_tm_lc args0.
Proof.
 introv Hlc Hfully_app.
 induction Hfully_app.
 apply all_tm_lc_nil. 
 (* apply all_tm_lc_nil.*)
 inversion Hlc; subst.
 apply all_tm_lc_cons; auto.
Qed.
 
Lemma all_tm_lc_app_l: forall l1 l2, 
  all_tm_lc (l1++l2) -> all_tm_lc l1.
Proof.
 intros. 
 induction l1.
 apply all_tm_lc_nil.
 rewrite <- app_comm_cons in H.
 inversion H; subst.
 eapply all_tm_lc_cons; eauto.
Qed.
 
 
Lemma step_matches_branches_progress: forall S e c n k args ags l1 l2 branches ctr0  c_tp t nc,
  wf_sig S ->
  fully_applied_args_ctr e (c,n)  args ->
  applied_value e ->
  wf_tm S S empty e t ->
  fully_applied_args t ctr0 ags  ->
  sig_has_ctr ctr0 k S ->
  get_ctrs_of  ctr0 S = Some (l1 ++ l2) ->
  In (c,n) l2 ->
  branches_cover_aux l2 branches ->
  sig_has_ctr (c,n) c_tp S ->
  has_level c_tp nc ->
  exists res, step_matches_branches e branches res.
Proof. 
 introv Hwf_sig Hfully_app_ctr
  Hvalue Hwf_e Hfully_app. intros Hwf_ctr0 Hget_ctrs Hin Hbrn_cover Hwf_c Hlevel.
 gen l1.
 induction Hbrn_cover; intros.
 Case "nil".
  inversion Hin.
 Case "cons".
   destruct (in_inv Hin). 
   Case "c = c0".
    subst. clear IHHbrn_cover.
     assert (Jin: In (c,n) (l1++(c,n)::l)). eapply in_or_app; eauto.
    assert (J:= step_matches_ctr_progress t0 Hwf_sig Hfully_app_ctr Hvalue Hwf_e Hwf_ctr0 Hget_ctrs Jin Hwf_c Hlevel).
    destruct J as [res [m [ args1 [args2 [Jstep [Jm Jinv]]]]]].
    destruct Jinv as [c_inst [args0 [bargs [Jcon [Jfully_app [Jeq [Jlen[ Jbv Jlc]]]]]]]].
     assert (Jcon' : constructs t t). 
       apply constructs_base. intro. inversion H; subst; inversion Hfully_app. 
      assert (J3:= construct_unique Jcon Jcon'). rewrite J3 in Jfully_app. 
      assert (Jarg: all_tm_lc args0). 
      assert (Jlct: lc t).
      assert (Jext : sig_extend S S). unfold sig_extend; auto.
      assert (Jwf:= wf_tm_tm Hwf_sig Hwf_e Jext).
      destruct Jwf. subst; unfold S_KIND; auto.
      destruct H.
       eapply (proj1 (wf_tm_branches_lc_mutual S S)); eauto.
      eapply  (lc_fully_applied_args_args_lc Jlct Jfully_app); eauto.
      rewrite Jeq in Jarg.  assert (Jarg': all_tm_lc bargs).
      eapply all_tm_lc_app_l; eauto.
      destruct bargs.  
      Case "bargs = nil".  simpl in Jlen.
      rewrite <- Jlen in Jstep.
      exists res. eapply step_matches_here ; eauto.
      Case "bargs<>nil".
      inversion Jbv; subst. inversion Jarg'; subst. inversion H3.
      
     Case "In c l".
    destruct (IHHbrn_cover H (l1++c0::nil)); auto. 
     rewrite Hget_ctrs.  rewrite app_ass. rewrite <- app_comm_cons; auto.  
    exists x.  eapply step_matches_earlier; eauto.
Qed. 

Lemma tm_has_level: forall t, exists n, has_level t n.
Proof.
 intros.
 induction t; subst; try solve [exists 0; apply has_level_base; auto; unfold not; intro; inversion H].
 destruct IHt2.
 exists (1+x); apply has_level_arr; auto.
Qed.

Lemma progress: forall S e t,
  wf_sig S -> wf_tm S S empty e t -> 
  (value e) \/ (exists e', step  e e').
Proof.
  intros S e t  Hwf_sig Hwf_e.
  remember (empty:env) as E. 
  induction Hwf_e; intros; try solve [left; auto]. 
       
  Case "app_value".
  destruct IHHwf_e1; auto.
  Case "e1 value".
    subst.
    assert  (Jvalue:= caonnical_forms_arr Hwf_sig Hwf_e1 H0).   
     destruct IHHwf_e2; auto.
       Case "value e2". 
          destruct Jvalue. 
          Case "e1 = abs".
          destruct H2 as [e0 Jeq_abs]. rewrite Jeq_abs.
          right. exists (e0^^e2). eapply step_app; eauto.
          Case "e1 applied_value".
           left. apply value_app; auto.       
       Case "e2 step".
         destruct H1 as [e2' Jstep].
         right. exists (tm_app e1 e2').  
         destruct Jvalue.  destruct H1 as [e0 Jeq_abs]; subst. eapply step_app_cong2; eauto.
         apply step_app_cong3; auto.

   Case "e1 steps".
   destruct H0. right. exists (tm_app x e2). eapply step_app_cong1; eauto.

  Case "PF_RETURN".
   destruct IHHwf_e1; auto.
   destruct H as [p' Jstep]. 
   right. exists (tm_app TM_PF_RETURN p'); apply step_pf_ret_cong; auto.
 
  Case "pf_bind". 
    destruct (IHHwf_e1); auto.     
     Case "value e1".
     destruct IHHwf_e2; auto.
      Case "value e2".
       right.  subst. assert (Jc:=  cononical_forms_T_PF Hwf_sig Hwf_e1 H0).
        destruct Jc as [e0 Jeq].
        exists (tm_app e2 e0).  rewrite Jeq. eapply step_pf_bind; eauto.
       Case "e2 step". subst.
        destruct H1 as [e2' Jstep].
        right. exists (tm_app(tm_app TM_PF_BIND e1) e2'). 
        apply step_pf_bind_cong2; auto.
      Case "cong1".
      right. destruct H0. exists (tm_app(tm_app TM_PF_BIND x) e2).
      apply step_pf_bind_cong1; auto.

 Case "TM_SAY".
  right. exists (tm_app TM_PF_RETURN (tm_app (tm_app PF_SIGN TM_SELF) P)).
  apply step_say; auto.

Case "TM_PIF".
  right. destruct (eq_tm_dec v1 v2).
  exists e1; apply  step_pif1; auto.
  exists e2; apply  step_pif2; auto.

Case "tm_mat_TYPE". 
clear IHHwf_e2 IHHwf_e3.
subst.
destruct IHHwf_e1; auto.
 Case "step_matches".
  right.
  unfold branches_cover in H3.
  destruct H3 as [l [Hget_ctrs Hbrn_cover_aux]].

  assert (J: (exists t, exists f, e = tm_abs t f) \/ exists c, exists n,  exists t, exists args', fully_applied_args_ctr e (c,n) args' /\ applied_value e /\In (c,n) l /\ sig_has_ctr (c,n) t S ). 
  eapply fully_app_type_gives_fully_app_ctr; eauto.
  apply constructs_base. unfold not. intros. inversion H3; subst. inversion H.
   destruct J.
    Case "e=tm_abs".
    destruct H3 as [tp [f Jeq]].
    subst;  inversion Hwf_e1; subst. inversion H. 
    Case "inv".
    destruct H3 as [c [n [c_tp [args1 [Jfully_app_ctr [Japp_val [Jin Jwf_c]]]]]]].
    assert (Jl:= tm_has_level c_tp).
    destruct Jl.
    assert (J: exists res, step_matches_branches e branches res).
    replace l with (nil++l) in Hget_ctrs; auto.
    eapply step_matches_branches_progress with (ctr0:=ctr0); eauto. 
    destruct  J as [res Jstep].
    exists res.  eapply step_mat; eauto. 
    
 Case "cong".
  destruct H4 as [e' Hstep].
  right. exists (tm_mat e' result_typ branches). apply step_mat_cong; auto.

Case "tm_mat_PROP". 
clear IHHwf_e2 IHHwf_e3.
subst.
destruct IHHwf_e1; auto.
 Case "step_matches".
  right.
  unfold branches_cover in H3.
  destruct H3 as [l [Hget_ctrs Hbrn_cover_aux]].

  assert (J: (exists t, exists f, e = tm_abs t f) \/ exists c, exists n,  exists t, exists args', fully_applied_args_ctr e (c,n) args' /\ applied_value e /\In (c,n) l /\ sig_has_ctr (c,n) t S ). 
  eapply fully_app_type_gives_fully_app_ctr; eauto.
  apply constructs_base. unfold not. intros. inversion H3; subst. inversion H.
   destruct J.
    Case "e=tm_abs".
    destruct H3 as [tp [f Jeq]].
    subst;  inversion Hwf_e1; subst. inversion H. 
    Case "inv".
    destruct H3 as [c [n [c_tp [args1 [Jfully_app_ctr [Japp_val [Jin Jwf_c]]]]]]].
    assert (Jl:= tm_has_level c_tp).
    destruct Jl.
    assert (J: exists res, step_matches_branches e branches res).
    replace l with (nil++l) in Hget_ctrs; auto.
    eapply step_matches_branches_progress with (ctr0:=ctr0); eauto. 
    destruct  J as [res Jstep].
    exists res.  eapply step_mat; eauto. 
    
 Case "cong".
  destruct H4 as [e' Hstep].
  right. exists (tm_mat e' result_typ branches). apply step_mat_cong; auto.

Case "tm_cast".
  clear IHHwf_e2. 
 destruct IHHwf_e1; auto.
 Case "value".
  right. subst. exists e; apply step_cast; auto.
 right. destruct H0. exists (tm_cast x t2); apply step_cast_cong; auto.
 
Qed.      

